﻿/*
multiDAC Export Plugin
Copyright (C) 2009 Felix Schmitt

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using PluginInterfaceLibrary;
using InternalData;

namespace ExportPluginLib
{
    public class ExportPlugin : IExportPlugin
    {
        private const string MY_VERSION = "0.1.8";

        /** description of this plugin **/
        public string description
        {
            get { return "Plugin for exporting and analysing result data"; }
        }

        /** name of this plugin **/
        public string name
        {
            get { return "Standard export plugin"; }
        }

        /** vendor of this plugin **/
        public string vendor
        {
            get { return "Felix Schmitt"; }
        }

        /** version of this plugin **/
        public string version
        {
            get { return MY_VERSION; }
        }

        /** indicates if this plugin should be configured prior to each use **/
        public bool requiresConfig
        {
            get { return false; }
        }

        /** constructor **/
        public ExportPlugin()
        {
        }

        /** estructor **/
        ~ExportPlugin()
        {
        }

        /**
         * configures this plugin, may call a config form
         * \return ERR_NOERR on success
         */
        public int configure()
        {
            return ErrorCodes.ERR_NOERR;
        }

        /**
         * export result data
         * \param runs the data from the main program (array of type DataRun)
         * \param frameCount total number of available frames
         * \param getFrameCallback callback to retrieve a frame from the main program
         * \param renderViewCallback callback to render a view of exported data back to the main program
         * \return ERR_NOERR on success
         */
        public int exportData(object[] runs, int frameCount, GetFrameCallback getFrameCallback,
            RenderViewCallback renderViewCallback)
        {
            if (runs == null)
                return ErrorCodes.ERR_NULL;

            /*if (runs.Length == 0)
            {
                System.Windows.Forms.MessageBox.Show("No input data. Start a processing run first.",
                    "No data", System.Windows.Forms.MessageBoxButtons.OK,
                    System.Windows.Forms.MessageBoxIcon.Information);
                return ErrorCodes.ERR_NOERR;
            }*/

            if (renderViewCallback != null && getFrameCallback != null && frameCount > 0)
            {
                Bitmap buffer = generateVectorView(runs);
                if (buffer != null)
                    renderViewCallback(buffer);
                else
                    return ErrorCodes.ERR_NULL;
            }

            return ErrorCodes.ERR_NOERR;
        }

        // this is only pseudo-useful
        private Bitmap generateVectorView(object[] runs)
        {
            if (runs == null)
                return null;

            try
            {
                Bitmap buffer = new Bitmap(320, 240);
                Graphics g = Graphics.FromImage(buffer);

                float offsetY = 15;

                // background and grid
                g.SetClip(new RectangleF(0.0f, 0.0f, (float)buffer.Width, (float)buffer.Height));
                g.FillRectangle(Brushes.White, g.ClipBounds);
                g.DrawString("Analyzed vectors count (alpha)", new Font(FontFamily.GenericSansSerif, 10.0f),
                    Brushes.Black, g.ClipBounds.Width / 2 - 80, 5);
                g.DrawLine(Pens.Black, 5, 25, 5, g.ClipBounds.Height - offsetY);
                g.DrawLine(Pens.Black, 5, g.ClipBounds.Height - offsetY, g.ClipBounds.Width - 5, g.ClipBounds.Height - offsetY);
                
                g.DrawString("start", new Font(FontFamily.GenericSansSerif, 8.0f),
                    Brushes.Black, 2, g.ClipBounds.Height - 12);
                g.DrawString("end", new Font(FontFamily.GenericSansSerif, 8.0f),
                    Brushes.Black, g.ClipBounds.Width - 25, g.ClipBounds.Height - 12);


                float gridHeight = g.ClipBounds.Height - 45;
                float gridWidth = g.ClipBounds.Width - 10;

                Color[] colorArray = { Color.Red, Color.Blue, Color.Green, Color.Yellow, Color.Gray };

                int colorCounter = 0;

                for (int i = 0; i < runs.Length; i++)
                {
                    DataRun run = runs[i] as DataRun;
                    int packageCount = run.getPackageCount();

                    if (colorCounter == colorArray.Length)
                        colorCounter = 0;
                    
                    if (packageCount > 0)
                    {
                        int maxVectors = run.getPackage(0).vectors.Length;
                        float vecHeight = gridHeight / Convert.ToSingle(maxVectors);

                        PointF lastPoint = new PointF(0, 0);

                        for (int j = 0; j < packageCount; j++)
                        {
                            DataPackage package = run.getPackage(j);
                            float vectors = Convert.ToSingle(package.vectors.Length);

                            PointF currentPoint = new PointF(Convert.ToSingle(gridWidth * j / Convert.ToSingle(packageCount) + 5.0f),
                                g.ClipBounds.Height - offsetY - (vectors * vecHeight - 1));

                            Pen pen = new Pen(colorArray[colorCounter]);

                            if (j == 0)
                                g.DrawEllipse(pen, currentPoint.X, currentPoint.Y, 2, 2);
                            else
                                g.DrawLine(pen, lastPoint, currentPoint);

                            lastPoint = currentPoint;
                        }

                        colorCounter++;
                    }
                }

                return buffer;
            }
            catch (Exception)
            {
                return null;
            }
        }
    }
}
