﻿/*
multiDAC Image Import Plugin
Copyright (C) 2009 Felix Schmitt

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Collections;
using System.Drawing;
using System.Drawing.Imaging;
using InternalData;
using PluginInterfaceLibrary;

namespace ImportPluginLib
{
    /**
     * \class ImportedImage
     * A single image imported by an ImportPlugin
     */
    class ImportedImage : IImportedFile
    {
        private Image _srcImage = null;
        private string _srcFile = "";

        /** the low-level object of this file **/
        public object srcObject
        {
            get
            {
                return _srcImage;
            }
        }

        /** size of one frame **/
        public Size dimension
        {
            get
            {
                if (_srcImage != null)
                    return _srcImage.PhysicalDimension.ToSize();
                else
                    return new Size(0, 0);
            }
        }

        /** number of frames this file represents **/
        public int frameCount
        {
            get
            {
                if (_srcImage != null)
                    return 1;
                else
                    return 0;
            }
        }

        /** path to the file **/
        public string srcFile
        {
            get { return _srcFile; }
        }

        /**
         * constructor
         * \param _srcFile location of the file to load
         */
        public ImportedImage(string _srcFile)
        {
            if (File.Exists(_srcFile))
            {
                this._srcFile = _srcFile;
                    try
                    {
                        _srcImage = Image.FromFile(_srcFile);
                    }
                    catch (OutOfMemoryException e)
                    {
                        throw e;
                    }
            }
        }

        /**
         * renders one frame of this file
         * \param frameNum index of frame
         * \param graphics Graphics objct to render to
         * \return indicates if rendering was successfull
         */
        public int renderFrame(int frameNum, Graphics graphics)
        {
            if (frameNum < 0 || frameNum >= frameCount)
                return ErrorCodes.ERR_INDEXBOUNDS;

            try
            {
                graphics.DrawImageUnscaled((Image)srcObject, 0, 0);
            }
            catch (ArgumentNullException)
            {
                return ErrorCodes.ERR_NULL;
            }

            return ErrorCodes.ERR_NOERR;
        }

        /**
         * creates a internal Frame object
         * \param frameNum index of frame to normalize
         * \return Frame object or null
         */
        public Frame normalizeFrame(int frameNum)
        {
            if (frameNum < 0 || frameNum >= frameCount)
                return null;

            Frame target = new Frame(dimension);

            unsafe
            {
                //set the number of bytes per pixel
                int pixelSize = 3;
                Bitmap bmp = ((Bitmap)srcObject);
                BitmapData bmpData = bmp.LockBits(new Rectangle(0, 0, 
                    bmp.Width, bmp.Height),
                    ImageLockMode.WriteOnly, 
                    PixelFormat.Format24bppRgb);

                for (int j = 0; j < dimension.Height; j++)
                {
                    byte* row = (byte*)bmpData.Scan0 + (j * bmpData.Stride);

                    for (int i = 0; i < dimension.Width; i++)
                    {
                        target.data[i, j] = (int)((row[i * pixelSize] * .11) +
                            (row[i * pixelSize + 1] * .59) +
                            (row[i * pixelSize + 2] * .3));
                    }
                }

                bmp.UnlockBits(bmpData);
            }

            return target;
        }
    }

    /**
     * \class ImportPlugin
     * An implementation of the IImportPlugin interface for importing images
     */
    public class ImportPlugin : IImportPlugin
    {
        private const string MY_VERSION = "0.1.8";

        private ArrayList _importedFiles = new ArrayList();
        private string[] validExt = { ".jpg", ".jpeg", ".bmp", ".png", ".wmf", ".gif" };

        /** description of this plugin **/
        public string description
        {
            get { return "Plugin for importing bitmap files"; }
        }

        /** name of this plugin **/
        public string name
        {
            get { return "Standard image loader"; }
        }

        /** vendor of this plugin **/
        public string vendor
        {
            get { return "Felix Schmitt"; }
        }

        /** version of this plugin **/
        public string version
        {
            get { return MY_VERSION; }
        }

        /** indicates if this plugin should be configured prior to each use **/
        public bool requiresConfig
        {
            get { return false; }
        }

        /** total number of frames from all ImportedFiles loaded by this plugin **/
        public int frameCount
        {
            get
            {
                if (importedFiles == null || importedFiles.Length == 0)
                    return 0;

                int tmpCount = 0;
                foreach (ImportedImage importedFile in importedFiles)
                    tmpCount += importedFile.frameCount;
                return tmpCount;
            }
        }

        /** list of imported files **/
        public object[] importedFiles
        {
            get { return _importedFiles.ToArray(); }
        }

        /** locations of imported files **/
        public string[] importedFilenames
        {
            get 
            {
                List<string> files = new List<string>();
                foreach (ImportedImage file in importedFiles)
                    files.Add(file.srcFile);
                return files.ToArray();
            }
        }

        /** whether this plugin supports loading multiple files **/
        public bool supportsMultipleFiles
        {
            get { return true; }
        }

        /** constructor **/
        public ImportPlugin()
        {
        }

        ~ImportPlugin()
        {
            clearFiles();
        }

        /**
         * configures this plugin, may call a config form
         * \return ERR_NOERR on success
         */
        public int configure()
        {
            return ErrorCodes.ERR_NOERR;
        }

        /**
         * checks if a requested extension is valid
         * \param ext the extension to check
         */
        private bool validExtension(string ext)
        {
            ext = ext.ToLower();
            
            foreach (string vExt in validExt)
                if (ext.CompareTo(vExt) == 0)
                    return true;

            return false;
        }

        /**
         * imports files to this plugin
         * \param filenames array of filenames to load
         * \param progressCallback callback indicating loading progress
         * \return true on success
         */
        public int loadFiles(string[] filenames, ProgressCallback progressCallback)
        {
            if (filenames.Length == 0)
                return ErrorCodes.ERR_NOERR;

            for (int i = 0; i < filenames.Length; i++)
            {
                if (File.Exists(filenames[i]))
                {
                    if (!validExtension(Path.GetExtension(filenames[i])))
                    {
                        Console.WriteLine("Error: Failed to load file: " + filenames[i] + ". Wrong file format.");
                        continue;
                    }

                    ImportedImage importedImage;
                    try
                    {
                        importedImage = new ImportedImage(filenames[i]);
                        if (_importedFiles.Count == 0 || importedImage.dimension == (_importedFiles[0] as IImportedFile).dimension)
                        {
                            progressCallback((i + 1) * 100 / filenames.Length, 
                                "Adding file " + filenames[i]);
                            _importedFiles.Add(importedImage);
                        }
                        else
                        {
                            Console.WriteLine("Error: Failed to load file: " + filenames[i] + ". Wrong dimension.");
                            continue;
                        }
                    }
                    catch (OutOfMemoryException e)
                    {
                        Console.WriteLine(e.Message);
                        return ErrorCodes.ERR_OUTOFMEMORY;
                    }
                    catch (NotSupportedException e)
                    {
                        Console.WriteLine(e.Message);
                        return ErrorCodes.ERR_NOTSUPPORTED;
                    }
                    
                }
                else
                    Console.WriteLine("Error: File not found: " + filenames[i]);
            }

            if (frameCount == 0)
            {
                progressCallback(0, "No frames loaded.");
                return ErrorCodes.ERR_UNKNOWN;
            }

            if (importedFiles.Length < filenames.Length)
                progressCallback(0, "Could not load all requested files.");
            else
                progressCallback(0, "Done");

            return ErrorCodes.ERR_NOERR;
        }

        /** drops all imported files. makes sure that all file handles are released. **/
        public void clearFiles()
        {
            if (_importedFiles != null)
            {
                foreach (IImportedFile file in _importedFiles)
                    if (file != null && file.srcObject != null)
                        (file.srcObject as Image).Dispose();

                try
                {
                    _importedFiles.Clear();
                }
                catch (NotSupportedException e)
                {
                    Console.WriteLine(e.Message);
                }
            }
            GC.Collect();
        }

        /**
         * resolves a global frame index to its corresponding ImportedFile and file frame index
         * \param globalFrameNum global index
         * \param importedFile reference to corresponding file
         * \param fileFrameNum reference to corresponding file frame index
         * \return ERR_NOERR if globalFrameNum could be resolved
         */
        private int getInternalFrame(int globalFrameNum, ref ImportedImage importedFile, ref int fileFrameNum)
        {
            int counter = 0;
            importedFile = null;
            fileFrameNum = -1;

            if (globalFrameNum < 0 || globalFrameNum >= frameCount)
                return ErrorCodes.ERR_INDEXBOUNDS;

            foreach (ImportedImage file in importedFiles)
            {
                if (globalFrameNum < counter + file.frameCount)
                {
                    importedFile = file;
                    fileFrameNum = globalFrameNum - counter;
                    return ErrorCodes.ERR_NOERR;
                }
                counter += file.frameCount;
            }

            return ErrorCodes.ERR_NOTFOUND;
        }

        /**
         * renders a frame
         * \param frameNum global frame index
         * \param graphics Graphics object to render to
         * \return ERR_NOERR in success
         */
        public int renderFrame(int frameNum, Graphics graphics)
        {
            if (graphics == null)
                return ErrorCodes.ERR_NULL;

            ImportedImage file = null;
            int fileFrameNum = 0;

            if (getInternalFrame(frameNum, ref file, ref fileFrameNum) == ErrorCodes.ERR_NOERR)
                return file.renderFrame(fileFrameNum, graphics);

            return ErrorCodes.ERR_NOTFOUND;
        }

        /**
         * returns an internal Frame object
         * \param frameNum global frame index
         * \return Frame object or null
         */
        public Frame normalizeFrame(int frameNum)
        {
            ImportedImage file = null;
            int fileFrameNum = 0;

            if (getInternalFrame(frameNum, ref file, ref fileFrameNum) == ErrorCodes.ERR_NOERR)
                return file.normalizeFrame(fileFrameNum);

            return null;
        }

        /**
         * dimension of a frame
         * \param frameNum global frame index
         * \param frameDimension reference to Size object
         * \return ERR_NOERR on success
         */
        public int getDimension(int frameNum, ref Size frameDimension)
        {
            ImportedImage file = null;
            int fileFrameNum = 0;
            frameDimension = new Size(0, 0);

            if (getInternalFrame(frameNum, ref file, ref fileFrameNum) == ErrorCodes.ERR_NOERR)
            {
                frameDimension = file.dimension;
                return ErrorCodes.ERR_NOERR;
            }

            return ErrorCodes.ERR_NOTFOUND;
        }
    }
}
