﻿/*
multiDAC Video Import Plugin
Copyright (C) 2009 Felix Schmitt

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Runtime.InteropServices;
using PluginInterfaceLibrary;
using InternalData;

using DirectShowLib;

namespace ImportPluginLib
{
    /**
     * \class ImportedVideo
     * A single video imported by a VideoImportPlugin
     */
    class ImportedVideo : IImportedFile
    {
        /**
         * \class VideoDescription
         * Internal description of an imported video file using DirectShow's Graphbuilder interface
         */
        private class VideoDescription
        {
            private IGraphBuilder graphBuilder = null;
            private ISampleGrabber sampleGrabber = null;

            private int stride;
            public Size resolution { get; internal set; }
            public int frames { get; internal set; }

            /** close and free DirectShow interfaces **/
            private void CloseInterfaces()
            {
                lock (this)
                {
                    if (sampleGrabber != null)
                    {
                        Marshal.ReleaseComObject(sampleGrabber);
                        sampleGrabber = null;
                    }
                    if (graphBuilder != null)
                    {
                        Marshal.ReleaseComObject(graphBuilder);
                        graphBuilder = null;
                    }
                }

                GC.Collect();
            }

            /**
             * creates the filter graph
             * \param filename the file to load
             */
            private void setupGraph(string filename)
            {
                int hr = 0;
                try
                {
                    graphBuilder = new FilterGraph() as IGraphBuilder;
                    sampleGrabber = new SampleGrabber() as ISampleGrabber;

                    graphBuilder.AddFilter(sampleGrabber as IBaseFilter, "samplegrabber");

                    AMMediaType mt = new AMMediaType();
                    mt.majorType = MediaType.Video;
                    mt.subType = MediaSubType.RGB24;
                    mt.formatType = FormatType.VideoInfo;
                    sampleGrabber.SetMediaType(mt);

                    hr = graphBuilder.RenderFile(filename, null);
                    DsError.ThrowExceptionForHR(hr);

                    IMediaSeeking mediaSeek = (IMediaSeeking)graphBuilder;
                    IVideoWindow videoWin = (IVideoWindow)graphBuilder;

                    videoWin.put_AutoShow(OABool.False);
                    sampleGrabber.SetOneShot(true);
                    sampleGrabber.SetBufferSamples(true);

                    long duration = 0;
                    mediaSeek.SetTimeFormat(TimeFormat.Frame);
                    mediaSeek.GetDuration(out duration);
                    frames = Convert.ToInt32(duration);

                    AMMediaType grab = new AMMediaType();
                    sampleGrabber.GetConnectedMediaType(grab);
                    VideoInfoHeader videoheader = (VideoInfoHeader)Marshal.PtrToStructure(grab.formatPtr, typeof(VideoInfoHeader));

                    resolution = new Size(videoheader.BmiHeader.Width, videoheader.BmiHeader.Height);
                    stride = videoheader.BmiHeader.ImageSize / resolution.Height;

                    DsUtils.FreeAMMediaType(grab);
                }
                finally
                {
                    GC.Collect();
                }
            }

            /**
             * constructor
             * \param filename the file to load
             */
            public VideoDescription(string filename)
            {
                try
                {
                    setupGraph(filename);
                }
                catch (Exception)
                {
                    ErrorCodes.error(ErrorCodes.ERR_UNKNOWN);
                }
            }

            /** destructor **/
            ~VideoDescription()
            {
                CloseInterfaces();
            }

            /**
             * extracts a frame from the video
             * \param frame number of frame to extract
             * \return the extracted bitmap or null
             * \bug #0009 this is a bit slowly
             */
            public Bitmap getFrame(int frame)
            {
                if (sampleGrabber == null)
                    return null;

                try
                {
                    int hr = 0;

                    IMediaSeeking mediaSeek = graphBuilder as IMediaSeeking;
                    hr = mediaSeek.SetPositions(frame, AMSeekingSeekingFlags.AbsolutePositioning, 
                        frame, AMSeekingSeekingFlags.AbsolutePositioning);
                    DsError.ThrowExceptionForHR(hr);

                    IMediaControl mediaCtrl = graphBuilder as IMediaControl;
                    IMediaEventEx mediaEvent = graphBuilder as IMediaEventEx;

                    mediaCtrl.Run();
                    EventCode evcode;
                    mediaEvent.WaitForCompletion(100, out evcode);
                    DsError.ThrowExceptionForHR(hr);

                    // get frame
                    IntPtr ip = IntPtr.Zero;
                    int iBuffSize = 0;

                    hr = sampleGrabber.GetCurrentBuffer(ref iBuffSize, ip);
                    DsError.ThrowExceptionForHR(hr);

                    ip = Marshal.AllocCoTaskMem(iBuffSize);

                    hr = sampleGrabber.GetCurrentBuffer(ref iBuffSize, ip);
                    DsError.ThrowExceptionForHR(hr);

                    int iBufSize2 = resolution.Width * resolution.Height * 3;

                    return new Bitmap(resolution.Width, resolution.Height, -stride, PixelFormat.Format24bppRgb,
                        (IntPtr)(ip.ToInt32() + iBufSize2 - stride));
                }
                catch (Exception)
                {
                    return null;
                }
            }
        }

        private VideoDescription _srcVideo = null;
        private string _srcFile = "";
        private Bitmap buffer = null;
        private int bufferedFrame = -1;

        /** the low-level object of this file **/
        public object srcObject
        {
            get
            {
                return _srcVideo;
            }
        }

        /** size of one frame **/
        public Size dimension
        {
            get
            {
                if (_srcVideo != null)
                    return _srcVideo.resolution;
                else
                    return new Size(0, 0);
            }
        }

        /** number of frames this file represents **/
        public int frameCount
        {
            get
            {
                if (_srcVideo != null)
                    return _srcVideo.frames;
                else
                    return 0;
            }
        }

        /** path to the file **/
        public string srcFile
        {
            get { return _srcFile; }
        }

        /**
         * constructor
         * \param _srcFile location of the file to load
         */
        public ImportedVideo(string _srcFile)
        {
            if (File.Exists(_srcFile))
            {
                this._srcFile = _srcFile;
                try
                {
                    _srcVideo = new VideoDescription(_srcFile);
                }
                catch (OutOfMemoryException e)
                {
                    throw e;
                }
                catch (NotSupportedException e)
                {
                    throw e;
                }
                catch (Exception e)
                {
                    throw e;
                }
            }
        }

        /**
         * renders one frame of this file
         * \param frameNum index of frame
         * \param graphics Graphics objct to render to
         * \return indicates if rendering was successfull
         */
        public int renderFrame(int frameNum, Graphics graphics)
        {
            if (frameNum < 0 || frameNum >= frameCount)
                return ErrorCodes.ERR_INDEXBOUNDS;

            try
            {
                if (bufferedFrame != frameNum)
                {
                    buffer = _srcVideo.getFrame(frameNum);
                    bufferedFrame = frameNum;
                }

                graphics.DrawImageUnscaled((Image)buffer, 0, 0);
            }
            catch (ArgumentNullException)
            {
                return ErrorCodes.ERR_NULL;
            }

            return ErrorCodes.ERR_NOERR;
        }

        /**
         * creates a internal Frame object
         * \param frameNum index of frame to normalize
         * \return Frame object or null
         */
        public Frame normalizeFrame(int frameNum)
        {
            if (frameNum < 0 || frameNum >= frameCount)
                return null;

            Frame target = new Frame(dimension);

            unsafe
            {
                //set the number of bytes per pixel
                int pixelSize = 3;

                if (bufferedFrame != frameNum)
                {
                    buffer = _srcVideo.getFrame(frameNum);
                    bufferedFrame = frameNum;
                }

                Bitmap bmp = ((Bitmap)buffer);
                BitmapData bmpData = bmp.LockBits(new Rectangle(0, 0,
                    bmp.Width, bmp.Height),
                    ImageLockMode.WriteOnly,
                    PixelFormat.Format24bppRgb);

                for (int j = 0; j < dimension.Height; j++)
                {
                    byte* row = (byte*)bmpData.Scan0 + (j * bmpData.Stride);

                    for (int i = 0; i < dimension.Width; i++)
                    {
                        target.data[i, j] = (int)((row[i * pixelSize] * .11) +
                            (row[i * pixelSize + 1] * .59) +
                            (row[i * pixelSize + 2] * .3));
                    }
                }

                bmp.UnlockBits(bmpData);
            }

            return target;
        }
    }

    /**
     * \class VideoImportPlugin
     * An implementation of the IImportPlugin interface for importing videos
     */
    public class VideoImportPlugin : IImportPlugin
    {
        private const string MY_VERSION = "0.1.8";
        private ArrayList _importedFiles = new ArrayList();
        private string[] validExt = { ".avi", ".mpg", ".mpeg", ".wmv"};

        /** description of this plugin **/
        public string description
        {
            get { return "plugin for importing a video file"; }
        }

        /** name of this plugin **/
        public string name
        {
            get { return "Standard video loader"; }
        }

        /** vendor of this plugin **/
        public string vendor
        {
            get { return "Felix Schmitt"; }
        }

        /** version of this plugin **/
        public string version
        {
            get { return MY_VERSION; }
        }

        /** indicates if thsi plugin should be configured prior to each use **/
        public bool requiresConfig
        {
            get { return false; }
        }

        /** total number of frames from all ImportedFiles loaded by this plugin **/
        public int frameCount
        {
            get
            {
                if (importedFiles == null || importedFiles.Length == 0)
                    return 0;

                int tmpCount = 0;
                foreach (ImportedVideo importedFile in importedFiles)
                    tmpCount += importedFile.frameCount;
                return tmpCount;
            }
        }

        /** list of imported files **/
        public object[] importedFiles
        {
            get { return _importedFiles.ToArray(); }
        }

        /** locations of imported files **/
        public string[] importedFilenames
        {
            get
            {
                List<string> files = new List<string>();
                foreach (ImportedVideo file in importedFiles)
                    files.Add(file.srcFile);
                return files.ToArray();
            }
        }

        /** whether this plugin supports loading multiple files **/
        public bool supportsMultipleFiles
        {
            get { return false; }
        }

        /** constructor **/
        public VideoImportPlugin()
        {
        }

        /** destructor **/
        ~VideoImportPlugin()
        {
            clearFiles();
        }

        /**
         * configures this plugin, may call a config form
         * \return ERR_NOERR on success
         */
        public int configure()
        {
            return ErrorCodes.ERR_NOERR;
        }

        /**
         * checks if a requested extension is valid
         * \param ext the extension to check
         */
        private bool validExtension(string ext)
        {
            ext = ext.ToLower();

            foreach (string vExt in validExt)
                if (ext.CompareTo(vExt) == 0)
                    return true;

            return false;
        }

        /**
         * imports files to this plugin
         * \param filenames array of filenames to load
         * \param progressCallback callback indicating loading progress
         * \return true on success
         */
        public int loadFiles(string[] filenames, ProgressCallback progressCallback)
        {
            if (filenames.Length == 0)
                return ErrorCodes.ERR_NOERR;

            string filename = filenames[0];

            if (File.Exists(filename))
            {
                if (!validExtension(Path.GetExtension(filename)))
                {
                    Console.WriteLine("Error: Failed to load file: " + filename + ". Wrong file format.");
                    return ErrorCodes.ERR_FORMATERROR;
                }

                ImportedVideo importedVideo;
                try
                {
                    importedVideo = new ImportedVideo(filename);
                    if (_importedFiles.Count == 0)
                    {
                        progressCallback(50, "Adding file " + filename);
                        _importedFiles.Add(importedVideo);
                    }
                    else
                    {
                        Console.WriteLine("Error: Failed to load file: " + filename + ". Wrong dimension.");
                        return ErrorCodes.ERR_FORMATERROR;
                    }
                }
                catch (OutOfMemoryException)
                {
                    ErrorCodes.error(ErrorCodes.ERR_OUTOFMEMORY);
                    return ErrorCodes.ERR_OUTOFMEMORY;
                }
                catch (NotSupportedException)
                {
                    ErrorCodes.error(ErrorCodes.ERR_NOTSUPPORTED);
                    return ErrorCodes.ERR_NOTSUPPORTED;
                }
                catch (Exception)
                {
                    ErrorCodes.error(ErrorCodes.ERR_UNKNOWN);
                    return ErrorCodes.ERR_UNKNOWN;
                }

            }
            else
                Console.WriteLine("Error: File not found: " + filename);

            if (frameCount == 0)
            {
                progressCallback(0, "No frames loaded.");
                return ErrorCodes.ERR_UNKNOWN;
            }

            if (importedFiles.Length < 1)
                progressCallback(0, "Could not load all requested files.");
            else
                progressCallback(0, "Done");

            return ErrorCodes.ERR_NOERR;
        }

        /** drops all imported files. makes sure that all file handles are released. **/
        public void clearFiles()
        {
            if (_importedFiles != null)
                try
                {
                    _importedFiles.Clear();
                }
                catch (NotSupportedException e)
                {
                    Console.WriteLine(e.Message);
                }
            GC.Collect();
        }

        /**
         * resolves a global frame index to its corresponding ImportedFile and file frame index
         * \param globalFrameNum global index
         * \param importedFile reference to corresponding file
         * \param fileFrameNum reference to corresponding file frame index
         * \return ERR_NOERR if globalFrameNum could be resolved
         */
        private int getInternalFrame(int globalFrameNum, ref ImportedVideo importedFile, ref int fileFrameNum)
        {
            int counter = 0;
            importedFile = null;
            fileFrameNum = -1;

            if (globalFrameNum < 0 || globalFrameNum >= frameCount)
                return ErrorCodes.ERR_INDEXBOUNDS;

            foreach (ImportedVideo file in importedFiles)
            {
                if (globalFrameNum < counter + file.frameCount)
                {
                    importedFile = file;
                    fileFrameNum = globalFrameNum - counter;
                    return ErrorCodes.ERR_NOERR;
                }
                counter += file.frameCount;
            }

            return ErrorCodes.ERR_NOTFOUND;
        }

        /**
         * renders a frame
         * \param frameNum global frame index
         * \param graphics Graphics object to render to
         * \return ERR_NOERR in success
         */
        public int renderFrame(int frameNum, Graphics graphics)
        {
            if (graphics == null)
                return ErrorCodes.ERR_NULL;

            ImportedVideo file = null;
            int fileFrameNum = 0;

            if (getInternalFrame(frameNum, ref file, ref fileFrameNum) == ErrorCodes.ERR_NOERR)
                return file.renderFrame(fileFrameNum, graphics);

            return ErrorCodes.ERR_NOTFOUND;
        }

        /**
         * returns an internal Frame object
         * \param frameNum global frame index
         * \return Frame object or null
         */
        public Frame normalizeFrame(int frameNum)
        {
            ImportedVideo file = null;
            int fileFrameNum = 0;

            if (getInternalFrame(frameNum, ref file, ref fileFrameNum) == ErrorCodes.ERR_NOERR)
                return file.normalizeFrame(fileFrameNum);

            return null;
        }

        /**
         * dimension of a frame
         * \param frameNum global frame index
         * \param frameDimension reference to Size object
         * \return ERR_NOERR on success
         */
        public int getDimension(int frameNum, ref Size frameDimension)
        {
            ImportedVideo file = null;
            int fileFrameNum = 0;
            frameDimension = new Size(0, 0);

            if (getInternalFrame(frameNum, ref file, ref fileFrameNum) == ErrorCodes.ERR_NOERR)
            {
                frameDimension = file.dimension;
                return ErrorCodes.ERR_NOERR;
            }

            return ErrorCodes.ERR_NOTFOUND;
        }
    }
}
