﻿/*
multiDAC - Movement and deformation analysis tool
Defines some internal classes, structures and types
Copyright (C) 2009 Felix Schmitt

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

using System;
using System.Drawing;
using System.Collections;

namespace InternalData
{
    public static class ErrorCodes
    {
        public static int ERR_NOERR = 0;
        public static int ERR_UNKNOWN = -1;
        public static int ERR_NULL = -2;
        public static int ERR_OUTOFMEMORY = -3;
        public static int ERR_FILENOTFOUND = -4;
        public static int ERR_FORMATERROR = -5;
        public static int ERR_ACCESSDENIED = -6;
        public static int ERR_INDEXBOUNDS = -7;
        public static int ERR_CANCELLED = -8;
        public static int ERR_NOTSUPPORTED = -9;
        public static int ERR_NOTFOUND = -10;
        public static int ERR_IO = -11;
        public static int ERR_OVERFLOW = -12;

        public static void error(int code)
        {
            try
            {
                Console.WriteLine("Error: Code " + code.ToString());
#if DEBUG
                System.Windows.Forms.MessageBox.Show("Error: Code " + code.ToString(), "Debug error",
                    System.Windows.Forms.MessageBoxButtons.OK, System.Windows.Forms.MessageBoxIcon.Error);
#endif
            }
            catch (System.IO.IOException)
            {
            }
        }
    }

    /**
     * \struct Vector
     * a 2-dimensional vector
     */
    public struct Vector
    {
        public Point start;
        public Point target;
        /** returns if this vector has valid coordinates (!= -1) **/
        public bool valid
        {
            get { return (start.X > -1 && start.Y > -1 && target.X > -1 && target.Y > -1); }
        }
    }

    /**
     * \class AFrame
     * abstract description of a frame object
     */
    public abstract class AFrame
    {
        public Size dimension;
        public int[,] data;
    }

    /**
     * \class Frame
     * implementation of an internal Frame
     */
    public class Frame : AFrame
    {
        /**
         * constructor
         * \param dimension size of the frame
         */
        public Frame(Size dimension)
        {
            this.dimension = dimension;
            data = new int[dimension.Width, dimension.Height];
        }
    }

    /**
     * \class IndexedNote
     * represents a note at a specific index
     */
    public class IndexedNote
    {
        public int index;
        public string note;

        /** constructor **/
        public IndexedNote()
        {
            index = -1;
            note = "";
        }

        /**
         * set note values
         * \param _index index
         * \param _note note
         */
        public void setValues(int _index, string _note)
        {
            index = _index;
            note = _note;
        }
    }

    /**
     * \class DataRun
     * represents a (processing) run
     */
    public class DataRun
    {
        public ArrayList dataPackages { get; internal set; }
        public DateTime duration { get; internal set; }
        public int initialVectors { get; internal set; }
        public int firstFrame { get; internal set; }
        public int lastFrame { get; internal set; }
        public string plugin { get; internal set; }
        public ArrayList notes { get; internal set; }

        /**
         * constructor
         * \param initialVectors number of vectors to start the run with
         * \param firstFrame index of first frame
         * \param lastFrame index of lastFrame
         * \param plugin identifier of the processing plugin
         */
        public DataRun(int initialVectors, int firstFrame, int lastFrame, string plugin)
        {
            dataPackages = new ArrayList();
            this.initialVectors = initialVectors;
            this.firstFrame = firstFrame;
            this.lastFrame = lastFrame;
            this.plugin = plugin;
            notes = new ArrayList();
        }

        /**
         * set duration time
         * \param duration the duration this run needed
         */
        public void setDuration(DateTime duration)
        {
            this.duration = duration;
        }

        /**
         * adds a DataPackage (one processing step)
         * \param package the package to add
         */
        public void addPackage(DataPackage package)
        {
            dataPackages.Add(package);
        }

        /**
         * get a data package from this run
         * \param index index of the package
         * \return the data package or null
         */
        public DataPackage getPackage(int index)
        {
            if (index >= 0 && index < dataPackages.Count)
                return (dataPackages[index] as DataPackage);
            else
                return null;
        }

        /**
         * number of packages
         * \return number of packages
         */
        public int getPackageCount()
        {
            if (dataPackages != null)
                return dataPackages.Count;
            else
                return 0;
        }

        /** drops all packages **/
        public void clearData()
        {
            dataPackages.Clear();
        }

        public void addNote(int index, string note)
        {
            if (notes != null)
            {
                for (int i = 0; i < notes.Count; i++)
                    if ((notes[i] as IndexedNote).index == index)
                        notes.RemoveAt(i);

                IndexedNote aNote = new IndexedNote();
                aNote.setValues(index, note);
                notes.Add(aNote);
            }
        }
    }

    /**
     * \class ProjectDataPackage
     * This class is used for xml serialization.
     * A DataPackage object is mapped to this class and serialized.
     */
    public class ProjectDataPackage
    {
        public int frame1;
        public int frame2;
        
        public ProjectDataPackage()
        {
        }

        /**
         * map data package to this class
         * \param package data package
         */
        public void setValues(DataPackage package)
        {
            frame1 = package.frame1;
            frame2 = package.frame2;
        }
    }

    /**
     * \class ProjectDataRun
     * This class is used for xml serialization.
     * A DataRun object is mapped to this class and serialized.
     */
    public class ProjectDataRun
    {
        public DateTime duration;
        public int initialVectors;
        public int firstFrame;
        public int lastFrame;
        public string plugin;
        public int packageCount;
        public ProjectDataPackage[] packages;
        public IndexedNote[] notes;

        public ProjectDataRun()
        {
            
        }

        /**
         * map data run to this class
         * \param run data run
         */
        public void setValues(DataRun run)
        {
            if (run != null)
            {
                duration = run.duration;
                initialVectors = run.initialVectors;
                firstFrame = run.firstFrame;
                lastFrame = run.lastFrame;
                plugin = run.plugin;
                packageCount = run.getPackageCount();

                packages = new ProjectDataPackage[packageCount];
                for (int i = 0; i < packageCount; i++)
                {
                    packages[i] = new ProjectDataPackage();
                    packages[i].setValues(run.getPackage(i));
                }

                notes = new IndexedNote[run.notes.Count];
                for (int i = 0; i < run.notes.Count; i++)
                {
                    notes[i] = new IndexedNote();
                    notes[i].setValues((run.notes[i] as IndexedNote).index,
                        (run.notes[i] as IndexedNote).note);
                }
            }
        }
    }

    /**
     * \class DataPackage
     * Stores the information of one processing step (two frames compared)
     */
    public class DataPackage
    {
        private int _frame1;
        private int _frame2;
        public Vector[] vectors;

        /** get index of frame 1 **/
        public int frame1
        {
            get { return _frame1; }
            internal set { _frame1 = value; }
        }

        /** get index of frame 2 **/
        public int frame2
        {
            get { return _frame2; }
            internal set { _frame2 = value; }
        }

        /**
         * constructor
         * \param frm1 index of frame 1
         * \param frm2 index of frame 2
         * \param vectorNum number of vectors
         */
        public DataPackage(int frm1, int frm2, int vectorNum)
        {
            frame1 = frm1;
            frame2 = frm2;
            if (vectorNum >= 0)
                vectors = new Vector[vectorNum];
            else
                vectors = new Vector[0];

            // invalidate all vectors
            for (int i = 0; i < vectorNum; i++)
            {
                vectors[i].start = new Point(-1, -1);
                vectors[i].target = new Point(-1, -1);
            }
        }
    }

    /** a delegate method for a progress callback used in various methods **/
    public delegate void ProgressCallback(int progress, string msg);

    /**
     * \enum PluginReloadAction
     * Indicates result of a reload plugins operation
     **/
    public enum PluginReloadAction { praLoaded, praFailed, praCancel };

    /**
     * \enum DetectionMode
     * How to setup the starting DataPackage's vectors
     */
    public enum DetectionMode { dmAutomaticMotion, dmAutomaticCorner, dmManual };

    /**
     * \enum TrackingMode
     * How to treat resulting vectors after each run
     */
    public enum TrackingMode { tmAutomaticMotion, tmAutomaticCorner, tmFollowing, tmStart };
}
