/*
multiDAC - Movement and deformation analysis tool
GPU computation kernels
Copyright (C) 2009 Felix Schmitt

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifdef LINUX
#define __declspec(dllexport)
#define __stdcall
#endif

struct Frame
{
	int* data;
	float* median;
	float* medianSqr;
};

const int BLOCK_SIZE = 16;

int frameWidth = 0;
int frameHeight = 0;

int intSize = 0;
int floatSize = 0;

__device__
Frame frames[2];

int lastFrame1 = -1;
int lastFrame2 = -1;

__device__
int* vecs;

typedef void (__stdcall *ProgressCallback)(int);

extern "C" __declspec(dllexport) int init(int frmWidth, int frmHeight)
{
	frameWidth = frmWidth;
	frameHeight = frmHeight;

	int dimension = frameWidth * frameHeight;
	intSize = dimension * sizeof(int);
	floatSize = dimension * sizeof(float);

	lastFrame1 = -1;
	lastFrame2 = -1;

	for (int i = 0; i < 2; i++)
	{
		cudaMalloc((void**)&(frames[i].data), intSize);
		cudaMalloc((void**)&(frames[i].median), floatSize);
		cudaMalloc((void**)&(frames[i].medianSqr), floatSize);

		/*cudaMallocPitch((void**)&(frames[i].data), &(frames[i].pitchData), 
		sizeof(int) * frameWidth, frameHeight);
		cudaMallocPitch((void**)&(frames[i].median), &(frames[i].pitchMedian), 
		sizeof(float) * frameWidth, frameHeight);
		cudaMallocPitch((void**)&(frames[i].medianSqr), &(frames[i].pitchMedianSqr), 
		sizeof(float) * frameWidth, frameHeight);*/
	}

	return 0;
}

extern "C" __declspec(dllexport) void deInit()
{
	for (int i = 0; i < 2; i++)
	{
		cudaFree(frames[i].data);
		cudaFree(frames[i].median);
		cudaFree(frames[i].medianSqr);
	}
}

extern "C" __declspec(dllexport) int testDevice(int* maxThreadsPerBlock, int* maxThreadsDim, int* maxGridSize,
												int* major, int* minor)
{
	cudaDeviceProp prop;
	int dev;
	
	int devices = 0;
	if (cudaGetDeviceCount(&devices) != cudaSuccess || devices <= 0)
		return -1;
	
	if (cudaGetDevice(&dev) != cudaSuccess)
		return -1;

	if (cudaGetDeviceProperties(&prop, dev) == cudaSuccess)
	{
		if (prop.major == 9999 && prop.minor == 9999)
			return -2; 
	
		*maxThreadsPerBlock = prop.maxThreadsPerBlock;
		maxThreadsDim[0] = prop.maxThreadsDim[0];
		maxThreadsDim[1] = prop.maxThreadsDim[1];
		maxThreadsDim[2] = prop.maxThreadsDim[2];
		maxGridSize[0] = prop.maxGridSize[0];
		maxGridSize[1] = prop.maxGridSize[1];
		maxGridSize[2] = prop.maxGridSize[2];
		*major = prop.major;
		*minor = prop.minor;

		return 0;
	}
	else
		return -1;
} 

__global__
void medians(int* data, float* median, float* medianSqr,
			 int frmWidth, int frmHeight, int halfKernel)
{
	int x = blockIdx.x * blockDim.x + threadIdx.x;
	int y = blockIdx.y * blockDim.y + threadIdx.y;

	int index = y + x * frmHeight;
	int kernel = halfKernel * 2 + 1;
	float kernelSqr = powf(kernel, 2.0f);

	if (index < frmWidth * frmHeight)
	{
		if (x >= halfKernel && x < frmHeight - halfKernel && 
			y >= halfKernel && y < frmWidth - halfKernel)
		{
			float sum = 0.0f;
			float sumSqr = 0.0f;

			for (int dx = -halfKernel; dx <= halfKernel; dx++)
				for (int dy = -halfKernel; dy <= halfKernel; dy++)
				{
					int dIndex = (y + dy) + (x + dx) * frmHeight;
					int value = data[dIndex];
					sum += value;
					sumSqr += powf(value, 2.0f);
				}

				median[index] = (float)(sum / kernelSqr);
				medianSqr[index] = (float)(sumSqr / kernelSqr);
		}
	}
}

extern "C" __declspec(dllexport) int computeMedians(int frameNum, int frameIndex, int* data, 
													float* median, float* medianSqr, int halfKernel)
{
	if (frameNum >= 2)
		return -1;

	if (frameNum == 0)
	{
		lastFrame1 = frameIndex;

		if (lastFrame2 == frameIndex)
		{
			cudaMemcpy(frames[0].data, frames[1].data, intSize, cudaMemcpyDeviceToDevice);
			cudaMemcpy(frames[0].median, frames[1].median, floatSize, cudaMemcpyDeviceToDevice);
			cudaMemcpy(frames[0].medianSqr, frames[1].medianSqr, floatSize, cudaMemcpyDeviceToDevice);
			return 0;
		}
	}
	else
		lastFrame2 = frameIndex;

	cudaMemcpy(frames[frameNum].data, data, intSize, cudaMemcpyHostToDevice);

	//cudaMemcpy2D(frames[frameNum].data, sizeof(int) * frameWidth, data, frames[frameNum].pitchData, 
	//	sizeof(int) * frameWidth, frameHeight, cudaMemcpyHostToDevice);

	// this is not completely correct, as frameWidth & frameHeight can't be expected to 
	// be dividable by BLOCK_SIZE
	dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
	dim3 dimGrid(frameWidth / dimBlock.x, frameHeight / dimBlock.y);

	medians<<<dimGrid, dimBlock>>>(frames[frameNum].data,
		frames[frameNum].median, frames[frameNum].medianSqr, frameWidth, frameHeight, 
		halfKernel);

	if (cudaThreadSynchronize() == cudaSuccess)
	{	
		cudaMemcpy(median, frames[frameNum].median, floatSize, cudaMemcpyDeviceToHost);
		cudaMemcpy(medianSqr, frames[frameNum].medianSqr, floatSize, cudaMemcpyDeviceToHost);

		/*cudaMemcpy2D(data, sizeof(int) * frameWidth, frames[frameNum].data, frames[frameNum].pitchData, 
		sizeof(int) * frameWidth, frameHeight, cudaMemcpyDeviceToHost);
		cudaMemcpy2D(median, sizeof(float) * frameWidth, frames[frameNum].median, frames[frameNum].pitchMedian, 
		sizeof(float) * frameWidth, frameHeight, cudaMemcpyDeviceToHost);
		cudaMemcpy2D(medianSqr, sizeof(float) * frameWidth, frames[frameNum].medianSqr, frames[frameNum].pitchMedianSqr, 
		sizeof(float) * frameWidth, frameHeight, cudaMemcpyDeviceToHost);*/

		return 0;
	}
	else
		return -1;
}

__global__
void correlate1(int srcX, int srcY, float* bpCoeffs, size_t pitch, int bpCoeffWidth,
				int* data1, float* median1, float* medianSqr1,
				int* data2, float* median2, float* medianSqr2,
				int frmWidth, int frmHeight, int halfKernel, int maxDelta)
{
	int dx = blockIdx.x * blockDim.x + threadIdx.x;
	int dy = blockIdx.y * blockDim.y + threadIdx.y;
	int tx = threadIdx.x;
	int ty = threadIdx.y;

	// this should not happen
	if((dx + dy * bpCoeffWidth) > (bpCoeffWidth * bpCoeffWidth))
		return;

	int srcXdx = srcX + dx - maxDelta;
	int srcYdy = srcY + dy - maxDelta;

	__shared__ float s_median2[BLOCK_SIZE][BLOCK_SIZE];
	s_median2[ty][tx] = median2[srcXdx * frmHeight + srcYdy];

	__syncthreads();

	float* row = (float*)((char*)bpCoeffs + dy * pitch);
	float bpCoeff = 0.0f;
	float median1Val = median1[srcX * frmHeight + srcY];

	// just some deltas, not all
	if ((srcXdx >= halfKernel) && (srcXdx < (frmWidth - halfKernel)) &&
		(srcYdy >= halfKernel) && (srcYdy < (frmHeight - halfKernel)))
	{
		float medianCross = 0.0f;

		int medianSum = 0;
		for (int mdx = -halfKernel; mdx <= halfKernel; mdx++)
			for (int mdy = -halfKernel; mdy <= halfKernel; mdy++)
			{
				medianSum += data1[(srcX + mdx) * frmHeight + (srcY + mdy)] * 
					data2[(srcXdx + mdx) * frmHeight + (srcYdy + mdy)];
			}

			medianCross = (float)(medianSum / (powf(halfKernel * 2 + 1, 2.0f)));

			float sum1 = medianCross - median1Val *
				s_median2[ty][tx];
				//median2[srcXdx * frmHeight + srcYdy];

			float sum2 = (medianSqr1[(srcX) * frmHeight + srcY] - 
				powf(median1Val, 2.0f)) *
				(medianSqr2[srcXdx * frmHeight + srcYdy] - 
				//powf(median2[srcXdx * frmHeight + srcYdy], 2.0f));
				powf(s_median2[ty][tx], 2.0f));

			if (sum2 < 0.0f)
				sum2 = -sum2;
			sum2 = sqrtf(sum2);

			if (!isinf(sum1) && !isnan(sum1) && isfinite(sum2) && sum2 != 0.0f)
			{
				bpCoeff = (float)(sum1 / sum2);

				if (isinf(bpCoeff) || isnan(bpCoeff))
					bpCoeff = 0.0f;
				else
					if (bpCoeff < 0.0f)
						bpCoeff = -bpCoeff;
			}
	}

	row[dx] = bpCoeff;
}

__global__
void initBpCoeffs(float* bpCoeffs, size_t pitch, int width)
{
	int dx = blockIdx.x * blockDim.x + threadIdx.x;
	int dy = blockIdx.y * blockDim.y + threadIdx.y;

	if((dx + dy * width) > (width * width))
		return;

	float* row = (float*)((char*)bpCoeffs + dy * pitch);
	row[dx] = 0.0f;
}

__device__
float* bpCoeffs;

// this is the only stable version at the moment
extern "C" __declspec(dllexport) int computeCorrelation1(int* vectors, int vectorCount, int halfKernel, int maxDelta, 
														 float minCoefficient, ProgressCallback progressCallback)
{
	if (vectorCount > 0)
	{
		if (minCoefficient < 0.0f)
			minCoefficient = 0.0f;
		if (minCoefficient > 1.0f)
			minCoefficient = 1.0f;

		int bpCoeffWidth = maxDelta * 2 + 1;
		int allocWidth = bpCoeffWidth;
		while (allocWidth % BLOCK_SIZE != 0)
			allocWidth++;

		int hostMemSize = sizeof(float) * allocWidth * allocWidth;
		float* hostMem = NULL;
		if (cudaHostAlloc((void**)&hostMem, hostMemSize, cudaHostAllocDefault) != cudaSuccess)
			return -1;

		int step = vectorCount / 5;
		if (step == 0)
			step = 1;

		if (progressCallback != NULL)
			(*progressCallback)(0);

		size_t pitch;
		if (cudaMallocPitch((void**)&bpCoeffs, &pitch, sizeof(float) * allocWidth, allocWidth) != cudaSuccess)
		{
			cudaFreeHost(hostMem);
			return -1;
		}

		for(int i = 0; i < vectorCount; i++)
		{
			int srcX = vectors[i * 4];
			int srcY = vectors[i * 4 + 1];

			if ((srcX < halfKernel) || (srcX >= (frameWidth - halfKernel)) ||
				(srcY < halfKernel) || (srcY >= (frameHeight - halfKernel)))
			{
				vectors[i * 4 + 2] = -1;
				vectors[i * 4 + 3] = -1;
				continue;
			}

			dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
			dim3 dimGrid(allocWidth / dimBlock.x, allocWidth / dimBlock.y);

			correlate1<<<dimGrid, dimBlock>>>(srcX, srcY, bpCoeffs, pitch, bpCoeffWidth,
				frames[0].data, frames[0].median, frames[0].medianSqr,
				frames[1].data, frames[1].median, frames[1].medianSqr,
				frameWidth, frameHeight, halfKernel, maxDelta);

			if (cudaThreadSynchronize() == cudaSuccess)
			{
				if (cudaMemcpy2D(hostMem, sizeof(float) * allocWidth, bpCoeffs, pitch, 
					sizeof(float) * allocWidth, allocWidth, cudaMemcpyDeviceToHost) != cudaSuccess)
				{
					cudaFree(bpCoeffs);
					cudaFreeHost(hostMem);
					return -1;
				}

				float bpMax = 0.0f;
				int dxMax = 0;
				int dyMax = 0;

				for (int dy = 0; dy < bpCoeffWidth; dy++)	
					for (int dx = 0; dx < bpCoeffWidth; dx++)
					{
						if (hostMem[dx + dy * allocWidth] > bpMax)
						{
							bpMax = hostMem[dx + dy * allocWidth];
							dxMax = dx - maxDelta;
							dyMax = dy - maxDelta;
						}
					}

					if (bpMax > minCoefficient)
					{
						vectors[i * 4 + 2] = srcX + dxMax;
						vectors[i * 4 + 3] = srcY + dyMax;
					}
					else
					{
						vectors[i * 4 + 2] = -1;
						vectors[i * 4 + 3] = -1;
					}
			}
			else
			{
				cudaFree(bpCoeffs);
				cudaFreeHost(hostMem);
				return -1;
			}


			// this progressCallback is very error-prone, as it results in a kind of
			// threading error which causes the plugin to 'forget' about vectors.
			// it seems that this depends on step and vectorCount (certain conf.
			// work, others not).
			// however, until solved, this callback should be disabled by passing NULL.
			if ((progressCallback != NULL) && (i >= step) && ((i + 1) % step == 0))
				(*progressCallback)((i+1) * 100 / vectorCount);
		}

		cudaFree(bpCoeffs);
		cudaFreeHost(hostMem);

		if (progressCallback != NULL)
			(*progressCallback)(100);
	}
	else
		return -1;

	return 0;
}

__global__
void correlate2(int* vectors, int vectorCount, 
				int* data1, float* median1, float* medianSqr1,
				int* data2, float* median2, float* medianSqr2,
				int frmWidth, int frmHeight, int halfKernel, 
				int maxDelta, float minCoefficient)
{
	int vecIdx = blockIdx.x * blockDim.x + threadIdx.x;
	int vecIdy = blockIdx.y * blockDim.y + threadIdx.y;

	int vecId = vecIdx + vecIdy * vectorCount;

	if (fmodf(vecId, 4) != 0 || vecId >= vectorCount * 4)
		return;

	int startX = vecId;
	int startY = vecId + 1;
	int targetX = vecId + 2;
	int targetY = vecId + 3;

	int srcX = vectors[startX];
	int srcY = vectors[startY];

	if ((srcX < halfKernel) || (srcX >= (frmWidth - halfKernel)) ||
		(srcY < halfKernel) || (srcY >= (frmHeight - halfKernel)))
	{
		vectors[targetX] = -1;
		vectors[targetY] = -1;
		return;
	}

	float bpCoeffMax = 0.0f;
	int coeffMaxX = 0;
	int coeffMaxY = 0;

	for (int dx = -maxDelta; dx <= maxDelta; dx++)
		for (int dy = -maxDelta; dy <= maxDelta; dy++)
		{
			int srcXdx = srcX + dx;
			int srcYdy = srcY + dy;

			// just some deltas, not all
			if ((srcXdx >= halfKernel) && (srcXdx < (frmWidth - halfKernel)) &&
				(srcYdy >= halfKernel) && (srcYdy < (frmHeight - halfKernel)))
			{
				float medianCross = 0.0f;

				int medianSum = 0;
				for (int mdx = -halfKernel; mdx <= halfKernel; mdx++)
					for (int mdy = -halfKernel; mdy <= halfKernel; mdy++)
						medianSum += data1[(srcX + mdx) * frmHeight + (srcY + mdy)] * 
						data2[(srcXdx + mdx) * frmHeight + (srcYdy + mdy)];

				medianCross = (float)(medianSum / (powf(halfKernel * 2 + 1, 2.0f)));

				float sum1 = medianCross - median1[srcX * frmHeight + srcY] * 
					median2[srcXdx * frmHeight + srcYdy];

				float sum2 = (medianSqr1[(srcX) * frmHeight + srcY] - 
					powf(median1[srcX * frmHeight + srcY], 2.0f)) *
					(medianSqr2[srcXdx * frmHeight + srcYdy] - 
					powf(median2[srcXdx * frmHeight + srcYdy], 2.0f));

				if (sum2 < 0.0f)
					sum2 = -sum2;
				sum2 = sqrtf(sum2);

				float bpCoeff = 0.0f;
				if (!isinf(sum1) && !isnan(sum1) && isfinite(sum2) && sum2 != 0.0f)
				{
					bpCoeff = (float)(sum1 / sum2);

					if (isinf(bpCoeff) || isnan(bpCoeff))
						bpCoeff = 0.0f;
					else
						if (bpCoeff < 0.0f)
							bpCoeff = -bpCoeff;
				}

				if (bpCoeff > bpCoeffMax)
				{
					bpCoeffMax = bpCoeff;
					coeffMaxX = srcXdx;
					coeffMaxY = srcYdy;
				}
			}
		}

		if (bpCoeffMax > minCoefficient)
		{
			vectors[targetX] = coeffMaxX;
			vectors[targetY] = coeffMaxY;
		}
		else
		{
			vectors[targetX] = -1;
			vectors[targetY] = -1;
		}

}

extern "C" __declspec(dllexport) int computeCorrelation2(int* vectors, int vectorCount, int halfKernel, 
														 int maxDelta, float minCoefficient, 
														 ProgressCallback progressCallback)
{
	if (vectorCount > 0)
	{
		if (minCoefficient < 0.0f)
			minCoefficient = 0.0f;
		if (minCoefficient > 1.0f)
			minCoefficient = 1.0f;

		int vecsSize = vectorCount * 4 * sizeof(int);
		cudaMalloc((void**)&vecs, vecsSize);

		cudaMemcpy(vecs, vectors, vecsSize, cudaMemcpyHostToDevice);

		dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
		dim3 dimGrid(vectorCount * 4 / dimBlock.x, vectorCount * 4 / dimBlock.y);

		correlate2<<<dimGrid, dimBlock>>>(vecs, vectorCount, 
			frames[0].data, frames[0].median, frames[0].medianSqr,
			frames[1].data, frames[1].median, frames[1].medianSqr,
			frameWidth, frameHeight, halfKernel, maxDelta, minCoefficient);

		cudaMemcpy(vectors, vecs, vecsSize, cudaMemcpyDeviceToHost);

		cudaFree(vecs);
	}
	else
		return -1;

	return 0;
}

__global__
void correlate3(int* data1, int* data2, 
			   float* median1, float* median2,
			   float* medianSqr1, float* medianSqr2,
			   int frmWidth, int frmHeight, int halfKernel, 
			   int maxDelta, float minCoefficient, int line, int2* results)
{
	int tx = threadIdx.x;
	int ty = threadIdx.y;
	int x = blockIdx.x * blockDim.x + threadIdx.x;
	int y = blockIdx.y * blockDim.y + threadIdx.y;

	y += line * BLOCK_SIZE;

	if (x < 0 || y < 0 || x >= frmWidth || y > frmHeight)
		return;

	//extern __shared__ int s_results[];
	__shared__ int2 s_results[BLOCK_SIZE][BLOCK_SIZE];

	__syncthreads();


	if ((x < halfKernel) || (x >= (frmWidth - halfKernel)) ||
		(y < halfKernel) || (y >= (frmHeight - halfKernel)))
	{
		s_results[ty][tx] = make_int2(-1, -1);
	}
	else
	{
		s_results[ty][tx] = make_int2(-1, -1);

		float bpCoeff = 0.0f;
		float bpCoeffMax = 0.0f;

		for (int dx = -maxDelta; dx <= maxDelta; dx++)
		{
			for (int dy = -maxDelta; dy <= maxDelta; dy++)
			{
				int xdx = x + dx;
				int ydy = y + dy;

				// just some deltas, not all
				if ((xdx >= halfKernel) && (xdx < (frmWidth - halfKernel)) &&
					(ydy >= halfKernel) && (ydy < (frmHeight - halfKernel)))
				{
					int medianSum = 0;
					for (int mdx = -halfKernel; mdx <= halfKernel; mdx++)
						for (int mdy = -halfKernel; mdy <= halfKernel; mdy++)
							medianSum += data1[(y + mdy) * frmWidth + (x + mdx)] * 
							data2[(ydy + mdy) * frmWidth + (xdx + mdx)];

					float sum1 = ((medianSum / (powf(halfKernel * 2 + 1, 2.0f))) - 
						median1[y * frmWidth + x] * 
						median2[ydy * frmWidth + xdx]);

					float sum2 = (medianSqr1[y * frmWidth + x] - 
						powf(median1[y * frmWidth + x], 2.0f)) *
						(medianSqr2[ydy * frmWidth + xdx] - 
						powf(median2[ydy * frmWidth + xdx], 2.0f));

					if (sum2 < 0.0f)
						sum2 = -sum2;
					sum2 = sqrtf(sum2);

					if (!isinf(sum1) && !isnan(sum1) && isfinite(sum2) && sum2 != 0.0f)
					{
						bpCoeff = (float)(sum1 / sum2);

						if (isinf(bpCoeff) || isnan(bpCoeff))
							bpCoeff = 0.0f;
						else
							if (bpCoeff < 0.0f)
								bpCoeff = -bpCoeff;
					}

					if (bpCoeff > bpCoeffMax)
					{
						bpCoeffMax = bpCoeff;
						s_results[ty][tx] = make_int2(xdx, ydy);
					}
				}
			}
		}

		/*if (bpCoeffMax > float minCoefficient)
		{
			s_results[ty][tx] = make_int2(coeffMaxX, coeffMaxY);
		}
		else
		{
			s_results[ty][tx] = make_int2(-1, -1);
		}*/
	}

	__syncthreads();

	results[y * frmWidth + x] = s_results[ty][tx];
}

__device__ int2* results;

extern "C" __declspec(dllexport) int computeCorrelation3(int* vectors, int vectorCount, int halfKernel, int maxDelta,
											 int minCoefficient, ProgressCallback progressCallback)
{	
	if (minCoefficient < 0)
		minCoefficient = 0;
	if (minCoefficient > 100)
		minCoefficient = 100;

	int memSize = sizeof(int2) * frameWidth * frameHeight;
	//int memOffset = frameWidth * frameHeight;

	int2* hostMem = (int2*)malloc(memSize);
	cudaMalloc((void**)&results, memSize);

	//int threadsPerBlock = BLOCK_SIZE * BLOCK_SIZE;
	//int sharedMemSize = threadsPerBlock * sizeof(int2);

	dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
	dim3 dimGrid(frameWidth / dimBlock.x, 1);

	if (progressCallback != NULL)
		(*progressCallback)(0);

	for (int line = 0; line < (frameHeight / BLOCK_SIZE); line++)
	{
		correlate3<<<dimGrid, dimBlock>>>(frames[0].data, frames[1].data,
			frames[0].median, frames[1].median, frames[0].medianSqr, frames[1].medianSqr,
			frameWidth, frameHeight, halfKernel, maxDelta, (float)minCoefficient / 100.0f, line, results);

		if (progressCallback != NULL)
			(*progressCallback)((line + 1) * 100 / (frameHeight / BLOCK_SIZE));
	}

	if (cudaThreadSynchronize() == cudaSuccess)
	{
		cudaMemcpy(hostMem, results, memSize, cudaMemcpyDeviceToHost);

		for (int i = 0; i < vectorCount; i++)
		{
			int base = i * 4;
			int x = vectors[base];
			int y = vectors[base + 1];

			vectors[base + 2] = hostMem[y * frameWidth + x].x;
			vectors[base + 3] = hostMem[y * frameWidth + x].y;
		}
	}

	if (progressCallback != NULL)
		(*progressCallback)(100);

	cudaFree(results);
	free(hostMem);
	return 0;
}