﻿/*
multiDAC - Movement and deformation analysis tool
Class for binary input/output
Copyright (C) 2009 Felix Schmitt

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using InternalData;

namespace Manager
{
    /**
     * \class BinaryIO
     * An abstract class used for binary I/O operations
     */
    public abstract class BinaryIO
    {
        /**
         * \enum BinaryIOMode
         * mode this class is currently used with
         */
        protected enum BinaryIOMode { ioWrite, ioRead };

        protected bool fileOpen;
        private BinaryIOMode mode;
        private BinaryWriter writer;
        private BinaryReader reader;
        
        /** constructor **/
        public BinaryIO()
        {
            fileOpen = false;
            writer = null;
            reader = null;
            mode = BinaryIOMode.ioRead;
        }

        /**
         * opens a file
         * \param filename file to open
         * \param mode the mode to open the file with (read/write)
         * \return ERR_NOERR on success
         */
        protected int open(string filename, BinaryIOMode mode)
        {
            if (fileOpen)
                return ErrorCodes.ERR_ACCESSDENIED;

            try
            {
                if (mode == BinaryIOMode.ioWrite)
                    writer = new BinaryWriter(new FileStream(filename, FileMode.OpenOrCreate));
                else
                    reader = new BinaryReader(new FileStream(filename, FileMode.Open));
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
                fileOpen = false;
                return ErrorCodes.ERR_UNKNOWN;
            }

            this.mode = mode;
            fileOpen = true;
            return ErrorCodes.ERR_NOERR;
        }

        /** closes opened files **/
        protected void close()
        {
            if (fileOpen)
            {
                if (mode == BinaryIOMode.ioWrite && writer != null)
                    writer.Close();

                if (mode == BinaryIOMode.ioRead && reader != null)
                    reader.Close();
            }

            fileOpen = false;
        }

        /**
         * writes version information to the file
         * \param version version number
         * \return ERR_NOERR on success
         */
        protected int writeVersionHeader(int version)
        {
            if (!fileOpen || mode != BinaryIOMode.ioWrite)
                return ErrorCodes.ERR_NOTSUPPORTED;

            try
            {
                writer.Write(version);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
                return ErrorCodes.ERR_ACCESSDENIED;
            }
            return ErrorCodes.ERR_NOERR;
        }

        /**
         * reads version information from file
         * \return version number or error code (negative)
         */
        protected int readVersionHeader()
        {
            if (!fileOpen || mode != BinaryIOMode.ioRead)
                return ErrorCodes.ERR_NOTSUPPORTED;

            int result = ErrorCodes.ERR_UNKNOWN;

            try
            {
                result = reader.ReadInt32();
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
                return ErrorCodes.ERR_ACCESSDENIED;
            }

            return result;
        }

        /**
         * writes a data package
         * \param package data package
         * \return ERR_NOERR on success
         */
        protected int writePackage(DataPackage package)
        {
            if (!fileOpen || mode != BinaryIOMode.ioWrite)
                return ErrorCodes.ERR_NOTSUPPORTED;

            if (package != null)
            {
                try
                {
                    writer.Write(package.frame1);
                    writer.Write(package.frame2);
                    writer.Write(package.vectors.Length);

                    foreach (Vector vector in package.vectors)
                    {
                        writer.Write(vector.start.X);
                        writer.Write(vector.start.Y);
                        writer.Write(vector.target.X);
                        writer.Write(vector.target.Y);
                    }
                }
                catch (Exception e)
                {
                    Console.WriteLine(e.Message);
                    return ErrorCodes.ERR_ACCESSDENIED;
                }
            }

            return ErrorCodes.ERR_NOERR;
        }

        /**
         * reads a data package
         * \param package reference to a data package
         * \return ERR_NOERR on success
         */
        protected int readPackage(ref DataPackage package)
        {
            if (!fileOpen || mode != BinaryIOMode.ioRead)
                return ErrorCodes.ERR_NOTSUPPORTED;

            try
            {
                int[] head = new int[3];
                head[0] = reader.ReadInt32();
                head[1] = reader.ReadInt32();
                head[2] = reader.ReadInt32();

                package = new DataPackage(head[0], head[1], head[2]);
                if (package == null)
                    return ErrorCodes.ERR_NULL;

                for (int i = 0; i < package.vectors.Length; i++)
                {
                    int[] buffer = new int[4];
                    buffer[0] = reader.ReadInt32();
                    buffer[1] = reader.ReadInt32();
                    buffer[2] = reader.ReadInt32();
                    buffer[3] = reader.ReadInt32();
                    
                    package.vectors[i].start.X = buffer[0];
                    package.vectors[i].start.Y = buffer[1];
                    package.vectors[i].target.X = buffer[2];
                    package.vectors[i].target.Y = buffer[3];
                }

            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
                return ErrorCodes.ERR_ACCESSDENIED;
            }

            return ErrorCodes.ERR_NOERR;
        }
    }

    /**
     * \class DataWriter
     * Implementation of the abstract BinaryIO class for writing data
     */
    public class DataWriter : BinaryIO
    {
        private const int version = 1;
        private string filename;

        /**
         * constructor
         * \param filename file to write to
         */
        public DataWriter(string filename)
        {
            this.filename = filename;
        }

        /**
         * writes all data to the file
         * \param dataProcessor DataProcessor object holding information to write
         * \param progressCallback callback function indicating write progress
         * \return ERR_NOERR on success
         */
        public int write(DataProcessor dataProcessor, ProgressCallback progressCallback)
        {
            int errResult = ErrorCodes.ERR_NOERR;

            if (dataProcessor == null)
                return ErrorCodes.ERR_NULL;

            if (progressCallback != null)
                progressCallback(0, "Saving to " + filename);

            errResult = open(filename, BinaryIOMode.ioWrite);
            if (errResult != ErrorCodes.ERR_NOERR)
                return errResult;

            errResult = writeVersionHeader(version);
            if (errResult != ErrorCodes.ERR_NOERR)
                return errResult;

            foreach (DataRun run in dataProcessor.dataRuns)
            {
                int packageCount = run.getPackageCount();

                for (int i = 0; i < packageCount; i++)
                {
                    DataPackage package = run.getPackage(i);
                    errResult = writePackage(package);

                    if (errResult != ErrorCodes.ERR_NOERR)
                    {
                        if (progressCallback != null)
                            progressCallback(0, "Failed to write to " + filename);
                        return errResult;
                    }

                    if (progressCallback != null)
                        progressCallback((i + 1) * 100 / packageCount, "Saving to " + filename);
                }
            }

            close();

            if (progressCallback != null)
                progressCallback(0, "Done");

            return ErrorCodes.ERR_NOERR;
        }
    }

    /**
     * \class DataReader
     * Implementation of the abstract BinaryIO class for reading data
     */
    public class DataReader : BinaryIO
    {
        private const int version = 1;
        private string filename;

        /**
         * constructor
         * \param filename file to read from
         */
        public DataReader(string filename)
        {
            this.filename = filename;
        }

        /**
         * reads all data from file
         * \param dataProcessor DataProcessor object to write data to
         * \param processingRuns array of processing runs
         * \param progressCallback callback function indication reading progress
         * \return ERR_NOERR on success
         */
        public int read(DataProcessor dataProcessor, ProjectDataRun[] processingRuns,
            ProgressCallback progressCallback)
        {
            int errResult = ErrorCodes.ERR_NOERR;

            if (dataProcessor == null)
                return ErrorCodes.ERR_NULL;

            if (progressCallback != null)
                progressCallback(0, "Loading from " + filename);

            errResult = open(filename, BinaryIOMode.ioRead);
            if (errResult != ErrorCodes.ERR_NOERR)
                return errResult;

            errResult = readVersionHeader();
            if (errResult != version) 
                return ErrorCodes.ERR_FORMATERROR;

            for (int j = 0; j < dataProcessor.dataRuns.Count; j++)
            {
                DataRun run = dataProcessor.getRun(j);
                int packageCount = processingRuns[j].packageCount;

                for (int i = 0; i < packageCount; i++)
                {
                    DataPackage package = new DataPackage(processingRuns[j].packages[i].frame1,
                        processingRuns[j].packages[i].frame2, 0);

                    errResult = readPackage(ref package);
                    if (errResult != ErrorCodes.ERR_NOERR)
                    {
                        if (progressCallback != null)
                            progressCallback(0, "Failed to read from " + filename);
                        return errResult;
                    }

                    run.addPackage(package);

                    if (progressCallback != null)
                        progressCallback((i + 1) * 100 / packageCount, "Loading from " + filename);
                }
            }

            close();

            if (progressCallback != null)
                progressCallback(0, "Done");

            return ErrorCodes.ERR_NOERR; ;
        }
    }

}
