﻿/*
multiDAC - Movement and deformation analysis tool
Class for managing program's configuration
Copyright (C) 2009 Felix Schmitt

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Xml;
using System.Xml.Serialization;
using System.Windows.Forms;
using InternalData;

namespace Manager
{
    /** delegate for callback method for loading/saving configuration **/ 
    public delegate void ConfigurationCallback(Configuration configuration);

    /**
     * \class Configuration
     * Represents a program configuration
     */
    public class Configuration
    {
        public string pathToPlugins;
        public int renderingLineWidth;
        public bool vectorsInSplitScreen;
        public bool includeSourceFilesByDefault;
        public bool askBeforeClosingModified;
        public bool askBeforeCancellingProcessing;
        public DetectionMode detectionMode;
        public TrackingMode trackingMode;
        public int maxVectors;

        private int _detectionSensitivity;

        public int DetectionSensitivity
        {
            get { return _detectionSensitivity; }
            set
            {
                if (value >= 1 && value <= 3)
                    _detectionSensitivity = value;
                else
                    _detectionSensitivity = 2;
            }
        }

        /** constructor **/
        public Configuration()
        {
            pathToPlugins = "";
            renderingLineWidth = 1;
            vectorsInSplitScreen = true;
            includeSourceFilesByDefault = false;
            askBeforeClosingModified = true;
            askBeforeCancellingProcessing = true;
            detectionMode = DetectionMode.dmAutomaticMotion;
            trackingMode = TrackingMode.tmAutomaticMotion;
            DetectionSensitivity = 2;
            maxVectors = 25;
        }
    }

    /**
     * \class ConfigManager
     * Manages the program configuration
     */
    public class ConfigManager
    {
        public Configuration configuration { get; internal set; }

        /** indicates if a configuration is loaded **/
        public bool loaded { get { return configuration != null; } }

        /** constructor **/
        public ConfigManager()
        {
            configuration = null;
        }

        /**
         * loads program configuration from file
         * \param filename file to load from
         * \param configCallback callback function which applies configuration to program
         * \param progressCallback callback function indicating loading progress
         * \return ERR_NOERR on success
         */
        public int load(string filename, ConfigurationCallback configCallback,
            ProgressCallback progressCallback)
        {
            if (progressCallback != null)
                progressCallback(0, "Loading configuration from " + filename);

            if (!File.Exists(filename))
            {
                configuration = new Configuration();
                return ErrorCodes.ERR_FILENOTFOUND;
            }

            XmlSerializer xmlSer = new XmlSerializer(typeof(Configuration));
            if (xmlSer == null)
                return ErrorCodes.ERR_NULL;

            TextReader stream;
            try
            {
                stream = new StreamReader(filename);
            }
            catch (Exception e)
            {
#if DEBUG
                MessageBox.Show(e.Message);
#endif
                MessageBox.Show("Failed to open " + filename + ".", "Error", MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
                return ErrorCodes.ERR_UNKNOWN;
            }

            try
            {
                configuration = (Configuration)xmlSer.Deserialize(stream);
                stream.Close();
                if (configuration == null)
                    return ErrorCodes.ERR_NULL;
            }
            catch (InvalidOperationException e)
            {
#if DEBUG
                MessageBox.Show(e.Message);
#endif
                MessageBox.Show("Failed to read configuration.", "Error", MessageBoxButtons.OK,
                    MessageBoxIcon.Error);

                // setup a default config so program can work
                configuration = new Configuration();
                return ErrorCodes.ERR_NOTSUPPORTED;
            }

            stream.Close();

            if (configCallback != null)
                configCallback(configuration);

            if (progressCallback != null)
                progressCallback(100, "Loading configuration from " + filename);

            return ErrorCodes.ERR_NOERR;
        }

        /**
         * writes program configuration to file
         * \param filename file to save to
         * \param configCallback callback function which applies program settings to Configuration object
         * \param progressCallback callback function indicating saving progress
         * \return ERR_NOERR on success
         */
        public int save(string filename, ConfigurationCallback configCallback,
            ProgressCallback progressCallback)
        {
            if (progressCallback != null)
                progressCallback(0, "Saving configuration to " + filename);

            XmlSerializer xmlSer;
            try
            {
                xmlSer = new XmlSerializer(typeof(Configuration));
            }
            catch (InvalidOperationException e)
            {
                MessageBox.Show("This is an internal bug.", "Error", MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
#if DEBUG
                MessageBox.Show(e.InnerException.Message);
#endif
                return ErrorCodes.ERR_NOTSUPPORTED;
            }

            if (xmlSer == null)
                return ErrorCodes.ERR_NULL;

            if (configuration == null)
                configuration = new Configuration();

            if (configCallback != null)
                configCallback(configuration);

            TextWriter stream;
            try
            {
                stream = new StreamWriter(filename);
            }
            catch (Exception e)
            {
#if DEBUG
                MessageBox.Show(e.Message);
#endif
                MessageBox.Show("Failed to open " + filename + ".", "Error", MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
                return ErrorCodes.ERR_UNKNOWN;
            }

            try
            {
                xmlSer.Serialize(stream, configuration);
            }
            catch (InvalidOperationException e)
            {
#if DEBUG
                MessageBox.Show(e.InnerException.Message);
#endif
                MessageBox.Show("Failed to write configuration.", "Error", MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
            }
            stream.Close();

            if (progressCallback != null)
                progressCallback(100, "Saving configuration to " + filename);

            return ErrorCodes.ERR_NOERR;
        }

        /**
         * displays the settings dialog
         * \param configCallback callback function which applies configuration to program
         * \param startTab index of the selected tab page of settingsForm
         * \return dialog result
         */
        public DialogResult showSettingsForm(int startTab, ConfigurationCallback configCallback)
        {
            ProgramSettingsForm settingsForm = new ProgramSettingsForm();
            settingsForm.refreshProgramSettings(startTab, configuration);

            DialogResult result = settingsForm.ShowDialog();

            if (result == DialogResult.OK)
            {
                if (configCallback != null)
                    configCallback(configuration);
            }

            settingsForm.Dispose();

            return result;
        }
    }
}
