﻿/*
multiDAC - Movement and deformation analysis tool
Management class for processing input data (calling processing plugins)
Copyright (C) 2009 Felix Schmitt

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.ComponentModel;
using InternalData;
using PluginInterfaceLibrary;

using AForge.Imaging;
using AForge.Vision.Motion;

namespace Manager
{
    /**
     * \class DataProcessor
     * Manager class for processing and storing results
     */
    public class DataProcessor
    {
        public delegate void TimeCallback(DateTime time);

        /**
         * \class FrameRange
         * Describes start and end frame of a processing run
         */
        public class FrameRange
        {
            public int start;
            public int end;

            /**
             * constructor
             * \param start start frame
             * \param end end frame
             */
            public FrameRange(int start, int end)
            {
                this.start = start;
                this.end = end;
            }
        }

        /**
         * \class DetectionParameters
         * provides parameters for classes used in motion/poi detection via AForge.NET
         */
        private class DetectionParameters
        {
            private int _sensitivity;

            public int gridSize { get; internal set; }
            public float detectionTrigger { get; internal set; }
            public int diffThreshold { get; internal set; }
            public int geomThreshold { get; internal set; }

            public int Sensitivity
            {
                get { return _sensitivity; }
                set
                {
                    switch (value)
                    {
                        case 1:
                            _sensitivity = value;
                            gridSize = 8;
                            detectionTrigger = 0.1f;
                            diffThreshold = 35;
                            geomThreshold = 10;
                            break;
                        case 3:
                            _sensitivity = value;
                            gridSize = 32;
                            detectionTrigger = 0.005f;
                            diffThreshold = 10;
                            geomThreshold = 30;
                            break;
                        default:
                            _sensitivity = value;
                            gridSize = 16;
                            detectionTrigger = 0.01f;
                            diffThreshold = 25;
                            geomThreshold = geomThreshold = 20;
                            break;
                    }
                }
            }

            /**
             * constructor
             * \param sensitivity ranging from 1 (low) to 3 (high)
             */
            public DetectionParameters(int sensitivity)
            {
                Sensitivity = sensitivity;
            }
        }

        public ArrayList dataRuns { get; internal set; }
        public int currentRunIndex { get; internal set; }
        
        private int _vectorCount;
        private bool cancellationPending = false;
        private bool cancellationForced = false;
        private BackgroundWorker backgroundWorker = null;
        private ProgressCallback internalProgressCallback;

        /** constructor **/
        public DataProcessor()
        {
            dataRuns = new ArrayList();
            createBackgroundWorker();
        }

        /**
         * cancels running processing operations
         * \param force indicates that pending result data is not needed anymore
         */
        public void cancelProcessing(bool force)
        {
            cancellationPending = true;
            cancellationForced = force;
        }

        /** creates the BackgroundWorker Thread object for asynchronious operations **/
        private void createBackgroundWorker()
        {
            if (backgroundWorker != null)
            {
                backgroundWorker.Dispose();
                backgroundWorker = null;
            }

            backgroundWorker = new BackgroundWorker();
            backgroundWorker.WorkerReportsProgress = true;
            backgroundWorker.WorkerSupportsCancellation = false;
            backgroundWorker.DoWork += worker_DoWork;
            backgroundWorker.ProgressChanged += worker_ProgressChanged;
        }

        /** returns the currently selected DataRun object or null **/
        public DataRun currentRun
        {
            get
            {
                if (dataRuns != null && dataRuns.Count > currentRunIndex)
                    return (dataRuns[currentRunIndex] as DataRun);
                else
                    return null;
            }
        }

        /**
         * sets the current data run
         * \param index run index
         * \return ERR_NOERR on success
         */
        public int setRun(int index)
        {
            if (index >= 0 && index < dataRuns.Count)
            {
                currentRunIndex = index;
                return ErrorCodes.ERR_NOERR;
            }
            else
                return ErrorCodes.ERR_INDEXBOUNDS;
        }

        /**
         * creates a new DataRun object and adds it to the dataRuns list
         * \param initialVectors number of vectors
         * \param firstFrame first frame
         * \param lastFrame last frame
         * \param plugin processing plugin identifier
         * \return created DataRun object
         */
        public DataRun addRun(int initialVectors, int firstFrame, int lastFrame, string plugin)
        {
            DataRun newRun = new DataRun(initialVectors, firstFrame, lastFrame, plugin);
            try
            {
                dataRuns.Add(newRun);
            }
            catch (NotSupportedException e)
            {
                throw e;
            }

            currentRunIndex = dataRuns.Count - 1;
            return newRun;
        }

        /**
         * returns a package from a run
         * \param runIndex index of run
         * \param packageIndex index of package in run
         * \return the data package or null
         */
        public DataPackage getPackage(int runIndex, int packageIndex)
        {
            DataRun run = getRun(runIndex);
            if (run != null)
                return run.getPackage(packageIndex - run.firstFrame);
            else
                return null;
        }

        /**
         * returns a data run
         * \param index index of run
         * \return the data run or null
         */
        public DataRun getRun(int index)
        {
            if (index >= 0 && index < dataRuns.Count)
                return (dataRuns[index] as DataRun);
            else
                return null;
        }

        /**
         * returns number of runs in the list
         * \return number of runs
         */
        public int getRunsCount()
        {
            if (dataRuns != null)
                return dataRuns.Count;
            else
                return 0;
        }

        /** drops all runs **/
        public void clearData()
        {
            if (dataRuns != null)
            {
                foreach (DataRun run in dataRuns)
                    if (run != null)
                        run.clearData();

                dataRuns.Clear();
            }
        }

        /**
         * deletes a single run from the list
         * \param index index of run to delete
         * \return ERR_NOERR on success
         */
        public int deleteRun(int index)
        {
            if (getRunsCount() > index)
            {
                if (getRun(index) != null)
                    getRun(index).clearData();

                if (dataRuns != null)
                {
                    try
                    {
                        dataRuns.RemoveAt(index);
                        return ErrorCodes.ERR_NOERR;
                    }
                    catch (Exception)
                    {
                        return ErrorCodes.ERR_UNKNOWN;
                    }
                }
                else
                    return ErrorCodes.ERR_NULL;
            }
            else
                return ErrorCodes.ERR_INDEXBOUNDS;
        }

        /** initial vector count (one dimension) **/
        public int vectorCount
        {
            get { return _vectorCount; }
            set
            {
                if (value < 1 || value > 100)
                    _vectorCount = 1;
                else
                    _vectorCount = value;
            }
        }

        /**
         * processes a range of frames using a processing plugin
         * \param range the range to process
         * \param pluginManager the plugin manager containing import and processing plugins
         * \param configuration the global program configuration
         * \param procPluginIndex needed to obtain plugin identifier
         * \param totalProgress callback function indicating overall progress
         * \param relativeProgress callback function indicating progress of one processing step
         * \param timeCallback callback function to update remaining time
         * \return ERR_NOERR on success
         */
        public int analyzeMedia(FrameRange range, PluginManager pluginManager, Configuration configuration,
            int procPluginIndex, ProgressCallback totalProgress, ProgressCallback relativeProgress,
            TimeCallback timeCallback)
        {
            int errResult = ErrorCodes.ERR_NOERR;

            bool vectorCountZero = false;
            cancellationPending = false;
            cancellationForced = false;

            if (pluginManager == null || configuration == null)
                return ErrorCodes.ERR_NULL;

            DetectionParameters detectionParams = new DetectionParameters(configuration.DetectionSensitivity);

            IProcessingPlugin processingPlugin = pluginManager.getPlugin(procPluginIndex,
                PluginManager.PluginType.ptProcessing) as IProcessingPlugin;
            IImportPlugin importPlugin = pluginManager.activeImportPlugin;

            if (timeCallback != null)
                timeCallback(DateTime.MinValue);

            if (range.end > range.start &&
                processingPlugin != null && importPlugin != null)
            {
                if (processingPlugin.requiresConfig)
                    processingPlugin.configure();

                DateTime start = DateTime.Now;

                if (totalProgress != null)
                    totalProgress(0, "Initializing...");

                System.Drawing.Size frameDimension = new System.Drawing.Size();
                errResult = importPlugin.getDimension(range.start, ref frameDimension);
                if (errResult != ErrorCodes.ERR_NOERR)
                    return errResult;

                if (processingPlugin.beginProcessing(frameDimension) != 0)
                {
                    if (totalProgress != null)
                        totalProgress(0, "Failed to initialize plugin");
                    return ErrorCodes.ERR_NOTSUPPORTED;
                }

                Frame frame1 = null;
                Frame frame2 = null;

                int startFrame = range.start;
                int endFrame = range.end;

                addRun(vectorCount * vectorCount, startFrame, endFrame,
                    pluginManager.getIdentifier(procPluginIndex, PluginManager.PluginType.ptProcessing));

                int packageCounter = 0;

                MotionDetector motionDetector = null;
                IMotionDetector detectionAlg = null;
                IMotionProcessing processingAlg = null;
                float gridWidthX = (float)frameDimension.Width / (float)detectionParams.gridSize;
                float gridWidthY = (float)frameDimension.Height / (float)detectionParams.gridSize;

                for (int currentFrame1 = startFrame; currentFrame1 < endFrame; currentFrame1++)
                {
                    if (cancellationPending)
                        break;

                    int currentFrame2 = currentFrame1 + 1;

                    DataPackage package;

                    if (currentFrame1 == startFrame)
                    {
                        List<Vector> vectorsList = new List<Vector>();

                        // run motion detection on frames if necessary
                        if (configuration.detectionMode == DetectionMode.dmAutomaticCorner)
                        {
                            try
                            {
                                SusanCornersDetector cornerDetector = new SusanCornersDetector(detectionParams.diffThreshold,
                                    detectionParams.geomThreshold);
                                Bitmap motionFrame = new Bitmap(frameDimension.Width, frameDimension.Height);
                                Graphics motionFrameGraphics = Graphics.FromImage(motionFrame);

                                errResult = pluginManager.activeImportPlugin.renderFrame(currentFrame1, motionFrameGraphics);
                                if (errResult != ErrorCodes.ERR_NOERR)
                                    return errResult;

                                Point[] poi = cornerDetector.ProcessImage(motionFrame);
                                if (poi.Length > 0)
                                {
                                    int poiIterator = 1;
                                    int maxVectors = configuration.maxVectors * configuration.maxVectors;
                                    int tmpPoiLength = poi.Length;

                                    // if poi length exceeds maxVectors * maxVectors, reduce points
                                    while (tmpPoiLength > maxVectors)
                                    {
                                        poiIterator++;
                                        tmpPoiLength = poi.Length / poiIterator;
                                    }

                                    for (int i = 0; i < poi.Length; i += poiIterator)
                                    {
                                        Vector v = new Vector();
                                        v.start = poi[i];
                                        vectorsList.Add(v);
                                    }
                                }
                            }
                            catch (Exception)
                            {
                                ErrorCodes.error(ErrorCodes.ERR_UNKNOWN);
                                return ErrorCodes.ERR_UNKNOWN;
                            }
                        }
                        else
                        {
                            if (configuration.detectionMode == DetectionMode.dmAutomaticMotion)
                            {
                                try
                                {
                                    detectionAlg = new SimpleBackgroundModelingDetector(true, true);
                                    (detectionAlg as SimpleBackgroundModelingDetector).DifferenceThreshold = 
                                        Convert.ToInt32(detectionParams.diffThreshold / 2);
                                    processingAlg = new GridMotionAreaProcessing(detectionParams.gridSize, detectionParams.gridSize, false);
                                    motionDetector = new MotionDetector(detectionAlg, processingAlg);

                                    Bitmap motionFrame = new Bitmap(frameDimension.Width, frameDimension.Height);
                                    Graphics motionFrameGraphics = Graphics.FromImage(motionFrame);

                                    errResult = pluginManager.activeImportPlugin.renderFrame(currentFrame1, motionFrameGraphics);
                                    if (errResult != ErrorCodes.ERR_NOERR)
                                        return errResult;
                                    motionDetector.ProcessFrame(motionFrame);

                                    errResult = pluginManager.activeImportPlugin.renderFrame(currentFrame2, motionFrameGraphics);
                                    if (errResult != ErrorCodes.ERR_NOERR)
                                        return errResult;
                                    motionDetector.ProcessFrame(motionFrame);
                                }
                                catch (Exception)
                                {
                                    ErrorCodes.error(ErrorCodes.ERR_UNKNOWN);
                                    return ErrorCodes.ERR_UNKNOWN;
                                }
                            }

                            // create the first data package
                            int vecNum = vectorCount;

                            int deltaX = frameDimension.Width / (vecNum + 1);
                            int deltaY = frameDimension.Height / (vecNum + 1);

                            for (int i = 0; i < vecNum; i++)
                                for (int j = 0; j < vecNum; j++)
                                {
                                    Vector v = new Vector();
                                    v.start.X = (i + 1) * deltaX;
                                    v.start.Y = (j + 1) * deltaY;

                                    if (configuration.detectionMode == DetectionMode.dmManual)
                                        vectorsList.Add(v);

                                    if (configuration.detectionMode == DetectionMode.dmAutomaticMotion && motionDetector != null)
                                    {
                                        int gridX = Convert.ToInt32(Math.Floor(v.start.X / gridWidthX));
                                        int gridY = Convert.ToInt32(Math.Floor(v.start.Y / gridWidthY));
                                        if ((processingAlg as GridMotionAreaProcessing).MotionGrid[gridX, gridY] > 
                                            detectionParams.detectionTrigger)
                                            vectorsList.Add(v);
                                    }
                                }
                        }

                        package = new DataPackage(currentFrame1, currentFrame2, vectorsList.Count);
                        for (int i = 0; i < vectorsList.Count; i++)
                            package.vectors[i] = vectorsList[i];

                        frame1 = importPlugin.normalizeFrame(currentFrame1);
                        frame2 = importPlugin.normalizeFrame(currentFrame2);
                    } //end of currentFrame1 == startFrame
                    else
                    {
                        package = null;
                        List<Vector> vectorsList = new List<Vector>();

                        if (configuration.trackingMode == TrackingMode.tmAutomaticCorner)
                        {
                            try
                            {
                                SusanCornersDetector cornerDetector = new SusanCornersDetector(detectionParams.diffThreshold,
                                    detectionParams.geomThreshold);
                                Bitmap motionFrame = new Bitmap(frameDimension.Width, frameDimension.Height);
                                Graphics motionFrameGraphics = Graphics.FromImage(motionFrame);

                                errResult = pluginManager.activeImportPlugin.renderFrame(currentFrame1, motionFrameGraphics);
                                if (errResult != ErrorCodes.ERR_NOERR)
                                    return errResult;

                                Point[] poi = cornerDetector.ProcessImage(motionFrame);
                                if (poi.Length > 0)
                                {
                                    int poiIterator = 1;
                                    int maxVectors = configuration.maxVectors * configuration.maxVectors;
                                    int tmpPoiLength = poi.Length;

                                    // if poi length exceeds maxVectors * maxVectors, reduce points
                                    while (tmpPoiLength > maxVectors)
                                    {
                                        poiIterator++;
                                        tmpPoiLength = poi.Length / poiIterator;
                                    }

                                    for (int i = 0; i < poi.Length; i += poiIterator)
                                    {
                                        Vector v = new Vector();
                                        v.start = poi[i];
                                        vectorsList.Add(v);
                                    }
                                }

                                package = new DataPackage(currentFrame1, currentFrame2, vectorsList.Count);
                                for (int i = 0; i < vectorsList.Count; i++)
                                    package.vectors[i] = vectorsList[i];
                            }
                            catch (Exception)
                            {
                                ErrorCodes.error(ErrorCodes.ERR_UNKNOWN);
                                return ErrorCodes.ERR_UNKNOWN;
                            }
                        }

                        if (configuration.trackingMode == TrackingMode.tmAutomaticMotion)
                        {
                            bool firstDetection = false;

                            if (detectionAlg == null)
                            {
                                detectionAlg = new SimpleBackgroundModelingDetector(true, true);
                                (detectionAlg as SimpleBackgroundModelingDetector).DifferenceThreshold =
                                    Convert.ToInt32(detectionParams.diffThreshold / 2);
                                firstDetection = true;
                            }

                            if (processingAlg == null)
                                processingAlg = new GridMotionAreaProcessing(detectionParams.gridSize, detectionParams.gridSize, false);
                            if (motionDetector == null)
                                motionDetector = new MotionDetector(detectionAlg, processingAlg);

                            Bitmap motionFrame = new Bitmap(frameDimension.Width, frameDimension.Height);
                            Graphics motionFrameGraphics = Graphics.FromImage(motionFrame);

                            if (firstDetection)
                            {
                                errResult = pluginManager.activeImportPlugin.renderFrame(currentFrame1, motionFrameGraphics);
                                if (errResult != ErrorCodes.ERR_NOERR)
                                    return errResult;
                                motionDetector.ProcessFrame(motionFrame);
                            }

                            errResult = pluginManager.activeImportPlugin.renderFrame(currentFrame2, motionFrameGraphics);
                            if (errResult != ErrorCodes.ERR_NOERR)
                                return errResult;
                            motionDetector.ProcessFrame(motionFrame);

                            int vecNum = vectorCount;

                            int deltaX = frameDimension.Width / (vecNum + 1);
                            int deltaY = frameDimension.Height / (vecNum + 1);

                            for (int i = 0; i < vecNum; i++)
                                for (int j = 0; j < vecNum; j++)
                                {
                                    Vector v = new Vector();
                                    v.start.X = (i + 1) * deltaX;
                                    v.start.Y = (j + 1) * deltaY;

                                    int gridX = Convert.ToInt32(Math.Floor(v.start.X / gridWidthX));
                                    int gridY = Convert.ToInt32(Math.Floor(v.start.Y / gridWidthY));
                                    if ((processingAlg as GridMotionAreaProcessing).MotionGrid[gridX, gridY] >
                                        detectionParams.detectionTrigger)
                                        vectorsList.Add(v);
                                }

                            package = new DataPackage(currentFrame1, currentFrame2, vectorsList.Count);
                            for (int i = 0; i < vectorsList.Count; i++)
                                package.vectors[i] = vectorsList[i];
                        } // end of tmAutomaticMotion

                        if (configuration.trackingMode == TrackingMode.tmFollowing)
                        {
                            DataPackage lastPackage = currentRun.getPackage(packageCounter);

                            // count vectors found in last processing step
                            int lastValidVectors = 0;
                            for (int i = 0; i < lastPackage.vectors.Length; i++)
                                if (lastPackage.vectors[i].valid)
                                    lastValidVectors++;

                            if (lastValidVectors == 0)
                            {
                                vectorCountZero = true;
                                break;
                            }

                            package = new DataPackage(currentFrame1, currentFrame2, lastValidVectors);

                            int tmpVectorIndex = 0;
                            for (int i = 0; i < lastPackage.vectors.Length; i++)
                                if (lastPackage.vectors[i].valid)
                                {
                                    package.vectors[tmpVectorIndex].start = lastPackage.vectors[i].target;
                                    tmpVectorIndex++;
                                }
                        }

                        if (configuration.trackingMode == TrackingMode.tmStart)
                        {
                            DataPackage lastPackage = currentRun.getPackage(0);
                            package = new DataPackage(currentFrame1, currentFrame2, lastPackage.vectors.Length);
                            for (int i = 0; i < lastPackage.vectors.Length; i++)
                                package.vectors[i].start = lastPackage.vectors[i].start;
                        }

                        frame1 = frame2;
                        frame2 = importPlugin.normalizeFrame(currentFrame2);
                        packageCounter++;
                    }

                    // this looks a bit creepy because currentFrame1/2 is zero-indexed while 'user frames' start with one
                    if (totalProgress != null)
                        totalProgress((currentFrame1 - startFrame + 1) * 100 / (endFrame - startFrame),
                            "Processing frames " + (currentFrame1 + 1).ToString() + " and " + (currentFrame2 + 1).ToString() +
                            " (" + (currentFrame1 - startFrame + 1).ToString() + " of " + (endFrame - startFrame).ToString() + " with " +
                            package.vectors.Length.ToString() + " vectors)");

                    if (relativeProgress != null)
                        relativeProgress(0, "");

                    DateTime startTime = DateTime.Now;

                    // process the frames
                    int procResult = processFrames(frame1, frame2, ref package, processingPlugin, relativeProgress);
                    // procResult is somehow plugin specific but should be 0 on success
                    if (procResult != 0)
                    {
                        System.Windows.Forms.MessageBox.Show("The processing plugin returned an error (Code: " +
                            procResult.ToString() + ")", "Error", System.Windows.Forms.MessageBoxButtons.OK,
                            System.Windows.Forms.MessageBoxIcon.Error);

                        return ErrorCodes.ERR_UNKNOWN;
                    }

                    if (cancellationForced)
                        break;

                    DateTime endTime = DateTime.Now;
                    try
                    {
                        DateTime duration = endTime.Subtract(startTime.TimeOfDay);
                        if (timeCallback != null)
                            timeCallback(new DateTime(duration.TimeOfDay.Ticks * (endFrame - currentFrame2)));
                    }
                    catch (ArgumentOutOfRangeException)
                    {
                        ErrorCodes.error(ErrorCodes.ERR_FORMATERROR);
                    }

                    currentRun.addPackage(package);
                    System.Windows.Forms.Application.DoEvents();
                }

                if (!cancellationForced)
                {
                    processingPlugin.endProcessing();

                    DateTime end = DateTime.Now;
                    currentRun.setDuration(end.Subtract(start.TimeOfDay));

                    if (vectorCountZero)
                        System.Windows.Forms.MessageBox.Show("There are no more vectors to analyse.", "Canceling",
                            System.Windows.Forms.MessageBoxButtons.OK, System.Windows.Forms.MessageBoxIcon.Information);

                    if (totalProgress != null)
                        totalProgress(100, "Done");
                    if (relativeProgress != null)
                        relativeProgress(100, "");
                }
                return ErrorCodes.ERR_NOERR;
            }
            else
                return ErrorCodes.ERR_NULL;
        }

        /**
         * \class WorkerArguments
         * object holding arguments for asynchronious processing **/
        private class WorkerArguments
        {
            public Frame frame1;
            public Frame frame2;
            public DataPackage dataPackage;
            public IProcessingPlugin processingPlugin;
            public ProgressCallback progressCallback;
            public int result;
        }

        /** running event of the BackgroundWorker **/
        private void worker_DoWork(object sender, DoWorkEventArgs e)
        {
            BackgroundWorker worker = sender as BackgroundWorker;
            WorkerArguments args = e.Argument as WorkerArguments;

            args.result = args.processingPlugin.processData(args.frame1, args.frame2, ref args.dataPackage,
                args.dataPackage.vectors.Length, ThreadProgress);
        }

        /** progress event of the BackgroundWorker **/
        private void worker_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            if (internalProgressCallback != null)
                internalProgressCallback(e.ProgressPercentage, "");
        }

        /** forwards plugin progress to the BackgroundWorker **/
        private void ThreadProgress(int progress, string msg)
        {
            backgroundWorker.ReportProgress(progress);
        }

        /**
         * processes two frames
         * \param frame1 first Frame object
         * \param frame2 second Frame object
         * \param dataPackage reference to data package for storing results
         * \param processingPlugin the processing plugin
         * \param progressCallback callback function indicating processing progress
         * \return ERR_NOERR on success
         * \bug #0008 GPU mode of processing plugin won't work with threads (probably context errors ?)
         */
        private int processFrames(Frame frame1, Frame frame2,
            ref DataPackage dataPackage, IProcessingPlugin processingPlugin, ProgressCallback progressCallback)
        {
            if (!processingPlugin.supportsThreading)
            {
                return processingPlugin.processData(frame1, frame2, ref dataPackage,
                    dataPackage.vectors.Length, progressCallback);
            }

            WorkerArguments args = new WorkerArguments();
            args.frame1 = frame1;
            args.frame2 = frame2;
            args.dataPackage = dataPackage;
            args.processingPlugin = processingPlugin;
            args.progressCallback = ThreadProgress;
            args.result = ErrorCodes.ERR_NOERR;
            internalProgressCallback = progressCallback;

            backgroundWorker.RunWorkerAsync(args);

            while (backgroundWorker.IsBusy)
                System.Windows.Forms.Application.DoEvents();

            if (args.result == ErrorCodes.ERR_NOERR)
                dataPackage = args.dataPackage;

            return args.result;
        }
    }
}
