﻿/*
multiDAC - Movement and deformation analysis tool
Class for managing plugin assemblies
Copyright (C) 2009 Felix Schmitt

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

using System;
using System.Collections;
using System.Text;
using System.IO;
using System.Reflection;
using PluginInterfaceLibrary;
using InternalData;

namespace Manager
{
    /**
     * \class PluginManager
     * Loads and manages plugins
     */
    public class PluginManager
    {
        /**
         * \enum PluginType
         * Defines the type of a plugin
         */
        public enum PluginType { ptUnknown, ptProcessing, ptImport, ptExport };

        /** delegate callback method used while reloading plugins **/ 
        public delegate PluginReloadAction ReloadPluginsCallback(bool initialReload);

        /**
         * \class InternalPlugin
         * Internal class used for storing a plugin with additional information
         */
        private class InternalPlugin
        {
            public PluginType pluginType { get; internal set; }
            public string filename { get; internal set; }
            public IPlugin plugin { get; internal set; }
            public string identifier { get; internal set; }

            /**
             * constructor
             * \param pluginType type of plugin
             * \param filename file containing plugin
             * \param plugin interface to the plugin
             */
            public InternalPlugin(PluginType pluginType, string filename, IPlugin plugin)
            {
                this.pluginType = pluginType;
                this.filename = filename;
                this.plugin = plugin;
                this.identifier = plugin.GetType().FullName;
            }
        }

        /** location of plugin directory **/
        public string pathToPlugins { get; internal set; }

        // the global plugin order is: Import > Processing > Export

        private ArrayList importPlugins;
        private int _importPluginIndex;

        private ArrayList processingPlugins;
        private int _processingPluginIndex;

        private ArrayList exportPlugins;
        private int _exportPluginIndex;

        /** number of all plugins loaded **/
        public int totalPluginCount {
            get { return importPluginCount + processingPluginCount + exportPluginCount; }
        }

        /** number of import plugins loaded **/
        public int importPluginCount
        {
            get { return importPlugins.Count; }
        }

        /** number of processing plugins loaded **/
        public int processingPluginCount
        {
            get { return processingPlugins.Count; }
        }

        /** number of export plugins loaded **/
        public int exportPluginCount
        {
            get { return exportPlugins.Count; }
        }

        /** index of active import plugin **/
        public int importPluginIndex
        {
            get { return _importPluginIndex; }
            set
            {
                if (importPlugins.Count > value && value >= 0)
                    _importPluginIndex = value;
            }
        }

        /** index of active processing plugin **/
        public int processingPluginIndex
        {
            get { return _processingPluginIndex; }
            set
            {
                if (processingPlugins.Count > value && value >= 0)
                    _processingPluginIndex = value;
            }
        }

        /** index of active export plugin **/
        public int exportPluginIndex
        {
            get { return _exportPluginIndex; }
            set
            {
                if (exportPlugins.Count > value && value >= 0)
                    _exportPluginIndex = value;
            }
        }

        /** active import plugin or null **/
        public IImportPlugin activeImportPlugin
        {
            get
            {
                if (importPluginIndex >= 0 && importPluginIndex < importPlugins.Count)
                    return ((importPlugins[importPluginIndex] as InternalPlugin).plugin as IImportPlugin);
                else
                    return null;
            }
        }

        /** active processing plugin or null **/
        public IProcessingPlugin activeProcessingPlugin
        {
            get
            {
                if (processingPluginIndex >= 0 && processingPluginIndex < processingPlugins.Count)
                    return ((processingPlugins[processingPluginIndex] as InternalPlugin).plugin as IProcessingPlugin);
                else
                    return null;
            }
        }

        /** active export plugin or null **/
        public IExportPlugin activeExportPlugin
        {
            get
            {
                if (exportPluginIndex >= 0 && exportPluginIndex < exportPlugins.Count)
                    return ((exportPlugins[exportPluginIndex] as InternalPlugin).plugin as IExportPlugin);
                else
                    return null;
            }
        }

        /** 
         * constructor 
         **/
        public PluginManager()
        {
            try
            {
#if DEBUG
                pathToPlugins = Path.GetDirectoryName(Assembly.GetCallingAssembly().Location) + "\\..\\..\\..\\plugins";
#else
                pathToPlugins = Path.GetDirectoryName(Assembly.GetCallingAssembly().Location) + "\\plugins";
#endif
                AppDomain.CurrentDomain.AppendPrivatePath("plugins");
            }
            catch (Exception)
            {
                ErrorCodes.error(ErrorCodes.ERR_IO);
                pathToPlugins = "";
            }

            importPlugins = new ArrayList();
            processingPlugins = new ArrayList();
            exportPlugins = new ArrayList();
        }

        /**
         * returns an identifier for a plugin
         * \param index index of plugin (in the local list)
         * \param pluginType type of plugin (to chose the list)
         * \return the identifier or null
         */
        public string getIdentifier(int index, PluginType pluginType)
        {
            if (pluginType == PluginType.ptUnknown || index < 0)
                return null;

            switch (pluginType)
            {
                case PluginType.ptImport:
                    if (index >= importPluginCount)
                        return null;

                    return Path.GetFileName((importPlugins[index] as InternalPlugin).filename)
                        + (importPlugins[index] as InternalPlugin).identifier;
                case PluginType.ptProcessing:
                    if (index >= processingPluginCount)
                        return null;

                    return Path.GetFileName((processingPlugins[index] as InternalPlugin).filename)
                        + (processingPlugins[index] as InternalPlugin).identifier;
                case PluginType.ptExport:
                    if (index >= exportPluginCount)
                        return null;

                    return Path.GetFileName((exportPlugins[index] as InternalPlugin).filename)
                        + (exportPlugins[index] as InternalPlugin).identifier;
                default:
                    return null;
            }
        }

        /**
         * returns a file path for a plugin
         * \param index index of plugin (in the local list)
         * \param pluginType type of plugin (to chose the list)
         * \return the path or null
         */
        public string getPath(int index, PluginType pluginType)
        {
            if (pluginType == PluginType.ptUnknown || index < 0)
                return null;

            switch (pluginType)
            {
                case PluginType.ptImport:
                    if (index >= importPluginCount)
                        return null;

                    return (importPlugins[index] as InternalPlugin).filename;
                case PluginType.ptProcessing:
                    if (index >= processingPluginCount)
                        return null;

                    return (processingPlugins[index] as InternalPlugin).filename;
                case PluginType.ptExport:
                    if (index >= exportPluginCount)
                        return null;

                    return (exportPlugins[index] as InternalPlugin).filename;
                default:
                    return null;
            }
        }

        /**
         * sets the active plugin
         * \param identifier identifier for plugin
         * \param pluginType type of plugin
         * \return ERR_NOERR on success
         */
        public int setActivePlugin(string identifier, PluginType pluginType)
        {
            if (identifier == null || pluginType == PluginType.ptUnknown)
                return ErrorCodes.ERR_NOTSUPPORTED;

            switch (pluginType)
            {
                case PluginType.ptImport:
                    for (int i = 0; i < importPluginCount; i++)
                        if (identifier.CompareTo(getIdentifier(i, PluginType.ptImport)) == 0)
                        {
                            importPluginIndex = i;
                            return ErrorCodes.ERR_NOERR;
                        }
                    break;
                case PluginType.ptProcessing:
                    for (int i = 0; i < processingPluginCount; i++)
                        if (identifier.CompareTo(getIdentifier(i, PluginType.ptImport)) == 0)
                        {
                            processingPluginIndex = i;
                            return ErrorCodes.ERR_NOERR;
                        }
                    break;
                case PluginType.ptExport:
                    for (int i = 0; i < exportPluginCount; i++)
                        if (identifier.CompareTo(getIdentifier(i, PluginType.ptExport)) == 0)
                        {
                            exportPluginIndex = i;
                            return ErrorCodes.ERR_NOERR;
                        }
                    break;
                default:
                    return ErrorCodes.ERR_FORMATERROR;
            }

            return ErrorCodes.ERR_NOTFOUND;
        }

        /**
         * configures plugin manager and calls the plugin settings dialog
         * \param reloadCallback callback function used if plugins have to be reloaded
         * \param progressCallback callback function indicating reload progress
         * \return indicates result of reload operation
         */
        public PluginReloadAction configure(ReloadPluginsCallback reloadCallback, ProgressCallback progressCallback)
        {
            if (progressCallback != null)
                progressCallback(0, "Done");
            
            PluginReloadAction result = PluginReloadAction.praLoaded;

            PluginManagerForm configForm = new PluginManagerForm(this);
            configForm.setFolderPath(pathToPlugins);
            if (configForm.ShowDialog() == System.Windows.Forms.DialogResult.OK)
            {
                string oldPathToPlugins = pathToPlugins;
                pathToPlugins = configForm.getFolderPath();

                if (pathToPlugins.CompareTo(oldPathToPlugins) != 0)
                {
                    result = reloadCallback(false);
                    if (result == PluginReloadAction.praCancel)
                    {
                        pathToPlugins = oldPathToPlugins;

                        if (progressCallback != null)
                            progressCallback(0, "Restored previous plugin path");
                    }
                }
            }

            configForm.Dispose();

            return result;
        }

        /**
         * get a plugin from its global index
         * \param index index of plugin
         * \return interface to plugin or null
         */
        public IPlugin getPlugin(int index)
        {
            if (index >= 0 && index < totalPluginCount)
            {
                if (index < importPluginCount)
                    return (importPlugins[index] as InternalPlugin).plugin;
                else
                    if (index < importPluginCount + processingPluginCount)
                        return (processingPlugins[index - importPluginCount] as InternalPlugin).plugin;
                    else
                        return (exportPlugins[index - importPluginCount - processingPluginCount] as InternalPlugin).plugin;
            }
            else
                return null;
        }

        /**
         * get plugin from local index
         * \param index index of plugin
         * \param pluginType type of plugin to chose list
         * \return interface to plugin or null
         */
        public IPlugin getPlugin(int index, PluginType pluginType)
        {
            if (index < 0)
                return null;

            switch (pluginType)
            {
                case PluginType.ptImport:
                    if (index < importPluginCount)
                        return (importPlugins[index] as InternalPlugin).plugin;
                    break;
                case PluginType.ptProcessing:
                    if (index < processingPluginCount)
                        return (processingPlugins[index] as InternalPlugin).plugin;
                    break;
                case PluginType.ptExport:
                    if (index < exportPluginCount)
                        return (exportPlugins[index] as InternalPlugin).plugin;
                    break;
                default:
                    return null;
            }

            return null;
        }

        /**
         * drops existing plugin and loads those in plugin directory
         * \param progressCallback callback function indicating load progress
         * \return ERR_NOERR on success
         */
        public int loadPlugins(ProgressCallback progressCallback)
        {
            try
            {
                importPlugins.Clear();
                processingPlugins.Clear();
                exportPlugins.Clear();
            }
            catch (NotSupportedException e)
            {
                Console.WriteLine(e.Message);
                return ErrorCodes.ERR_NOTSUPPORTED;
            }

            string[] pluginFiles;
            try
            {
                pluginFiles = Directory.GetFiles(pathToPlugins, "*.dll", SearchOption.AllDirectories);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
                return ErrorCodes.ERR_IO;
            }

            if (pluginFiles.Length == 0)
                return ErrorCodes.ERR_INDEXBOUNDS;

            for (int i = 0; i < pluginFiles.Length; i++)
            {
                try
                {
                    Assembly pluginDll = Assembly.LoadFile(pluginFiles[i]);
                    if (pluginDll != null)
                    {
                        progressCallback((i + 1) * 100 / pluginFiles.Length,
                            "Loading plugins");

                        foreach (Type type in pluginDll.GetExportedTypes())
                            if (type.IsClass && !type.IsAbstract && type.GetInterface(typeof(IPlugin).FullName) != null)
                            {
                                if (type.GetInterface(typeof(IImportPlugin).FullName) != null)
                                {
                                    progressCallback((i + 1) * 100 / pluginFiles.Length,
                                        "Loading plugin " + Path.GetFileName(pluginFiles[i]));

                                    object importPlugin = Activator.CreateInstance(type);
                                    InternalPlugin internalPlugin = new InternalPlugin(PluginType.ptImport,
                                        pluginFiles[i], importPlugin as IPlugin);
                                    importPlugins.Add(internalPlugin);
                                }

                                if (type.GetInterface(typeof(IProcessingPlugin).FullName) != null)
                                {
                                    progressCallback((i + 1) * 100 / pluginFiles.Length,
                                        "Loading plugin " + Path.GetFileName(pluginFiles[i]));

                                    object processingPlugin = Activator.CreateInstance(type);
                                    InternalPlugin internalPlugin = new InternalPlugin(PluginType.ptProcessing,
                                        pluginFiles[i], processingPlugin as IPlugin);
                                    processingPlugins.Add(internalPlugin);
                                }

                                if (type.GetInterface(typeof(IExportPlugin).FullName) != null)
                                {
                                    progressCallback((i + 1) * 100 / pluginFiles.Length,
                                        "Loading plugin " + Path.GetFileName(pluginFiles[i]));

                                    object exportPlugin = Activator.CreateInstance(type);
                                    InternalPlugin internalPlugin = new InternalPlugin(PluginType.ptExport,
                                        pluginFiles[i], exportPlugin as IPlugin);
                                    exportPlugins.Add(internalPlugin);
                                }
                            }
                    }
                }
                catch (Exception e)
                {
                    Console.WriteLine(e.Message);
                    continue;
                }
            }

            progressCallback(0, "Done");

            return ErrorCodes.ERR_NOERR;
        }
    }
}
