﻿/*
multiDAC - Movement and deformation analysis tool
Class for managing project (files)
Copyright (C) 2009 Felix Schmitt

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Xml;
using System.Xml.Serialization;
using System.Windows.Forms;
using InternalData;
using PluginInterfaceLibrary;
using ICSharpCode.SharpZipLib.Zip;

namespace Manager
{
    /** 
     * delegate callback method used during loading/saving projects 
     * \param project the project to operate with
     * \return the projects new 'modified' state
     **/
    public delegate bool ProjectOperationCallback(Project project);

    /**
     * \class Project
     * Internal representation of all project parameters
     * optimized for xml serialization
     */
    public class Project
    {
        public string name;
        public string user;
        public string description;
        public DateTime creationDate;
        public List<string> sourceFiles;
        public bool includeSourceFiles;
        public int initialVectors;
        public string importPlugin;
        public ProjectDataRun[] processingRuns;

        /** constructor **/
        public Project()
        {
            name = "untitled";
            user = "";
            description = "";
            creationDate = DateTime.Now;
            sourceFiles = new List<string>();
            includeSourceFiles = false;
            initialVectors = 25;
            importPlugin = "";
        }

        /**
         * sets the filenames of the source files
         * \param sourceFiles string array of filenames
         * \bug #0007 filenames of sourcefiles shouldn't be saved with path if
         * includeSourceFiles is true
         */
        public void setSourceFiles(string[] sourceFiles)
        {
            if (this.sourceFiles != null)
            {
                this.sourceFiles.Clear();

                if (sourceFiles != null)
                    for (int i = 0; i < sourceFiles.Length; i++)
                        this.sourceFiles.Add(sourceFiles[i]);
            }
        }
    }

    /**
     * \class ProjectManager
     * Manages project specific operations (loading/saving/settings)
     */
    public class ProjectManager
    {
        public static string PROJ_ARCHIVE_EXT = ".mpz";
        public static string PROJ_XML_EXT = ".mpr";
        public static string PROJ_BINARY_EXT = ".mbp";

        public enum ProjectFileType { pftSingleFiles, pftArchive };

        public bool modified { get; internal set; }
        public bool projectLoaded { get; internal set; }
        public Project project { get; internal set; }
        public string location;
        public bool includeSrcFilesByDefault;
        public ProjectFileType fileType { get; internal set; }
        
        private DataProcessor dataProcessor;
        private PluginManager pluginManager;

        public List<string> cachedDirectories { get; internal set; }

        // 1 MB
        private static long MAX_INCLUDE_SIZE = 1024 * 1024;

        /**
         * constructor
         * \param dataProcessor the data processor
         * \param pluginManager the plugin manager
         */
        public ProjectManager(DataProcessor dataProcessor, PluginManager pluginManager)
        {
            this.dataProcessor = dataProcessor;
            this.pluginManager = pluginManager;
            modified = false;
            projectLoaded = false;
            project = null;
            location = "";
            includeSrcFilesByDefault = false;
            fileType = ProjectFileType.pftArchive;
            cachedDirectories = new List<string>();
        }

        /** sets modified to true **/
        public void modify()
        {
            modified = true;
        }

        /**
         * creates a new project
         * \param name project's name
         * \param createdCallback callback function after project was created
         */
        public void create(string name, ProjectOperationCallback createdCallback)
        {
            project = new Project();
            project.name = name;
            project.includeSourceFiles = includeSrcFilesByDefault;
            location = "";
            fileType = ProjectFileType.pftArchive;
            
            if (createdCallback != null)
                createdCallback(project);

            modified = false;
            projectLoaded = true;
        }

        /** closes existing project **/
        public void close()
        {
            modified = false;
            projectLoaded = false;
            project = null;
            location = "";
            deleteCache();
        }

        /**
         * delete files on disk
         * \param files string array with filenames to delete
         * \return ERR_NOERR on success
         */
        private int deleteFiles(string[] files)
        {
            if (files == null)
                return ErrorCodes.ERR_NULL;

            try
            {
                foreach (string file in files)
                    File.Delete(file);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
                return ErrorCodes.ERR_IO;
            }

            return ErrorCodes.ERR_NOERR;
        }

        /**
         * delete cached temporary directories
         * \return ERR_NOERR on success
         */
        private int deleteCache()
        {
            if (cachedDirectories != null)
            {
                try
                {
                    if (cachedDirectories.Count > 0)
                    {
                        foreach (string dir in cachedDirectories)
                            Directory.Delete(dir, true);
                    }
                }
                catch (Exception e)
                {
                    Console.WriteLine(e.Message);
                    return ErrorCodes.ERR_IO;
                }

                cachedDirectories.Clear();

                return ErrorCodes.ERR_NOERR;
            }
            else
                return ErrorCodes.ERR_NULL;
        }

        /**
         * loads a project file (compressed archive)
         * \param archiveFilename file to load from (compressed archive)
         * \param loadedCallback callback function to apply loaded project to program
         * \param progressCallback callback function indicating loading progress
         * \return ERR_NOERR on success
         */
        public int load(string archiveFilename, ProjectOperationCallback loadedCallback,
            ProgressCallback progressCallback)
        {
            int errResult = ErrorCodes.ERR_NOERR;

            if (progressCallback != null)
                progressCallback(0, "Loading " + archiveFilename);

            bool loadFromArchive = Path.GetExtension(archiveFilename).CompareTo(PROJ_ARCHIVE_EXT) == 0;
            if (loadFromArchive)
                fileType = ProjectFileType.pftArchive;
            else
                fileType = ProjectFileType.pftSingleFiles;

            List<string> filesList = new List<string>();
            string xmlFilename;
            string dataFilename;

            if (loadFromArchive)
            {
                errResult = loadArchive(archiveFilename, filesList, cachedDirectories, progressCallback);

                if (errResult != ErrorCodes.ERR_NOERR || filesList.Count == 0)
                {
                    MessageBox.Show("Failed to load " + archiveFilename + ".", "Error", MessageBoxButtons.OK,
                        MessageBoxIcon.Error);
                    if (progressCallback != null)
                        progressCallback(100, "Failed to load " + archiveFilename);
                    return errResult;
                }

            }

            /**
             * \todo to be changed
             */
            if (loadFromArchive)
            {
                xmlFilename = filesList[0];
                dataFilename = filesList[1];
            }
            else
            {
                xmlFilename = archiveFilename;
                dataFilename = xmlFilename + PROJ_BINARY_EXT;
            }

            XmlSerializer xmlSer = new XmlSerializer(typeof(Project));
            if (xmlSer == null)
            {
                if (loadFromArchive)
                    deleteFiles(filesList.ToArray());
                return ErrorCodes.ERR_NULL;
            }

            TextReader stream;
            try
            {
                stream = new StreamReader(xmlFilename);
            }
            catch (Exception e)
            {
                MessageBox.Show(e.Message);
                MessageBox.Show("Failed to open " + xmlFilename + ".", "Error", MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
                if (loadFromArchive)
                    deleteFiles(filesList.ToArray());
                return ErrorCodes.ERR_IO;
            }

            try
            {
                project = (Project)xmlSer.Deserialize(stream);
            }
            catch (InvalidOperationException e)
            {
#if DEBUG
                MessageBox.Show(e.Message);
#endif
                MessageBox.Show("Failed to read project data.", "Error", MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
                stream.Close();
                if (loadFromArchive)
                    deleteFiles(filesList.ToArray());
                return ErrorCodes.ERR_NOTSUPPORTED;
            }

            stream.Close();

            // clear data
            dataProcessor.clearData();
            try
            {
                for (int i = 0; i < project.processingRuns.Length; i++)
                {
                    ProjectDataRun run = project.processingRuns[i] as ProjectDataRun;

                    DataRun actualRun = dataProcessor.addRun(run.initialVectors, run.firstFrame, 
                        run.lastFrame, run.plugin);
                    actualRun.setDuration(run.duration);
                    foreach (IndexedNote note in run.notes)
                        actualRun.addNote(note.index, note.note);
                }

                DataReader reader = new DataReader(dataFilename);
                if (reader == null || reader.read(dataProcessor, project.processingRuns, progressCallback) != 0)
                {
                    dataProcessor.clearData();
                    MessageBox.Show("Could not read binary data.", "Error", MessageBoxButtons.OK,
                        MessageBoxIcon.Error);
                }

                // load source files
                List<string> files = project.sourceFiles;

                if (progressCallback != null)
                    progressCallback(50, "Loading " + archiveFilename);

                if (pluginManager.activeImportPlugin != null)
                {
                    pluginManager.activeImportPlugin.clearFiles();
                }

                if (loadFromArchive)
                    deleteFiles(filesList.ToArray());
                
                if (progressCallback != null)
                    progressCallback(100, "Loading " + archiveFilename);

                modified = false;
                if (loadedCallback != null)
                    modified = loadedCallback(project);
                
                location = archiveFilename;

                projectLoaded = true;

            }
            catch (Exception e)
            {
#if DEBUG
                Console.WriteLine(e.Message);
#endif
                MessageBox.Show("Could not load project.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                close();
                if (loadFromArchive)
                    deleteFiles(filesList.ToArray());
                return ErrorCodes.ERR_ACCESSDENIED;
            }

            if (loadFromArchive)
                deleteFiles(filesList.ToArray());

            return ErrorCodes.ERR_NOERR;
        }

        /**
         * saves a project to a file (compressed archive)
         * \param archiveFilename file to save to (compressed archive)
         * \param savingCallback callback function used to apply project info to saved object
         * \param progressCallback callback function indicating saving progress
         * \return ERR_NOERR on success
         */
        public int save(string archiveFilename, ProjectOperationCallback savingCallback,
            ProgressCallback progressCallback)
        {
            if (progressCallback != null)
                progressCallback(0, "Saving " + archiveFilename);

            bool saveToArchive = Path.GetExtension(archiveFilename).CompareTo(PROJ_ARCHIVE_EXT) == 0;

            if (saveToArchive)
                fileType = ProjectFileType.pftArchive;
            else
                fileType = ProjectFileType.pftSingleFiles;

            List<string> filesList = new List<string>();
            string xmlFilename;
            string dataFilename;

            if (saveToArchive)
            {
                /**
                 * \bug #0006 SOLVED if filename part of xmlFilename already exists in destinaion directory,
                 * no files are added to the zip file, therefor the underscore
                 */
                xmlFilename = Path.GetTempPath() + "_" + Path.GetFileNameWithoutExtension(archiveFilename) +
                    PROJ_XML_EXT;
                dataFilename = xmlFilename + PROJ_BINARY_EXT;
                filesList.Add(xmlFilename);
                filesList.Add(dataFilename);
            }
            else
            {
                xmlFilename = archiveFilename;
                dataFilename = xmlFilename + PROJ_BINARY_EXT;
            }

            XmlSerializer xmlSer;
            try
            {
                xmlSer = new XmlSerializer(typeof(Project));
            }
            catch (InvalidOperationException e)
            {
                MessageBox.Show("This is an internal bug.", "Error", MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
#if DEBUG
                MessageBox.Show(e.InnerException.Message);
#endif
                return ErrorCodes.ERR_NOTSUPPORTED;
            }

            if (xmlSer == null)
                return ErrorCodes.ERR_NULL;

            if (project == null)
                return ErrorCodes.ERR_NULL;

            location = archiveFilename;
            if (savingCallback != null)
                savingCallback(project);

            // test on source files filesize
            if (saveToArchive && project.includeSourceFiles)
                foreach (string sourcefile in project.sourceFiles)
                {
                    try
                    {
                        FileInfo fileInfo = new FileInfo(sourcefile);
                        if (fileInfo.Length > MAX_INCLUDE_SIZE)
                        {
                            if (MessageBox.Show("Some of your source files are very large. Including them would take" +
                                " time and consume a lot of disk space.\nDo you want to include them anyway?",
                                "Warning", MessageBoxButtons.YesNo, MessageBoxIcon.Warning,
                                MessageBoxDefaultButton.Button2) == DialogResult.No)
                                project.includeSourceFiles = false;
                            break;
                        }
                    }
                    catch (Exception)
                    {
#if DEBUG
                        MessageBox.Show("Error: Could not retrieve file info for " + sourcefile, "Error", MessageBoxButtons.OK,
                            MessageBoxIcon.Error);
#endif
                        ErrorCodes.error(ErrorCodes.ERR_UNKNOWN);
                    }
                }

            TextWriter stream;
            try
            {
                stream = new StreamWriter(xmlFilename);
            }
            catch (Exception e)
            {
#if DEBUG
                MessageBox.Show(e.Message);
#endif
                MessageBox.Show("Failed to open " + xmlFilename + ".", "Error", MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
                return ErrorCodes.ERR_IO;
            }

            try
            {
                xmlSer.Serialize(stream, project);
            }
            catch (InvalidOperationException e)
            {
#if DEBUG
                MessageBox.Show(e.InnerException.Message);
#endif
                MessageBox.Show("Failed to write project data.", "Error", MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
            }
            stream.Close();

            if (progressCallback != null)
                progressCallback(50, "Saving " + archiveFilename);

            // write binary data
            DataWriter writer = new DataWriter(dataFilename);
            if (writer == null || writer.write(dataProcessor, progressCallback) != ErrorCodes.ERR_NOERR)
                MessageBox.Show("Could not write binary data.", "Error", MessageBoxButtons.OK,
                    MessageBoxIcon.Error);

            if (saveToArchive)
            {
                if (saveArchive(archiveFilename, filesList, progressCallback) != 0)
                {
                    MessageBox.Show("Failed to save project archive.", "Error", MessageBoxButtons.OK,
                        MessageBoxIcon.Error);
                    if (progressCallback != null)
                        progressCallback(0, "Failed to save project archive.");
                }
                else
                    if (progressCallback != null)
                        progressCallback(0, "Done");

                deleteFiles(filesList.ToArray());
            }
            else
                if (progressCallback != null)
                    progressCallback(0, "Done");

            modified = false;

            return ErrorCodes.ERR_NOERR;
        }

        /**
         * displays the project settings dialog
         * \return returns button the user pressed
         */
        public DialogResult showForm()
        {
            ProjectManagerForm projectForm = new ProjectManagerForm();
            projectForm.refreshProjectInfo(project, fileType, location, pluginManager.activeImportPlugin);
            DialogResult result = projectForm.ShowDialog();
            projectForm.Dispose();

            return result;
        }

        /**
         * loads a compressed multidac project archive
         * \param filename the archive file
         * \param files list of files extracted from the archive
         * \param cachedDirs list of cached directories (e.g. including source files) 
         * \param progressCallback progress function indicating loading progress
         * \return ERR_NOERR on success
         */
        private int loadArchive(string filename, List<string> files, List<string> cachedDirs,
            ProgressCallback progressCallback)
        {
            if (filename != null && files != null && cachedDirs != null)
            {
                if (!File.Exists(filename))
                    return ErrorCodes.ERR_FILENOTFOUND;

                ZipFile zipFile;
                try
                {
                    zipFile = new ZipFile(filename);

                    if (progressCallback != null)
                        progressCallback(25, "Decompressing data");

                    foreach (ZipEntry entry in zipFile)
                    {
                        string fileOnDisk;
                            
                        if (entry.IsDirectory)
                        {
                            fileOnDisk = Path.GetTempPath() + "multidac/" + entry.Name;
                            Directory.CreateDirectory(fileOnDisk);
                            cachedDirs.Add(fileOnDisk);
                        }

                        if (entry.IsFile)
                        {
                            if (Path.GetDirectoryName(entry.Name).CompareTo("") == 0)
                                fileOnDisk = Path.GetTempPath() + entry.Name;
                            else
                                fileOnDisk = Path.GetTempPath() + "multidac/" + entry.Name;
                            
                            FileStream fs = new FileStream(fileOnDisk, FileMode.Create);
                            if (fs == null)
                                return ErrorCodes.ERR_NULL;

                            byte[] buffer = new byte[entry.Size];
                            zipFile.GetInputStream(entry).Read(buffer, 0, Convert.ToInt32(entry.Size));
                            fs.Write(buffer, 0, Convert.ToInt32(entry.Size));
                            fs.Close();

                            // only add top level entries to this list
                            if (Path.GetDirectoryName(entry.Name) == "")
                                files.Add(fileOnDisk);
                        }
                    }

                    zipFile.Close();
                }
                catch (Exception e)
                {
#if DEBUG
                    Console.WriteLine(e.Message);
#endif
                    ErrorCodes.error(ErrorCodes.ERR_UNKNOWN);
                    return ErrorCodes.ERR_UNKNOWN;
                }

                return ErrorCodes.ERR_NOERR;
            }
            else
                return ErrorCodes.ERR_NULL;
        }

        /**
         * saves all project files into a compressed multidac archive file
         * \param filename archive file
         * \param files list of files to be added to the archive
         * \param progressCallback callback function indicating saving progress
         * \return ERR_NOERR on success
         */
        private int saveArchive(string filename, List<string> files, ProgressCallback progressCallback)
        {
            if (filename != null && files != null)
            {
                ZipFile zipFile;
                try
                {
                    zipFile = ZipFile.Create(filename);

                    if (zipFile != null)
                    {
                        if (progressCallback != null)
                            progressCallback(60, "Compressing project files");

                        zipFile.BeginUpdate();
                        foreach (string file in files)
                            zipFile.Add(file, Path.GetFileName(file));
                        
                        if (project.includeSourceFiles)
                        {
                            if (progressCallback != null)
                                progressCallback(80, "Compressing source files");

                            zipFile.AddDirectory("sourcefiles");
                            foreach (string file in project.sourceFiles)
                                zipFile.Add(file, "sourcefiles/" + Path.GetFileName(file));
                        }

                        zipFile.CommitUpdate();
                        zipFile.Close();                        
                    }
                }
                catch (Exception e)
                {
#if DEBUG
                    Console.WriteLine(e.Message);
#endif
                    ErrorCodes.error(ErrorCodes.ERR_UNKNOWN);
                    return ErrorCodes.ERR_UNKNOWN;
                }

                return ErrorCodes.ERR_NOERR;
            }
            else
                return ErrorCodes.ERR_NULL;
        }
    }
}
