﻿/*
multiDAC - Movement and deformation analysis tool
Renderer for displaying frames and result data (vectors)
Copyright (C) 2009 Felix Schmitt

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using PluginInterfaceLibrary;
using InternalData;

namespace Manager
{
    /**
     * \class Renderer
     * rendering class for displaying loaded images, result data etc.
     */
    public class Renderer
    {
        public bool enabled { get; internal set; }
        public Size dimension { get; internal set; }
        public RectangleF scaling { get; internal set; }
        public RectangleF destination { get; internal set; }
        public bool fitWindow { get; internal set; }
        public bool draggable { get; internal set; }
        public bool vectorsInSplitScreen;

        private bool updatePending = true;
        private Bitmap overlayBuffer1 = null;
        private Bitmap overlayBuffer2 = null;

        private int _lineWidth;
        private float _zoom;
        private Size dragMaximum;
        private Size _draggedOffset;

        /** line width of result vectors **/
        public int lineWidth
        {
            get { return _lineWidth; }
            set
            {
                _lineWidth = value;
                if (value < 0)
                    _lineWidth = 0;
                if (value > 10)
                    _lineWidth = 10;
            }
        }

        /** the zoom value the frame is displayed with (if fitWindow == false) **/
        public int Zoom
        {
            get
            {
                return Convert.ToInt32(scaling.Width * 100.0f);
            }
            set
            {
                if (value > 0 && value < 1000)
                {
                    fitWindow = false;
                    _zoom = value / 100.0f;
                    _draggedOffset = new Size(0, 0);
                }
            }
        }

        /** specifies the image's displacement after dragging it **/
        public Size DraggedOffset
        {
            get
            {
                return _draggedOffset;
            }
            set
            {
                if (value.Width < -dragMaximum.Width)
                    _draggedOffset.Width = -dragMaximum.Width;
                else
                    if (value.Width > dragMaximum.Width)
                        _draggedOffset.Width = dragMaximum.Width;
                    else
                        _draggedOffset.Width = value.Width;

                if (value.Height < -dragMaximum.Height)
                    _draggedOffset.Height = -dragMaximum.Height;
                else
                    if (value.Height > dragMaximum.Height)
                        _draggedOffset.Height = dragMaximum.Height;
                    else
                        _draggedOffset.Height = value.Height;
            }
        }

        /** constructor **/
        public Renderer()
        {
            lineWidth = 1;
            enabled = true;
            fitWindow = true;
            reset();
        }

        /** enables renderer **/
        public void enable()
        {
            enabled = true;
        }

        /** disables renderer **/
        public void disable()
        {
            enabled = false;
        }

        public void update()
        {
            updatePending = true;
        }

        /** resets renderer to default values **/
        public void reset()
        {
            scaling = new RectangleF(0, 0, 1.0f, 1.0f);
            dimension = new Size(0, 0);
            destination = new Rectangle(0, 0, 0, 0);
            _zoom = 1.0f;
            _draggedOffset = new Size(0, 0);
            draggable = false;
            dragMaximum = new Size(0, 0);
            vectorsInSplitScreen = true;
            update();
        }

        /** sets fitWindow property to true **/
        public void zoomFitWindow()
        {
            fitWindow = true;
            _draggedOffset = new Size(0, 0);
        }

        /**
         * draws a single result vector
         * \param graphics Graphics object to render to
         * \param vector vector data to render
         * \param scaleRect scaling dimensions (ignored for scale == false)
         * \param color the Color of the vector
         * \return ERR_NOERR on success
         */
        private int renderVector(Graphics graphics, Vector vector, RectangleF scaleRect, Color color)
        {
            if (graphics != null)
            {
                Pen p = new Pen(color, lineWidth);
                int r = 3;

                try
                {
                    /*graphics.DrawEllipse(p, new RectangleF(scaleRect.Width*(vector.start.X - 1), 
                        scaleRect.Height*(vector.start.Y - 1), 
                        scaleRect.Width * 2, scaleRect.Height * 2));*/
                    graphics.DrawLine(p, scaleRect.Width * vector.start.X - r, scaleRect.Height * vector.start.Y,
                        scaleRect.Width * vector.start.X + r, scaleRect.Height * vector.start.Y);
                    graphics.DrawLine(p, scaleRect.Width * vector.start.X, scaleRect.Height * vector.start.Y - r,
                        scaleRect.Width * vector.start.X, scaleRect.Height * vector.start.Y + r);

                    graphics.DrawLine(p,
                        new PointF(scaleRect.Width * vector.start.X, scaleRect.Height * vector.start.Y),
                        new PointF(scaleRect.Width * vector.target.X, scaleRect.Height * vector.target.Y));

                }
                catch (ArgumentNullException)
                {
                    return ErrorCodes.ERR_NULL;
                }

                return ErrorCodes.ERR_NOERR;
            }
            else
                return ErrorCodes.ERR_NULL;
        }

        /**
         * displays "[No frames]"
         * \param graphics Graphics object to render to
         * \param clipRectangle clipping bounds
         * \return ERR_NOERR on success
         */
        private int renderNoFrames(Graphics graphics, RectangleF clipRectangle)
        {
            if (graphics != null)
            {
                try
                {
                    graphics.DrawString("[No frames]", new Font(FontFamily.GenericSansSerif, 10.0f), Brushes.White,
                        new PointF(clipRectangle.Width / 2 - 20, clipRectangle.Height / 2));
                }
                catch (ArgumentNullException)
                {
                    return ErrorCodes.ERR_NULL;
                }

                return ErrorCodes.ERR_NOERR;
            }
            else
                return ErrorCodes.ERR_NULL;
        }

        /**
         * displays "[No import plugin]"
         * \param graphics Graphics object to render to
         * \param clipRectangle clipping bounds
         * \return ERR_NOERR on success
         */
        private int renderNoPlugin(Graphics graphics, RectangleF clipRectangle)
        {
            if (graphics != null)
            {
                try
                {
                    graphics.DrawString("[No import plugin]", new Font(FontFamily.GenericSansSerif, 10.0f), Brushes.White,
                        new PointF(clipRectangle.Width / 2 - 35, clipRectangle.Height / 2));
                }
                catch (ArgumentNullException)
                {
                    return ErrorCodes.ERR_NULL;
                }

                return ErrorCodes.ERR_NOERR;
            }
            else
                return ErrorCodes.ERR_NULL;
        }

        /*private int renderFrameInfo(Graphics graphics, RectangleF clipRectangle, Size frameInfo)
        {
            if (graphics != null)
            {
                try
                {
                    graphics.DrawString(frameInfo.Width.ToString() + "x" + frameInfo.Height.ToString() + "px",
                        new Font(FontFamily.GenericSansSerif, 7.0f), Brushes.White,
                        new PointF(clipRectangle.Width - 60, clipRectangle.Height - 20));
                    graphics.DrawString(Zoom.ToString() + "%",
                        new Font(FontFamily.GenericSansSerif, 7.0f), Brushes.White,
                        new PointF(clipRectangle.Width - 60, clipRectangle.Height - 30));
                }
                catch (ArgumentNullException)
                {
                    return ErrorCodes.ERR_NULL;
                }

                return ErrorCodes.ERR_NOERR;

            }
            else
                return ErrorCodes.ERR_NULL;
        }*/

        /**
         * renders crossed lines indicating the mouse's position in both screens
         * \param graphics Graphics object to render to
         * \param bufferSize size of the target bitmap
         * \param scaling scaling parameters to be applied
         * \param mousePos the current mouse position (mousePos : image == 1:1)
         * \return ERR_NOERR on success
         */
        private int renderMouse(Graphics graphics, SizeF bufferSize, RectangleF scaling, Point mousePos)
        {
            if (graphics != null)
            {
                Pen p = new Pen(Color.Green, 1);

                try
                {
                    graphics.DrawLine(p, mousePos.X * scaling.Width, 0,
                        mousePos.X * scaling.Width, bufferSize.Height);
                    graphics.DrawLine(p, 0, mousePos.Y * scaling.Height,
                        bufferSize.Width, mousePos.Y * scaling.Height);
                }
                catch (ArgumentNullException)
                {
                    return ErrorCodes.ERR_NULL;
                }

                return ErrorCodes.ERR_NOERR;
            }
            else
                return ErrorCodes.ERR_NULL;
        }

        /**
         * calculates dimensional and scaling parameters
         * \param dimension size of the source images
         * \param screenSize sizeof the target screen (both screens)
         * \param dstScreenWidth width of the target screen (one screen)
         * \param fitWindow if content is fittet into teh clip rectangle
         * \param dstRect reference to destination rectangle
         * \param scaleRect reference to scaling rectangle
         * \return ERR_NOERR on success
         */
        private int calculateDimensions(Size dimension, SizeF screenSize, float dstScreenWidth, bool fitWindow,
            ref RectangleF dstRect, ref RectangleF scaleRect)
        {
            try
            {
                if (fitWindow)
                {
                    dstRect.Width = dstScreenWidth;
                    dstRect.Height = dimension.Height * (dstScreenWidth / dimension.Width);
                    if (dstRect.Height > screenSize.Height)
                    {
                        dstRect.Height = screenSize.Height;
                        dstRect.Width = dimension.Width * (screenSize.Height / dimension.Height);
                    }

                    dstRect.X = dstScreenWidth / 2 - dstRect.Width / 2;
                    dstRect.Y = screenSize.Height / 2 - dstRect.Height / 2;

                    scaleRect.X = dstRect.X;
                    scaleRect.Y = dstRect.Y;
                    scaleRect.Width = dstRect.Width / dimension.Width;
                    scaleRect.Height = dstRect.Height / dimension.Height;
                }
                else
                {
                    dstRect.Width = dimension.Width * _zoom;
                    dstRect.Height = dimension.Height * _zoom;

                    dstRect.X = (dstScreenWidth - dstRect.Width) / 2;
                    dstRect.Y = (screenSize.Height - dstRect.Height) / 2;

                    scaleRect.X = dstRect.X;
                    scaleRect.Y = dstRect.Y;
                    scaleRect.Width = dstRect.Width / dimension.Width;
                    scaleRect.Height = dstRect.Height / dimension.Height;
                }

                draggable = false;
                dragMaximum = new Size(0, 0);

                if (dstScreenWidth >= dstRect.Width)
                    _draggedOffset = new Size(0, _draggedOffset.Height);
                else
                {
                    draggable = true;
                    dragMaximum.Width = Convert.ToInt32((dstRect.Width - dstScreenWidth) / 2);
                }

                if (screenSize.Height >= dstRect.Height)
                    _draggedOffset = new Size(_draggedOffset.Width, 0);
                else
                {
                    draggable = true;
                    dragMaximum.Height = Convert.ToInt32((dstRect.Height - screenSize.Height) / 2);
                }

                // because dragMaximum could have updated this
                DraggedOffset = DraggedOffset;

                scaleRect.X += _draggedOffset.Width;
                scaleRect.Y += _draggedOffset.Height;

                dstRect.X += _draggedOffset.Width;
                dstRect.Y += _draggedOffset.Height;
            }
            catch (OverflowException)
            {
                return ErrorCodes.ERR_OVERFLOW;
            }
            catch (DivideByZeroException)
            {
                return ErrorCodes.ERR_FORMATERROR;
            }
            catch (Exception)
            {
                return ErrorCodes.ERR_UNKNOWN;
            }

            return ErrorCodes.ERR_NOERR;
        }

        /**
         * retrieves source images from the plugin
         * \param data DataPackage containing vectors
         * \param importPlugin the plugin to get images from
         * \param dimension size of the source images
         * \param destination target location and size
         * \param gUnit GraphicsUnit used for drawing on Graphics object
         * \param frameNum frame number of the first image
         * \param splitScreen indicates if two images are required
         * \param renderVectors indicates if vectors should be displayed
         * \param renderChangesOnly indicates if only moving vectors should be displayed
         */
        private int updateImages(DataPackage data, IImportPlugin importPlugin, Size dimension, RectangleF destination, 
            GraphicsUnit gUnit, int frameNum, bool splitScreen, bool renderVectors, bool renderChangesOnly)
        {
            int errResult = ErrorCodes.ERR_NOERR;

            Bitmap srcBuffer1 = null;
            Bitmap srcBuffer2 = null;

            // grab the source frames
            Graphics srcBuffer1Graphics = null;
            Graphics srcBuffer2Graphics = null;

            try
            {
                srcBuffer1 = new Bitmap(dimension.Width, dimension.Height);
                srcBuffer1Graphics = Graphics.FromImage(srcBuffer1);

                if (splitScreen)
                {
                    srcBuffer2 = new Bitmap(dimension.Width, dimension.Height);
                    srcBuffer2Graphics = Graphics.FromImage(srcBuffer2);
                    srcBuffer2Graphics.FillRectangle(Brushes.Black, new Rectangle(0, 0, dimension.Width, dimension.Height));
                }
            }
            catch (Exception)
            {
                return ErrorCodes.ERR_UNKNOWN;
            }

            errResult = importPlugin.renderFrame(frameNum, srcBuffer1Graphics);
            if (errResult != ErrorCodes.ERR_NOERR)
                return errResult;

            if (splitScreen && frameNum + 1 < importPlugin.frameCount && srcBuffer2Graphics != null)
            {
                errResult = importPlugin.renderFrame(frameNum + 1, srcBuffer2Graphics);
                if (errResult != ErrorCodes.ERR_NOERR)
                    return errResult;
            }

            Graphics overlayBuffer1Graphics = null;
            Graphics overlayBuffer2Graphics = null;

            // render the source frames to their destination size and add results data etc. 
            try
            {
                overlayBuffer1 = new Bitmap(Convert.ToInt32(destination.Width), Convert.ToInt32(destination.Height));
                overlayBuffer1Graphics = Graphics.FromImage(overlayBuffer1);

                if (splitScreen)
                {
                    overlayBuffer2 = new Bitmap(Convert.ToInt32(destination.Width), Convert.ToInt32(destination.Height));
                    overlayBuffer2Graphics = Graphics.FromImage(overlayBuffer2);
                }
            }
            catch (Exception)
            {
                return ErrorCodes.ERR_UNKNOWN;
            }

            try
            {
                overlayBuffer1Graphics.DrawImage(srcBuffer1, new RectangleF(0, 0, destination.Width, destination.Height),
                    srcBuffer1.GetBounds(ref gUnit), gUnit);


                if (splitScreen && overlayBuffer2Graphics != null && srcBuffer2 != null)
                {
                    overlayBuffer2Graphics.DrawImage(srcBuffer2, new RectangleF(0, 0, destination.Width, destination.Height),
                        srcBuffer2.GetBounds(ref gUnit), gUnit);
                }
            }
            catch (ArgumentNullException)
            {
                return ErrorCodes.ERR_NULL;
            }

            if (renderVectors && data != null && data.vectors.Length > 0)
                for (int i = 0; i < data.vectors.Length; i++)
                    if (data.vectors[i].valid)
                    {
                        if (renderChangesOnly && data.vectors[i].start == data.vectors[i].target)
                            continue;

                        errResult = renderVector(overlayBuffer1Graphics, data.vectors[i], scaling, Color.Red);
                        if (errResult != ErrorCodes.ERR_NOERR)
                        {
                            ErrorCodes.error(errResult);
                            break;
                        }

                        if (splitScreen && vectorsInSplitScreen && overlayBuffer2Graphics != null)
                        {
                            errResult = renderVector(overlayBuffer2Graphics, data.vectors[i], scaling, Color.LightGray);
                            if (errResult != ErrorCodes.ERR_NOERR)
                            {
                                ErrorCodes.error(errResult);
                                break;
                            }
                        }
                    }

            return ErrorCodes.ERR_NOERR;
        }

        /**
         * draws current frame, result data or messages
         * \param graphics Graphics object to render to
         * \param clipRectangle clipping bounds
         * \param importPlugin active import plugin
         * \param frameNum index of current frame
         * \param data data package with result vectors
         * \param mousePos position of cursor over the screens
         * \param renderVectors set if vectors should be rendered
         * \param renderChangesOnly set if only moving vectors should be rendered
         * \param splitScreen set if the next frame should be rendered, too
         * \return ERR_NOERR on success
         */
        public int render(Graphics graphics, RectangleF clipRectangle, IImportPlugin importPlugin, int frameNum,
            DataPackage data, Point mousePos, bool renderVectors, bool renderChangesOnly, bool splitScreen)
        {
            int errResult = ErrorCodes.ERR_NOERR;

            if (!enabled)
                return ErrorCodes.ERR_NOERR;

            scaling = new RectangleF(0, 0, 1.0f, 1.0f);

            SizeF screenSize = new SizeF(clipRectangle.Width, clipRectangle.Height);
            float dstScreenWidth = screenSize.Width;
            bool doRenderMousePos = mousePos.X != -1 && mousePos.Y != -1;

            if (splitScreen)
                dstScreenWidth = dstScreenWidth / 2.0f;

            if (importPlugin != null)
            {
                // retrieve dimension of source frames
                Size srcDimension = new Size(0, 0);
                errResult = importPlugin.getDimension(frameNum, ref srcDimension);
                if (errResult != ErrorCodes.ERR_NOERR)
                {
                    renderNoFrames(graphics, clipRectangle);
                    draggable = false;
                    return errResult;
                }

                dimension = srcDimension;
                GraphicsUnit gUnit = GraphicsUnit.Display;

                // calculate scaling and destination dimensions
                RectangleF dstRect = new RectangleF(0, 0, dimension.Width, dimension.Height);
                RectangleF scaleRect = new RectangleF(0, 0, 1.0f, 1.0f);

                errResult = calculateDimensions(dimension, screenSize, dstScreenWidth, fitWindow, ref dstRect, ref scaleRect);
                if (errResult != ErrorCodes.ERR_NOERR)
                    return errResult;

                scaling = scaleRect;
                destination = dstRect;

                // this should avoid unneccessary loading
                if (updatePending)
                {
                    // grab the source images
                    errResult = updateImages(data, importPlugin, dimension, destination, gUnit, frameNum, 
                        splitScreen, renderVectors, renderChangesOnly);

                    if (errResult != ErrorCodes.ERR_NOERR)
                    {
                        ErrorCodes.error(errResult);
                        return errResult;
                    }

                    updatePending = false;
                }

                Bitmap mouseBuffer = null;

                if (splitScreen && doRenderMousePos)
                {
                    mouseBuffer = new Bitmap(Convert.ToInt32(destination.Width), Convert.ToInt32(destination.Height));
                    Graphics mouseGraphics = Graphics.FromImage(mouseBuffer);
                    mouseGraphics.FillRectangle(Brushes.Transparent, 0, 0, destination.Width, destination.Height);

                    errResult = renderMouse(mouseGraphics, destination.Size, scaling, mousePos);
                    if (errResult != ErrorCodes.ERR_NOERR)
                    {
                        ErrorCodes.error(errResult);
                        return errResult;
                    }
                }

                // now map the final images on a buffer representing the target Graphics object
                try
                {
                    Bitmap finalBuffer = new Bitmap(Convert.ToInt32(dstScreenWidth), Convert.ToInt32(screenSize.Height));
                    Graphics.FromImage(finalBuffer).DrawImage(overlayBuffer1, destination.X, destination.Y);
                    
                    if (doRenderMousePos && mouseBuffer != null)
                        Graphics.FromImage(finalBuffer).DrawImage(mouseBuffer, destination.X, destination.Y);

                    graphics.DrawImageUnscaled(finalBuffer, 0, 0);

                    if (splitScreen && overlayBuffer2 != null)
                    {
                        Graphics.FromImage(finalBuffer).DrawImage(overlayBuffer2, destination.X, destination.Y);
                        
                        if (doRenderMousePos && mouseBuffer != null)
                            Graphics.FromImage(finalBuffer).DrawImage(mouseBuffer, destination.X, destination.Y);
                        
                        graphics.DrawImageUnscaled(finalBuffer, Convert.ToInt32(dstScreenWidth), 0);

                        // border line between frames
                        graphics.DrawLine(Pens.Black, dstScreenWidth, 0, dstScreenWidth, screenSize.Height);
                    }

                    /*if (renderInfo)
                        renderFrameInfo(graphics, clipRectangle, dimension);*/
                }
                catch (ArgumentNullException)
                {
                    return ErrorCodes.ERR_NULL;
                }
            }
            else
            {
                errResult = renderNoPlugin(graphics, clipRectangle);
                draggable = false;
            }

            return errResult;
        }
    }
}
