﻿/*
multiDAC Processing Plugin
Copyright (C) 2009 Felix Schmitt

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.IO;
using System.Reflection;
using System.Runtime.InteropServices;
using PluginInterfaceLibrary;
using InternalData;

namespace ProcessingPluginLib
{
    /**
     * \class ProcesingPlugin
     * An implementation of the IProcessingPlugin interface.
     * This plugin uses libProcCuda.dll for processing data on Cuda compatible GPUs.
     */
    public class ProcessingPlugin : IProcessingPlugin
    {
        private const string MY_VERSION = "0.1.8";

#if x64
        private const string CUDA_DLL = "libProcCuda64.dll";
#else
        private const string CUDA_DLL = "libProcCuda.dll";
#endif

        [DllImport("kernel32.dll", CharSet = CharSet.Auto)]
        static extern IntPtr LoadLibrary(string lpFileName);

        [DllImport(CUDA_DLL)]
        static extern int init(int frmWidth, int frmHeight);

        [DllImport(CUDA_DLL)]
        static extern void deInit();

        [DllImport(CUDA_DLL)]
        static extern int computeMedians(int frameNum, int frameIndex, int[,] data,
            float[,] median, float[,] medianSqr, int halfKernel);

        [DllImport(CUDA_DLL)]
        static extern int testDevice(ref int maxThreadsPerBlock, int[] maxThreadsDim, int[] maxGridSize,
            ref int major, ref int minor);

        [StructLayout(LayoutKind.Sequential)]
        public struct DllVector
        {
            public int startX;
            public int startY;
            public int targetX;
            public int targetY;
        }

        /** delegate callback method for indicating progress of C dlls **/
        private delegate void ProgressCallbackC(int progress);

        private ProgressCallback progCallbackMain;

        [DllImport(CUDA_DLL)]
        static extern int computeCorrelation1(int[,] vectors, int vectorCount, int halfKernel, int maxDelta,
            float minCoefficient, ProgressCallbackC progressCallback);

        [DllImport(CUDA_DLL)]
        static extern int computeCorrelation2(int[,] vectors, int vectorCount, int halfKernel, int maxDelta,
            float minCoefficient, ProgressCallbackC progressCallback);

        private bool gpuEnable = true;
        private bool gpuDllNotFound = false;
        private IntPtr dllPointer = IntPtr.Zero;
        private int gpuMode = 0;
        private bool preserveUnfound = false;
        private int kernelWidth = 11;
        private int searchFactor = 3;
        private float minCoefficient = 0.5f;

        private static int BLOCK_SIZE = 16;

        /**
         * \class BPFrame
         * Bravais-Pearson correlation compatible frame
         */
        private class BPFrame : AFrame
        {
            private Size allocDimension;

            public float[,] median { get; internal set; }
            public float[,] medianSqr { get; internal set; }

            /**
             * constructor
             * \param frameNum 0 or 1, depending on which frame it represents
             * \param frameIndex index of the frame
             * \param package data package with initial vectors
             * \param frame the Frame object used to generate this BPFrame
             * \param maxDelta maximum with of displacement, used for correlation
             * \param halfKernelWidth width of the correlation kernel (n-1)/2
             * \param gpuEnable set to true to enable GPU calculations
             * \param progressCallback callback function indicating processing progress
             */
            public BPFrame(int frameNum, int frameIndex, DataPackage package,
                Frame frame, int maxDelta, int halfKernelWidth, bool gpuEnable,
                ProgressCallback progressCallback)
            {
                this.dimension = frame.dimension;
                allocDimension = this.dimension;

                //this.data = frame.data;

                while (allocDimension.Width % BLOCK_SIZE != 0)
                    allocDimension.Width++;
                while (allocDimension.Height % BLOCK_SIZE != 0)
                    allocDimension.Height++;

                data = new int[allocDimension.Width, allocDimension.Height];
                median = new float[allocDimension.Width, allocDimension.Height];
                medianSqr = new float[allocDimension.Width, allocDimension.Height];

                // this speeds up the whole thing
                int[,] dataSqr = new int[allocDimension.Width, allocDimension.Height];

                for (int i = 0; i < dimension.Width; i++)
                    for (int j = 0; j < dimension.Height; j++)
                    {
                        data[i, j] = frame.data[i, j];
                        dataSqr[i, j] = (int)Math.Pow(frame.data[i, j], 2.0f);
                        median[i, j] = -1;
                        medianSqr[i, j] = -1;
                    }

                // choose between GPU based and CPU based code
                if (gpuEnable)
                {
                    computeMedians(frameNum, frameIndex, data, median, medianSqr, halfKernelWidth);
                }
                else
                {
                    for (int vectorIndex = 0; vectorIndex < package.vectors.Length; vectorIndex++)
                    {
                        Vector vector = package.vectors[vectorIndex];
                        System.Drawing.Point src = vector.start;

                        float hkwPow = (float)Math.Pow((halfKernelWidth * 2 + 1), 2);

                        // only calc necessary median values (never twice)
                        for (int j = -maxDelta; j <= maxDelta; j++)
                            for (int i = -maxDelta; i <= maxDelta; i++)
                            {
                                int px = src.X + i;
                                int py = src.Y + j;

                                if ((px >= halfKernelWidth) && (px < frame.dimension.Width - halfKernelWidth) &&
                                    (py >= halfKernelWidth) && (py < frame.dimension.Height - halfKernelWidth) &&
                                    (median[px, py] == -1) && (medianSqr[px, py] == -1))
                                {

                                    int sum = 0;
                                    int sumSqr = 0;
                                    for (int dy = -halfKernelWidth; dy <= halfKernelWidth; dy++)
                                        for (int dx = -halfKernelWidth; dx <= halfKernelWidth; dx++)
                                        {
                                            sum += data[px + dx, py + dy];
                                            sumSqr += dataSqr[px + dx, py + dy];
                                        }

                                    median[px, py] = (float)(sum / hkwPow);
                                    medianSqr[px, py] = (float)(sumSqr / hkwPow);
                                }
                            }

                        if (progressCallback != null)
                            progressCallback(25 * frameNum + (vectorIndex + 1) * 25 / package.vectors.Length,
                                "Preprocessing...");
                    }
                }
            }
        }

        /** name of this plugin **/
        public string name
        {
            get { return "Correlation plugin"; }
        }

        /** vendor of this plugin **/
        public string vendor
        {
            get { return "Felix Schmitt"; }
        }

        /** description of this plugin **/
        public string description
        {
            get { return "Plugin for processing data using grayscale correlation. Supports processing on GPU."; }
        }

        /** version of this plugin **/
        public string version
        {
            get { return MY_VERSION; }
        }

        /** if this plugin requires configure() being called before processing **/
        public bool requiresConfig
        {
            get { return false; }
        }

        /** if this plugin supports being called in another thread **/
        public bool supportsThreading
        {
            get { return !gpuEnable; }
        }

        /** detect system architechture **/
        private bool X64Arch
        {
            get { return Marshal.SizeOf(typeof(IntPtr)) == 8; }
        }

        /** constructor **/
        public ProcessingPlugin()
        {
            if (!X64Arch)
            {
                gpuDllNotFound = false;
                Environment.CurrentDirectory = Path.GetDirectoryName(Assembly.GetAssembly(this.GetType()).Location);
                string dllPath = Path.GetFullPath(@CUDA_DLL);

                if (File.Exists(dllPath))
                {
                    dllPointer = LoadLibrary(dllPath);
                    if (dllPointer == IntPtr.Zero)
                        gpuDllNotFound = true;
#if !DEBUG
                    dllPath = Path.GetFullPath(@"cudart.dll");
                    dllPointer = LoadLibrary(dllPath);
                    if (dllPointer == IntPtr.Zero)
                        gpuDllNotFound = true;
#endif
                }
                else
                    gpuDllNotFound = true;
            }
        }

        /**
         * configures this plugin
         * \return ERR_NOERR on success
         **/
        public int configure()
        {
            ConfigForm configForm = new ConfigForm(X64Arch);
            configForm.gpuEnable = gpuEnable;
            configForm.gpuMode = gpuMode;
            configForm.preserveUnfound = preserveUnfound;
            configForm.kernelWidth = kernelWidth;
            configForm.searchFactor = searchFactor;
            configForm.minCoefficient = Convert.ToInt32(minCoefficient * 100.0f);

            try
            {
                if (configForm.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                {
                    gpuEnable = configForm.gpuEnable;
                    gpuMode = configForm.gpuMode;
                    preserveUnfound = configForm.preserveUnfound;
                    kernelWidth = configForm.kernelWidth;
                    searchFactor = configForm.searchFactor;
                    minCoefficient = Convert.ToSingle(configForm.minCoefficient) / 100.0f;
                }
            }
            catch (Exception)
            {
                configForm.Dispose();
                return ErrorCodes.ERR_UNKNOWN;
            }

            configForm.Dispose();
            return ErrorCodes.ERR_NOERR;
        }

        /**
         * should be called before processing starts
         * used to initialize some plugin parameters and test device functionality
         * \param frameDimension size of the frames
         * \return 0 on success
         */
        public int beginProcessing(Size frameDimension)
        {
            int foo = 0;
            int[] foo2 = new int[2];

            if (X64Arch || gpuDllNotFound || testDevice(ref foo, foo2, foo2, ref foo, ref foo) != 0)
                gpuEnable = false;

            if (gpuEnable)
            {
                Size allocDimension = frameDimension;

                while (allocDimension.Width % BLOCK_SIZE != 0)
                    allocDimension.Width++;
                while (allocDimension.Height % BLOCK_SIZE != 0)
                    allocDimension.Height++;

                int initResult = init(allocDimension.Width, allocDimension.Height);
                if (initResult != 0)
                    return ErrorCodes.ERR_UNKNOWN;
            }

            return ErrorCodes.ERR_NOERR;
        }

        /**
         * should be called after processing
         * used to clean up plugin
         * \return 0 on success
         */
        public int endProcessing()
        {
            if (gpuEnable)
                deInit();

            return ErrorCodes.ERR_NOERR;
        }

        /**
         * calculates te median value between two frames
         * \param frame1 first frame
         * \param frame2 second frame
         * \param centerX1 x coordinate in frame1
         * \param centerY1 y coordinate in frame1
         * \param centerX2 x coordinate in frame2
         * \param centerY2 y coordinate in frame2
         * \param halfWidth half kernel width
         * \return the median value
         */
        private float medianCrossFrames(BPFrame frame1, BPFrame frame2, int centerX1, int centerY1,
            int centerX2, int centerY2, int halfWidth)
        {
            int medianSum = 0;
            for (int dy = -halfWidth; dy <= halfWidth; dy++)
                for (int dx = -halfWidth; dx <= halfWidth; dx++)
                    medianSum += frame1.data[centerX1 + dx, centerY1 + dy] *
                        frame2.data[centerX2 + dx, centerY2 + dy];

            return (float)(medianSum / (Math.Pow(halfWidth * 2 + 1, 2)));
        }

        /**
         * returns the maximum in the Bravais-Pearson coefficient matrix
         * \param bpCoeffMatrix a 2-dimensional array to search in
         * \param resX reference to result position x in the array
         * \param resY reference to result position y in the array
         * \return the maximum coefficient
         */
        private float getMaxCoeff(float[,] bpCoeffMatrix, ref int resX, ref int resY)
        {
            float tmpMax = 0;
            resX = 0;
            resY = 0;

            for (int i = 0; i < bpCoeffMatrix.GetLength(0); i++)
                for (int j = 0; j < bpCoeffMatrix.GetLength(1); j++)
                    if (bpCoeffMatrix[i, j] > tmpMax)
                    {
                        tmpMax = bpCoeffMatrix[i, j];
                        resX = i;
                        resY = j;
                    }
            return tmpMax;
        }

        /**
         * this function is called from libProcCuda.dll
         * \param progress progress of processing in the dll
         * 
         * \bug #0005 this is still buggy (null-ref-exceptions)
         */
        private void dllProgressCallback(int progress)
        {
            if (progCallbackMain != null)
            {
                if (progress < 0)
                    progress = 0;
                if (progress > 100)
                    progress = 100;
                progCallbackMain(progress, "");
            }
        }

        /**
         * processes two frames and stores data
         * \param frame1 first frame object
         * \param frame2 second frame object
         * \param vectorData reference to data package with initial and result vectors
         * \param dataLen number of initial vectors
         * \param progressCallback callback function indicating processing progress
         * \return 0 on success
         * 
         * \bug #0003 gpuMode loses vectors (from time to time ?)
         * \bug #0004 cpu mode doesn't generate any vectors on Linux
         */
        public int processData(Frame frame1, Frame frame2,
            ref DataPackage vectorData, int dataLen, ProgressCallback progressCallback)
        {
            int halfKernelWidth = (int)((kernelWidth - 1) / 2);
            int maxDelta = (int)((searchFactor - 1) * kernelWidth / 2);

            progCallbackMain = progressCallback;

            if (progressCallback != null)
                progressCallback(0, "Initializing frames...");

            if (dataLen > 0)
            {
                BPFrame bpFrame1 = bpFrame1 = new BPFrame(0, vectorData.frame1, vectorData, frame1,
                        maxDelta, halfKernelWidth, gpuEnable, progressCallback);
                BPFrame bpFrame2 = bpFrame2 = new BPFrame(1, vectorData.frame2, vectorData, frame2,
                       maxDelta, halfKernelWidth, gpuEnable, progressCallback);

                if (gpuEnable)
                {
                    int[,] dllVectors = new int[dataLen, 4];
                    for (int i = 0; i < dataLen; i++)
                    {
                        dllVectors[i, 0] = vectorData.vectors[i].start.X;
                        dllVectors[i, 1] = vectorData.vectors[i].start.Y;
                        dllVectors[i, 2] = -1;
                        dllVectors[i, 3] = -1;
                    }

                    int gpuResult = 0;

                    // see kernels.cu for explanation why progressCallback is null.
                    switch (gpuMode)
                    {
                        case 0:
                            gpuResult = computeCorrelation1(dllVectors, dataLen, halfKernelWidth, maxDelta,
                                minCoefficient, null);
                            break;
                        case 1:
                            gpuResult = computeCorrelation2(dllVectors, dataLen, halfKernelWidth, maxDelta,
                                minCoefficient, null);
                            break;
                    }

                    if (gpuResult != 0)
                        return gpuResult;

                    for (int i = 0; i < dataLen; i++)
                    {
                        vectorData.vectors[i].start.X = dllVectors[i, 0];
                        vectorData.vectors[i].start.Y = dllVectors[i, 1];

                        if (!preserveUnfound || (dllVectors[i, 2] != -1 && dllVectors[i, 3] != -1))
                        {
                            vectorData.vectors[i].target.X = dllVectors[i, 2];
                            vectorData.vectors[i].target.Y = dllVectors[i, 3];
                        }
                        else
                            if (dllVectors[i, 2] == -1 || dllVectors[i, 3] == -1)
                            {
                                vectorData.vectors[i].target.X = dllVectors[i, 0];
                                vectorData.vectors[i].target.Y = dllVectors[i, 1];
                            }

                    }
                } // end of GPU code
                else
                {
                    for (int i = 0; i < dataLen; i++)
                    {
                        System.Drawing.Point src = vectorData.vectors[i].start;

                        if ((src.X < halfKernelWidth) || (src.X >= frame1.dimension.Width - halfKernelWidth) ||
                            (src.Y < halfKernelWidth) || (src.Y >= frame1.dimension.Height - halfKernelWidth))
                        {
                            vectorData.vectors[i].target.X = -1;
                            vectorData.vectors[i].target.Y = -1;
                            continue;
                        }

                        //float[,] bpCoeffMatrix = new float[searchSize - kernelWidth + 1, searchSize - kernelWidth + 1];

                        float bpCoeffMax = 0;
                        int coeffMaxX = 0;
                        int coeffMaxY = 0;

                        for (int dx = -maxDelta; dx <= maxDelta; dx++)
                            for (int dy = -maxDelta; dy <= maxDelta; dy++)
                                // just some deltas, not all
                                if ((src.X + dx >= halfKernelWidth) && (src.X + dx < frame1.dimension.Width - halfKernelWidth) &&
                                    (src.Y + dy >= halfKernelWidth) && (src.Y + dy < frame1.dimension.Height - halfKernelWidth))
                                {
                                    float medianCross = medianCrossFrames(bpFrame1, bpFrame2,
                                        src.X, src.Y, src.X + dx, src.Y + dy, halfKernelWidth);

                                    float sum1 = medianCross - bpFrame1.median[src.X, src.Y] * bpFrame2.median[src.X + dx, src.Y + dy];
                                    float sum2 = (bpFrame1.medianSqr[src.X, src.Y] - (float)Math.Pow(bpFrame1.median[src.X, src.Y], 2)) *
                                        (bpFrame2.medianSqr[src.X + dx, src.Y + dy] - (float)Math.Pow(bpFrame2.median[src.X + dx, src.Y + dy], 2));

                                    float bpCoeff = 0;

                                    if (sum2 > 0)
                                    {
                                        sum2 = (float)Math.Sqrt(sum2);
                                        if (sum2 != 0)
                                            bpCoeff = Math.Abs(sum1 / sum2);
                                    }

                                    if (bpCoeff > bpCoeffMax)
                                    {
                                        bpCoeffMax = bpCoeff;
                                        coeffMaxX = src.X + dx;
                                        coeffMaxY = src.Y + dy;
                                    }

                                    //bpCoeffMatrix[dx + maxDelta, dy + maxDelta] = bpCoeff;
                                }

                        // get max bpCoeff from Matrix
                        //int targetX = 0;
                        //int targetY = 0;
                        //float maxBPCoeff = getMaxCoeff(bpCoeffMatrix, ref targetX, ref targetY);
                        float maxBPCoeff = bpCoeffMax;
                        if (maxBPCoeff > minCoefficient)
                        {
                            vectorData.vectors[i].target.X = coeffMaxX;//targetX - maxDelta + vectorData.vectors[i].start.X;
                            vectorData.vectors[i].target.Y = coeffMaxY;//targetY - maxDelta + vectorData.vectors[i].start.Y;
                        }
                        else
                        {
                            if (!preserveUnfound)
                            {
                                vectorData.vectors[i].target.X = -1;
                                vectorData.vectors[i].target.Y = -1;
                            }
                            else
                            {
                                vectorData.vectors[i].target.X = vectorData.vectors[i].start.X;
                                vectorData.vectors[i].target.Y = vectorData.vectors[i].start.Y;
                            }
                        }

                        if (progressCallback != null)
                            progressCallback(50 + (i + 1) * 50 / dataLen, "Processing...");
                    }
                } // end of CPU code

            } // end of (dataLen > 0)

            return ErrorCodes.ERR_NOERR;
        }

    }
}
