﻿/*
multiDAC - Movement and deformation analysis tool
Custom FramesTrackBar component
Copyright (C) 2009 Felix Schmitt

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using InternalData;

namespace multidac
{
    /**
     * \class FramesTrackBar
     * A custom track bar for frame ranges, markers etc.
     */
    public class FramesTrackBar : Control
    {
        /**
         * \class Marker
         * A marker used to set start and end positions and additonal information
         */
        public class Marker
        {
            /**
             * \enum MarkerType
             * type of marker
             */
            public enum MarkerType { mtLeft, mtRight, mtNote };
            private const int MAX_MARKER_WIDTH = 10;
            
            private int _markerWidth;
            /** width of marker (for drawing) **/
            public int MarkerWidth
            {
                get
                {
                    return _markerWidth;
                }
                set
                {
                    _markerWidth = value / 2;
                    if (_markerWidth > MAX_MARKER_WIDTH)
                        _markerWidth = MAX_MARKER_WIDTH;
                }
            }

            private int _index;
            /** position of marker **/
            public int Index
            {
                get { return _index; }
                set
                {
                    if (value >= 0)
                        _index = value;
                }
            }

            private string _note;
            /** a text note attached to the marker (not yet used) **/
            public string Note
            {
                get { return _note; }
                set
                {
                    if (value != null)
                        _note = value;
                    else
                        _note = "";
                }
            }

            public Brush brush { get; internal set; }
            public MarkerType markerType { get; internal set; }

            /**
             * constructor
             * \param index position
             * \param note attached string note
             * \param markerType type of marker
             */
            public Marker(int index, string note, MarkerType markerType)
            {
                this.Index = index;
                this.Note = note;
                this.markerType = markerType;
                MarkerWidth = MAX_MARKER_WIDTH;

                switch (markerType)
                {
                    case MarkerType.mtLeft:
                        brush = Brushes.Red;
                        break;
                    case MarkerType.mtRight:
                        brush = Brushes.Red;
                        break;
                    default:
                        brush = Brushes.Blue;
                        break;
                }
            }

            /**
             * returns bounds of marker (position and dimension) in FramesTrackBar
             * \param ticksSpace current ticksSpace of FramesTrackBar
             * \param markerHeight height of marker region
             * \param headHeight height of frames/head region
             * \return bounds of the marker
             */
            public Rectangle getBounds(double ticksSpace, int markerHeight, int headHeight)
            {
                int posX = Convert.ToInt32(ticksSpace * (Index + 1.0f));
                return new Rectangle(posX - MarkerWidth, headHeight + markerHeight / 2 - MarkerWidth,
                    posX + MarkerWidth, headHeight + markerHeight / 2 + MarkerWidth);
            }

            /**
             * renders a marker
             * \param g Graphics object to render to
             * \param ticksSpace current ticksSpace of FramesTrackBar
             * \param markerHeight height of marker region
             * \param headHeight height of frames/head region
             * \param offset the offset to the borders from the positioner's head
             */
            public void render(Graphics g, double ticksSpace, int markerHeight, int headHeight, int offset)
            {
                if (g != null)
                {
                    float posX = Convert.ToSingle(offset + ticksSpace * (Index + 1.0f));
                    float centerY = headHeight + markerHeight / 2;

                    switch (markerType)
                    {
                        case MarkerType.mtRight:
                            PointF[] points1 = { new PointF(posX - MarkerWidth, centerY - MarkerWidth), 
                                                new PointF(posX + MarkerWidth, centerY), 
                                                new PointF(posX - MarkerWidth, centerY + MarkerWidth),
                                                new PointF(posX - MarkerWidth, centerY - MarkerWidth)
                                              };
                            g.FillPolygon(brush, points1);
                            break;
                        case MarkerType.mtLeft:
                            PointF[] points2 = { new PointF(posX + MarkerWidth, centerY - MarkerWidth), 
                                                new PointF(posX - MarkerWidth, centerY), 
                                                new PointF(posX + MarkerWidth, centerY + MarkerWidth),
                                                new PointF(posX + MarkerWidth, centerY - MarkerWidth)
                                              };
                            g.FillPolygon(brush, points2);
                            break;
                        case MarkerType.mtNote:
                            g.FillRectangle(brush, posX - MarkerWidth, centerY - MarkerWidth,
                                2 * MarkerWidth, 2 * MarkerWidth);
                            Pen exclamationPen = new Pen(Color.White, 2f);
                            g.DrawLine(exclamationPen, posX, centerY - MarkerWidth + 1,
                                posX, centerY + MarkerWidth / 2 - 1);
                            g.DrawEllipse(exclamationPen, posX - 1, centerY + MarkerWidth / 2,
                                1, 1);
                            break;
                    }
                }
            }
        }

        /**
         * \class Positioner
         * A custom trackbar thumb used in FramesTrackBar
         */
        class Positioner
        {
            /**
             * \enum PositionerState
             * current state
             */
            public enum PositionerState { psNormal, psPressed, psInvisible };
            public PositionerState state;

            private Size headSize;
            private Pen headPen;
            private double currentPositionX;

            private const int HEAD_WIDTH = 9;

            public int HeadOffset
            {
                get { return headSize.Width; }//Convert.ToInt32(Math.Floor(Convert.ToDouble(headSize.Width) / 2.0f)); }
            }

            /** returns bounds of positioner (position + dimension) **/
            public RectangleF Bounds
            {
                get
                {
                    return new RectangleF(Convert.ToSingle(currentPositionX - headSize.Width), 0,
                        headSize.Width * 2, headSize.Height);
                }
            }

            /** returns position (x value) **/
            public double Position
            {
                get
                {
                    return currentPositionX;
                }
                set
                {
                    if (value >= 0)
                        currentPositionX = value + HeadOffset;
                }
            }

            /**
             * constructor
             * \param headSize size of positioner
             */
            public Positioner(Size headSize)
            {
                setHeadSize(headSize);
                headPen = new Pen(Color.Black, 2);
                currentPositionX = HeadOffset;
                state = PositionerState.psInvisible;
            }

            /**
             * sets size of positioner
             * \param headSize new size
             */
            private void setHeadSize(Size headSize)
            {
                this.headSize.Width = HEAD_WIDTH;
                this.headSize.Height = headSize.Height;
            }

            /**
             * renders the positioner
             * \param g Graphics object to render to
             * \param markerHeight height of marker region
             */
            public void render(Graphics g, int markerHeight)
            {
                if (g != null && state != PositionerState.psInvisible)
                {
                    System.Windows.Forms.VisualStyles.TrackBarThumbState thumbState;

                    if (state == PositionerState.psNormal)
                    {
                        headPen.Color = Color.Black;
                        thumbState = System.Windows.Forms.VisualStyles.TrackBarThumbState.Normal;
                    }
                    else
                    {
                        headPen.Color = Color.Green;
                        thumbState = System.Windows.Forms.VisualStyles.TrackBarThumbState.Pressed;
                    }

                    Rectangle thumbRect = new Rectangle(Convert.ToInt32(Position - headSize.Width), 1,
                        headSize.Width * 2, headSize.Height + 1);
                    
                    Rectangle ticksRect = new Rectangle(Convert.ToInt32(Position - headSize.Width / 3 - 1),
                        headSize.Height / 3, headSize.Width * 2 / 3, headSize.Height / 3);

                    int thumbTicksNum = 3;

                    TrackBarRenderer.DrawHorizontalThumb(g, thumbRect, thumbState);
                    TrackBarRenderer.DrawHorizontalTicks(g, ticksRect, thumbTicksNum,
                         System.Windows.Forms.VisualStyles.EdgeStyle.Bump);

                    //g.DrawRectangle(headPen, Position - headSize.Width, 1, headSize.Width * 2, headSize.Height);
                    g.DrawLine(headPen, Convert.ToSingle(Position - 1), headSize.Height,
                        Convert.ToSingle(Position - 1), headSize.Height + markerHeight);
                }
            }
        }

        /**
         * \class ValueChangedArgs
         * EventArgs used if FramesTrackBar's Value is changed
         */
        public class ValueChangedArgs : EventArgs
        {
            private int _position;

            /**
             * constructor
             * \param position index/Value
             */
            public ValueChangedArgs(int position)
            {
                this.Position = position;
            }

            /** the index/Value **/
            public int Position
            {
                get
                {
                    return _position;
                }
                internal set
                {
                    _position = value;
                }
            }
        }

        /** called if FramesTrackBar's Value is changed **/
        public event EventHandler<ValueChangedArgs> OnValueChanged;

        private int _framesCount;
        /** number of frames **/
        public int Frames
        {
            get
            {
                return _framesCount;
            }
            set
            {
                if (value >= 0)
                {
                    _framesCount = value;
                    Value = 0;

                    if (positioner != null)
                    {
                        if (value == 0)
                            positioner.state = Positioner.PositionerState.psInvisible;
                        else
                            positioner.state = Positioner.PositionerState.psNormal;
                    }

                    updateTicksSpace();
                    removeObsoleteMarkers();
                    Invalidate();
                }
            }
        }
        private int _currentFrame;
        /** the current index **/
        public int Value
        {
            get
            {
                return _currentFrame;
            }
            set
            {
                if (value >= 0)
                {
                    _currentFrame = value;

                    if (positioner != null)
                        if (useTicks)
                            positioner.Position = (_currentFrame + 1) * ticksSpace;
                        else
                            try
                            {
                                positioner.Position = (_currentFrame + 1) * (Width - 2.0f * positioner.HeadOffset) / Frames;
                            }
                            catch (DivideByZeroException)
                            {
                                positioner.Position = 0;
                            }

                    if (OnValueChanged != null)
                        OnValueChanged(this, new ValueChangedArgs(_currentFrame));

                    Invalidate();
                }
            }
        }

        private double ticksSpace;
        private int headHeight;
        private Pen defaultPen;
        private Font font;
        private Positioner positioner;
        private int currentMarkerHeight;
        private ToolTip toolTip;
        private bool useTicks { get { return ticksSpace > 0.0f; } }
        
        public List<Marker> markers { get; internal set; }
        public bool dragging { get; internal set; }

        /**
         * constructor
         * \param frames initial number of frames
         * \param trackBarSize site of the track bar control
         */
        public FramesTrackBar(int frames, Size trackBarSize)
        {
            this.Location = new Point(10, 10);
            this.Size = trackBarSize;
            this.BackColor = Color.LightGray;
            this.DoubleBuffered = true;

            Frames = frames;

            defaultPen = new Pen(Color.Black, 1);
            font = new Font(FontFamily.GenericMonospace, 10);
            
            Value = 0;
            headHeight = 15;
            currentMarkerHeight = 0;
            dragging = false;
            updateTicksSpace();

            positioner = new Positioner(new Size(Convert.ToInt32(ticksSpace), headHeight));
            markers = new List<Marker>();
            toolTip = new ToolTip();

            
        }

        /**
         * returns a marker by its position
         * \param index marker position
         * \return the marker or null
         */
        public Marker getMarkerFromIndex(int index)
        {
            if (markers == null || index < 0 || index >= Frames)
                return null;

            foreach (Marker marker in markers)
                if (marker != null && marker.Index == index)
                    return marker;

            return null;
        }

        /**
         * adds a marker
         * \param index the marker position
         * \param note a note attached to the marker
         * \param markerType the type of the marker
         * \return ERR_NOERR on success
         */
        public int addMarker(int index, string note, Marker.MarkerType markerType)
        {
            if (index < 0 || index >= Frames)
                return ErrorCodes.ERR_INDEXBOUNDS;

            Marker tmpMarker = getMarkerFromIndex(index);
            if (tmpMarker != null)
                return ErrorCodes.ERR_ACCESSDENIED;

            // these types must occur unique
            if (markerType == Marker.MarkerType.mtLeft || markerType == Marker.MarkerType.mtRight)
                for (int i = markers.Count - 1; i >= 0; i--)
                {
                    Marker marker = markers[i];
                    if (marker.markerType == markerType)
                        markers.RemoveAt(i);
                }

            markers.Add(new Marker(index, note, markerType));
            return ErrorCodes.ERR_NOERR;
        }

        /**
         * removes a marker
         * \param index the marker position
         */
        public void removeMarker(int index)
        {
            Marker tmpMarker = getMarkerFromIndex(index);
            if (tmpMarker != null && markers != null)
                markers.Remove(tmpMarker);
        }

        /** drops all markers **/
        public void removeAllMarkers()
        {
            if (markers != null)
                markers.Clear();
        }

        /** removes all markers with invalid Index values **/
        private void removeObsoleteMarkers()
        {
            if (markers != null)
            {
                for (int i = markers.Count - 1; i >= 0; i--)
                {
                    Marker marker = markers[i];
                    if (marker == null || marker.Index < 0 || marker.Index >= Frames)
                        markers.RemoveAt(i);
                }
            }
        }

        /** 
         * allows editing a marker
         * \param index the marker position
         * \result true if note was changed, false otherwise
         */
        public bool editMarker(int index)
        {
            Marker tmpMarker = getMarkerFromIndex(index);
            if (tmpMarker != null)
            {
                InputDialog inputDialog = new InputDialog();
                inputDialog.initDialog("Note", "Enter your note", tmpMarker.Note);
                if (inputDialog.ShowDialog() == DialogResult.OK)
                {
                    tmpMarker.Note = inputDialog.getInput();
                    return true;
                }
                else
                    return false;
            }
            return false;
        }

        /** set ticksSpace due to control size **/
        private void updateTicksSpace()
        {
            if (positioner != null)
                ticksSpace = (this.Size.Width - 2 * positioner.HeadOffset) / (Frames + 1);
            else
                ticksSpace = (this.Size.Width) / (Frames + 1);

            if (positioner != null)
                if (useTicks)
                    positioner.Position = (Value + 1) * ticksSpace;
                else
                    positioner.Position = (Value + 1) * (Width - 2 * positioner.HeadOffset) / Frames;
        }

        /**
         * draws the ticks
         * \param g Graphics object to render to
         * 
         * \bug #0001 SOLVED adjust number of ticks for large Frames values
         */
        private void renderTicks(Graphics g)
        {
            if (g != null)
            {
                bool renderNumbers = ticksSpace > g.MeasureString(Frames.ToString(), font).Width / 2;

                if (useTicks)
                {
                    for (int i = 1; i <= Frames; i++)
                    {
                        int tickHeight = 2;

                        if (i % 2 == 0)
                        {
                            tickHeight *= 2;
                            if (renderNumbers)
                                g.DrawString(i.ToString(), font, Brushes.Black, positioner.HeadOffset + 
                                    Convert.ToSingle(i * ticksSpace - g.MeasureString(i.ToString(), font).Width / 2), 2);
                        }

                        g.DrawLine(defaultPen, positioner.HeadOffset + Convert.ToSingle(i * ticksSpace), 1,
                            positioner.HeadOffset + Convert.ToSingle(i * ticksSpace), tickHeight);
                    }
                }
                else
                {
                    g.DrawLine(new Pen(Color.Gray, 2.0f), positioner.HeadOffset, headHeight / 2,
                        Width - positioner.HeadOffset, headHeight / 2);
                }
            }
        }

        /**
         * draws markers
         * \param g Graphics object to render to
         * \param markerHeight height of markers region
         */
        private void renderMarkers(Graphics g, int markerHeight)
        {
            if (markers != null)
            {
                double markersTicksSpace = ticksSpace;
                if (!useTicks)
                    markersTicksSpace = (double)Width / (double)Frames;

                foreach (Marker marker in markers)
                    if (marker != null && marker.Index >= 0 && marker.Index < Frames)
                        marker.render(g, markersTicksSpace, markerHeight, headHeight, positioner.HeadOffset);
            }
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            currentMarkerHeight = e.ClipRectangle.Height - headHeight;

            e.Graphics.Clear(this.BackColor);
            e.Graphics.DrawRectangle(defaultPen, 0, 0, e.ClipRectangle.Width - 1, e.ClipRectangle.Height - 1);

            renderTicks(e.Graphics);

            e.Graphics.DrawLine(defaultPen, 0, headHeight + 1, e.ClipRectangle.Width, headHeight + 1);

            if (markers != null)
                renderMarkers(e.Graphics, e.ClipRectangle.Height - headHeight);

            if (positioner != null)
                positioner.render(e.Graphics, e.ClipRectangle.Height - headHeight);
        }

        protected override void OnMouseDown(MouseEventArgs e)
        {
            dragging = false;
            if (positioner != null)
            {
                if (positioner.Bounds.Contains(e.Location))
                {
                    positioner.state = Positioner.PositionerState.psPressed;
                    dragging = true;
                    Invalidate();
                }
                else
                    if (e.Location.Y <= headHeight)
                    {
                        int tmpFramePos = Value;
                        try
                        {
                            if (useTicks)
                                tmpFramePos = Convert.ToInt32(Math.Floor((e.Location.X - positioner.HeadOffset - ticksSpace / 2.0f) /
                                    ticksSpace));
                            else
                                tmpFramePos = Convert.ToInt32(Math.Floor(Convert.ToDouble((e.Location.X - positioner.HeadOffset) * Frames / 
                                    (Width - 2 * positioner.HeadOffset))));
                        }
                        catch (OverflowException)
                        {
                            ErrorCodes.error(ErrorCodes.ERR_OVERFLOW);
                        }

                        if (tmpFramePos < 0)
                            tmpFramePos = 0;
                        else
                            if (tmpFramePos >= Frames)
                                tmpFramePos = Frames - 1;

                        Value = tmpFramePos;
                    }
            }
        }

        protected override void OnMouseUp(MouseEventArgs e)
        {
            if (positioner != null)
                positioner.state = Positioner.PositionerState.psNormal;

            dragging = false;
            if (OnValueChanged != null)
                OnValueChanged(this, new ValueChangedArgs(_currentFrame));

            Invalidate();
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            if (positioner != null && positioner.state == Positioner.PositionerState.psPressed)
            {
                if (e.Location.X != positioner.Position)
                {
                    int tmpPos;
                    if (useTicks)
                        tmpPos = Convert.ToInt32(Math.Floor((e.Location.X - positioner.HeadOffset - ticksSpace / 2.0f) /
                            ticksSpace));
                    else
                        tmpPos = Convert.ToInt32(Math.Floor(Convert.ToDouble((e.Location.X - positioner.HeadOffset) * Frames / 
                            (Width - 2 * positioner.HeadOffset))));

                    if (tmpPos >= 0 && tmpPos <= Frames - 1 && tmpPos != Value)
                        Value = tmpPos;

                    /*if (Value < Frames - 1 &&
                        e.Location.X > positioner.Position + ticksSpace / 2.0f)
                        Value++;
                    else
                        if (Value > 0 &&
                            e.Location.X < positioner.Position - ticksSpace / 2.0f)
                            Value--;*/
                }
            }

            bool foundActiveNote = false;

            if (!dragging)
                foreach (Marker marker in markers)
                {
                    double markersTicksSpace = ticksSpace;
                    if (!useTicks)
                        markersTicksSpace = (double)Width / (double)Frames;

                    Rectangle markerBounds = marker.getBounds(markersTicksSpace,
                        currentMarkerHeight, headHeight);
                    Rectangle actualBounds = new Rectangle(markerBounds.X, markerBounds.Y,
                        markerBounds.Width - markerBounds.X, markerBounds.Height - markerBounds.Y);

                    if (actualBounds.Contains(e.Location))
                    {
                        foundActiveNote = true;
                        toolTip.Show((marker.Index + 1).ToString() +": " + marker.Note, this, e.X + 20, e.Y);
                    }
                }

            if (!foundActiveNote)
                toolTip.Hide(this);

            this.Invalidate();
        }

        protected override void OnResize(EventArgs e)
        {
            base.OnResize(e);

            updateTicksSpace();
            Invalidate();
        }

    }
}
