﻿/*
multiDAC - Movement and deformation analysis tool
Copyright (C) 2009 Felix Schmitt

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Reflection;
using System.IO;
#if !LINUX
using Microsoft.VisualBasic;
#else
using System.Security.Principal;
#endif
using Manager;
using InternalData;
using PluginInterfaceLibrary;

namespace multidac
{
    public partial class MainForm : Form
    {
        Renderer renderer;
        PluginManager pluginManager;
        DataProcessor dataProcessor;
        ProjectManager projectManager;
        ProcessingForm processingForm;
        ConfigManager configManager;

        List<Bitmap> views;

        private FramesTrackBar framesTrackBar;

        private const string VERSION = "0.1.9";
#if LINUX
        private string CONFIG_FILENAME = "/home/" + WindowsIdentity.GetCurrent().Name.Split('/')[1] + 
            "/.multidac.xml";
#else
        private string CONFIG_FILENAME = Microsoft.VisualBasic.FileIO.SpecialDirectories.CurrentUserApplicationData +
            "\\.multidac.xml";
#endif

        private bool draggingFrame = false;
        private Point draggingStart = new Point(0, 0);
        private int currentFrame = -1;
        private bool updatingGraphics;
        private Point contentMousePos = new Point(-1, -1);
        private bool _isPlaying = false;

        private bool IsPlaying
        {
            get { return _isPlaying; }
            set
            {
                _isPlaying = value;
                playFramesButton.Enabled = !_isPlaying;
                pauseFramesButton.Enabled = _isPlaying;
            }
        }

        private void progressCallback(int progress, string msg)
        {
            statusProgressBar.Value = progress;
            statusBarLabel.Text = msg.ToString();
            statusBar.Refresh();
        }

        private void processingCallback(int progress, string msg)
        {
            processingForm.setTotalProgress(progress, msg.ToString());
        }

        private void perFrameCallback(int progress, string msg)
        {
            processingForm.setRelativeProgress(progress);
        }

        private bool projectCreatedCallback(Project project)
        {
            this.Text = "multiDAC - " + project.name;
            vectorCountTrackBar.Value = 25;
            dataProcessor.vectorCount = vectorCountTrackBar.Value;
            vectorCountLabel.Text = "Vectors: " + (vectorCountTrackBar.Value * vectorCountTrackBar.Value).ToString();
            framesTrackBar.Frames = 0;
            framesTrackBar.Value = 0;

            progressCallback(0, "Done");
            return false;
        }

        private bool projectLoadedCallback(Project project)
        {
            bool projectModified = false;

            this.Text = "multiDAC - " + project.name;

            if (project.initialVectors >= vectorCountTrackBar.Minimum &&
                project.initialVectors <= vectorCountTrackBar.Maximum)
                vectorCountTrackBar.Value = project.initialVectors;
            else
                vectorCountTrackBar.Value = 25;

            dataProcessor.vectorCount = vectorCountTrackBar.Value;
            vectorCountLabel.Text = "Vectors: " + (vectorCountTrackBar.Value * vectorCountTrackBar.Value).ToString();
            if (pluginManager.setActivePlugin(project.importPlugin, PluginManager.PluginType.ptImport) != ErrorCodes.ERR_NOERR &&
                project.sourceFiles.Count > 0)
            {
                if (MessageBox.Show("Couldn't find the project's import plugin.\nDo you want to choose a plugin" +
                    " from the list?", "Error", MessageBoxButtons.YesNo,
                    MessageBoxIcon.Error) == DialogResult.Yes)
                {
                    PluginsListForm listForm = new PluginsListForm();
                    string[] plugins = new string[pluginManager.importPluginCount];
                    for (int i = 0; i < pluginManager.importPluginCount; i++)
                    {
                        IPlugin importPlugin = pluginManager.getPlugin(i, PluginManager.PluginType.ptImport);
                        plugins[i] = importPlugin.name + ": " + importPlugin.description;
                    }
                    listForm.fillPluginsList(plugins);

                    if (listForm.ShowDialog() == DialogResult.OK)
                    {
                        pluginManager.setActivePlugin(pluginManager.getIdentifier(listForm.selectedPlugin,
                            PluginManager.PluginType.ptImport), PluginManager.PluginType.ptImport);
                        project.importPlugin = pluginManager.getIdentifier(listForm.selectedPlugin,
                            PluginManager.PluginType.ptImport);
                        projectModified = true;
                    }
                }
            }

            if (pluginManager.activeImportPlugin != null)
            {
                if (!project.includeSourceFiles && pluginManager.activeImportPlugin.loadFiles(project.sourceFiles.ToArray(),
                    progressCallback) != ErrorCodes.ERR_NOERR)
                    MessageBox.Show("Failed to load all files.", "Error", MessageBoxButtons.OK,
                        MessageBoxIcon.Error);

                if (project.includeSourceFiles)
                {
                    List<string> tmpSourceFiles = new List<string>();
                    /**
                     * \todo to be changed
                     */
                    foreach (string sourceFile in project.sourceFiles)
                        tmpSourceFiles.Add(projectManager.cachedDirectories[0] + Path.GetFileName(sourceFile));

                    if (pluginManager.activeImportPlugin.loadFiles(tmpSourceFiles.ToArray(),
                        progressCallback) != ErrorCodes.ERR_NOERR)
                        MessageBox.Show("Failed to load all files.", "Error", MessageBoxButtons.OK,
                        MessageBoxIcon.Error);
                }
            }

            if (pluginManager.activeImportPlugin != null && pluginManager.activeImportPlugin.frameCount > 0)
                framesTrackBar.Frames = pluginManager.activeImportPlugin.frameCount;
            else
                framesTrackBar.Frames = 0;

            framesTrackBar.Value = 0;

            progressCallback(0, "Done");
            return projectModified;
        }

        private bool projectSavingCallback(Project project)
        {
            if (pluginManager.activeImportPlugin != null)
                project.setSourceFiles(pluginManager.activeImportPlugin.importedFilenames);

            project.initialVectors = dataProcessor.vectorCount;
            string identifier = pluginManager.getIdentifier(pluginManager.importPluginIndex,
                PluginManager.PluginType.ptImport);
            if (identifier != null)
                project.importPlugin = identifier;
            else
                project.importPlugin = "NULL";

            project.processingRuns = new ProjectDataRun[dataProcessor.dataRuns.Count];
            for (int i = 0; i < dataProcessor.dataRuns.Count; i++)
            {
                project.processingRuns[i] = new ProjectDataRun();
                project.processingRuns[i].setValues(dataProcessor.getRun(i));
            }

            return false;
        }

        private void configLoadedCallback(Configuration configuration)
        {
            if (configuration != null)
            {
                //pluginManager.pathToPlugins = configuration.pathToPlugins;
                renderer.lineWidth = configuration.renderingLineWidth;
                renderer.vectorsInSplitScreen = configuration.vectorsInSplitScreen;
                projectManager.includeSrcFilesByDefault = configuration.includeSourceFilesByDefault;
            }
        }

        private void configSavingCallback(Configuration configuration)
        {
            configuration.pathToPlugins = pluginManager.pathToPlugins;
        }

        private void timeRemainingCallback(DateTime time)
        {
            processingForm.setTimeLabel(time);
        }

        private void modify()
        {
            projectManager.modify();
            modifiedLabel.Text = "modified";
        }

        private void refreshProjectLayout()
        {
            if (!projectManager.projectLoaded)
            {
                toolStripContainer.ContentPanel.Hide();
                this.Text = "multiDAC";
                progressCallback(0, "Done");
                modifiedLabel.Text = "";
            }
            else
            {
                toolStripContainer.ContentPanel.Show();
                if (projectManager.location != "")
                    this.Text = "multiDAC - " + Path.GetFileName(projectManager.location);
                else
                    this.Text = "multiDAC - " + projectManager.project.name;
                refreshPluginsGroup();
                refreshFrameInfo();
                refreshDataRunsList();
                refreshDataViewTree();
                refreshRendering(false);

                if (projectManager.modified)
                    modifiedLabel.Text = "modified";
                else
                    modifiedLabel.Text = "";
            }

            saveToolStripMenuItem.Enabled = projectManager.projectLoaded;
            saveAsToolStripMenuItem.Enabled = projectManager.projectLoaded;
            saveProjectButton.Enabled = projectManager.projectLoaded;

            importFilesButton.Enabled = projectManager.projectLoaded && pluginManager.activeImportPlugin != null;
            importToolStripMenuItem.Enabled = projectManager.projectLoaded && pluginManager.activeImportPlugin != null;

            captureMediaButton.Enabled = false; //projectManager.projectLoaded;
            captureMediaToolStripMenuItem.Enabled = false; //projectManager.projectLoaded;

            processDataButton.Enabled = projectManager.projectLoaded && pluginManager.activeProcessingPlugin != null;
            processingToolStripMenuItem.Enabled = projectManager.projectLoaded && pluginManager.activeProcessingPlugin != null;

            viewToolStripMenuItem.Enabled = projectManager.projectLoaded;
            closeToolStripMenuItem.Enabled = projectManager.projectLoaded;
            propertiesToolStripMenuItem.Enabled = projectManager.projectLoaded;
            exportToolStripMenuItem.Enabled = projectManager.projectLoaded;

            playFramesButton.Enabled = !IsPlaying;
            pauseFramesButton.Enabled = IsPlaying;

            refreshFrameInfo();
            this.Refresh();
        }

        private void setGUIEnabled(bool guiEnabled)
        {
            pausePlayingFrames();
            toolStripContainer.ContentPanel.Enabled = guiEnabled;
            mainMenu.Enabled = guiEnabled;
            mainToolStrip.Enabled = guiEnabled;
        }

        private PluginReloadAction reloadPlugins(bool initialReload)
        {
            if (!Directory.Exists(pluginManager.pathToPlugins))
                return PluginReloadAction.praFailed;

            if (!initialReload && projectManager.projectLoaded && pluginManager.totalPluginCount > 0)
            {
                if (MessageBox.Show("Reloading plugins.\nThis will drop existing frames and data. Continue?",
                    "Warning", MessageBoxButtons.YesNo, MessageBoxIcon.Warning) == DialogResult.No)
                    return PluginReloadAction.praCancel;
            }

            setGUIEnabled(false);

            
            int result = pluginManager.loadPlugins(progressCallback);

            if (!initialReload && projectManager.projectLoaded)
                dataProcessor.clearData();

            if (framesTrackBar != null)
            {
                framesTrackBar.Frames = 0;
                currentFrame = 0;
            }

            renderer.reset();
            if (!fitWindowToolStripMenuItem.Checked)
                fitWindowToolStripMenuItem_Click(this, null);
            refreshPluginsGroup();
            refreshFrameInfo();
            refreshDataRunsList();
            refreshDataViewTree();
            refreshRendering(true);

            if (!initialReload && projectManager.projectLoaded)
                modify();

            refreshProjectLayout();

            setGUIEnabled(true);

            if (result == ErrorCodes.ERR_NOERR && pluginManager.totalPluginCount > 0)
                return PluginReloadAction.praLoaded;
            else
                return PluginReloadAction.praFailed;
        }

        private void refreshRendering(bool fullUpdate)
        {
            if (renderer != null && fullUpdate)
                renderer.update();

            renderingBox.Invalidate();
        }

        private void refreshFrameInfo()
        {
            frameIndexLabel.Text = "No frames";

            if (pluginManager.activeImportPlugin != null && pluginManager.activeImportPlugin.frameCount > 0)
            {
                frameIndexLabel.Text = "Frame " + (currentFrame + 1).ToString() + "/" +
                    pluginManager.activeImportPlugin.frameCount;

                markersButton.Enabled = true;
            }
            else
                markersButton.Enabled = false;

            FramesTrackBar.Marker marker = framesTrackBar.getMarkerFromIndex(currentFrame);
            jumpStartMarkerToolStripMenuItem.Enabled = dataProcessor.currentRun != null;
            jumpEndMarkerToolStripMenuItem.Enabled = dataProcessor.currentRun != null;
            addStartMarkerButton.Enabled = marker == null;
            addEndMarkerButton.Enabled = marker == null;
            removeMarkerButton.Enabled = marker != null;

            notesToolStripMenuItem.Enabled = (dataProcessor != null) && (dataProcessor.currentRun != null);
            addNoteMarkerButton.Enabled = marker == null;
            editNoteMarkerButton.Enabled = (marker != null) &&
                (marker.markerType == FramesTrackBar.Marker.MarkerType.mtNote);

            framesTrackBar.Invalidate();

            fitWindowToolStripMenuItem.Checked = renderer.fitWindow;
            setFrameInfoLabel(new Point(-1, -1));
        }

        private void refreshPluginsGroup()
        {
            processingComboBox.Items.Clear();
            processDataWithToolStripMenuItem.DropDownItems.Clear();

            for (int i = 0; i < pluginManager.processingPluginCount; i++)
            {
                IPlugin plugin = pluginManager.getPlugin(i, PluginManager.PluginType.ptProcessing);
                processingComboBox.Items.Add(plugin.name);
                ToolStripMenuItem newItem = new ToolStripMenuItem(plugin.name);
                newItem.Name = plugin.name;
                newItem.Tag = i;
                newItem.Click += processDataToolStripMenuItem_Click;
                processDataWithToolStripMenuItem.DropDownItems.Add(newItem);
            }

            if (processingComboBox.Items.Count > pluginManager.processingPluginIndex)
                processingComboBox.SelectedIndex = pluginManager.processingPluginIndex;

            importToolStripMenuItem.DropDownItems.Clear();
            importFilesButton.DropDownItems.Clear();

            for (int i = 0; i < pluginManager.importPluginCount; i++)
            {
                IPlugin plugin = pluginManager.getPlugin(i, PluginManager.PluginType.ptImport);
                ToolStripMenuItem newItem1 = new ToolStripMenuItem(plugin.name);
                ToolStripMenuItem newItem2 = new ToolStripMenuItem(plugin.name);

                newItem1.Name = plugin.name;
                newItem2.Name = plugin.name;
                newItem1.Tag = i;
                newItem2.Tag = i;
                newItem1.Click += importFiles;
                newItem2.Click += importFiles;
                importToolStripMenuItem.DropDownItems.Add(newItem1);
                importFilesButton.DropDownItems.Add(newItem2);
            }

            exportPluginsToolStripMenuItem.DropDownItems.Clear();
            //exportPluginsButton.DropDownItems.Clear();

            for (int i = 0; i < pluginManager.exportPluginCount; i++)
            {
                IPlugin plugin = pluginManager.getPlugin(i, PluginManager.PluginType.ptExport);
                ToolStripMenuItem newItem1 = new ToolStripMenuItem(plugin.name);
                //ToolStripMenuItem newItem2 = new ToolStripMenuItem(plugin.name);

                newItem1.Name = plugin.name;
                //newItem2.Name = plugin.name;
                newItem1.Tag = i;
                //newItem2.Tag = i;
                newItem1.Click += exportResults;
                //newItem2.Click += importFiles;
                exportPluginsToolStripMenuItem.DropDownItems.Add(newItem1);
                //importFilesButton.DropDownItems.Add(newItem2);
            }
        }

        private void refreshDataRunsList()
        {
            renderer.disable();

            int selected = 0;
            if (dataRunsGrid.SelectedRows.Count > 0)
                selected = dataRunsGrid.SelectedRows[0].Index;

            dataRunsGrid.Rows.Clear();
            for (int i = 0; i < dataProcessor.dataRuns.Count; i++)
            {
                DataRun run = dataProcessor.getRun(i);
                dataRunsGrid.Rows.Add((i + 1).ToString(), run.duration.TimeOfDay.TotalSeconds.ToString() + " s",
                    run.initialVectors.ToString(), (run.firstFrame + 1).ToString(),
                    (run.lastFrame + 1).ToString());
            }

            if (dataRunsGrid.Rows.Count > selected)
                dataRunsGrid.Rows[selected].Selected = true;

            deleteSelectedToolStripMenuItem.Enabled = dataRunsGrid.SelectedRows.Count > 0;
            deleteAllToolStripMenuItem1.Enabled = dataRunsGrid.Rows.Count > 0;

            renderer.enable();
        }

        private void refreshDataViewTree()
        {
            dataViewTree.Nodes.Clear();
            fillDataTree();
        }

        private void setPlugins()
        {
            if (processingComboBox.SelectedIndex != -1 && processingComboBox.SelectedIndex < pluginManager.processingPluginCount)
                pluginManager.processingPluginIndex = processingComboBox.SelectedIndex;
        }

        private void nextFrame()
        {
            if (pluginManager.activeImportPlugin != null)
            {
                if (currentFrame < pluginManager.activeImportPlugin.frameCount - 1)
                {
                    framesTrackBar.Value = currentFrame + 1;
                }
            }
        }

        private void prevFrame()
        {
            if (pluginManager.activeImportPlugin != null)
            {
                if (currentFrame > 0)
                {
                    framesTrackBar.Value = currentFrame - 1;
                }
            }
        }

        private void firstFrame()
        {
            if (pluginManager.activeImportPlugin != null)
            {
                framesTrackBar.Value = 0;
            }
        }

        private void lastFrame()
        {
            if (pluginManager.activeImportPlugin != null)
            {
                framesTrackBar.Value = pluginManager.activeImportPlugin.frameCount - 1;
            }
        }

        private void playFrames()
        {
            if (!IsPlaying && framesTrackBar.Frames > 0)
            {
                if (framesTrackBar.Value == framesTrackBar.Frames - 1)
                    framesTrackBar.Value = 0;

                IsPlaying = true;
                while (framesTrackBar.Value < framesTrackBar.Frames - 1 && IsPlaying)
                {
                    DateTime start = DateTime.Now;
                    while (IsPlaying && (TimeSpan.FromMilliseconds(100.0f).CompareTo(DateTime.Now - start) > 0 ||
                        updatingGraphics))
                        Application.DoEvents();

                    framesTrackBar.Value++;
                }

                IsPlaying = false;
            }
        }

        private void pausePlayingFrames()
        {
            if (IsPlaying)
            {
                IsPlaying = false;
            }
        }

        private void jumpStartMarker()
        {
            if (pluginManager.activeImportPlugin != null && dataProcessor.currentRun != null)
                framesTrackBar.Value = dataProcessor.currentRun.firstFrame;
        }

        private void jumpEndMarker()
        {
            if (pluginManager.activeImportPlugin != null && dataProcessor.currentRun != null)
                framesTrackBar.Value = dataProcessor.currentRun.lastFrame;
        }

        private void computeFrameRange(DataProcessor.FrameRange range)
        {
            if (range == null)
                return;

            if (framesTrackBar.markers.Count == 0)
            {
                range.start = 0;
                range.end = pluginManager.activeImportPlugin.frameCount - 1;
                return;
            }
            else
            {
                int lastStartMarkerIndex = 0;
                int lastEndMarkerIndex = pluginManager.activeImportPlugin.frameCount - 1;

                foreach (FramesTrackBar.Marker marker in framesTrackBar.markers)
                {
                    if (marker.markerType == FramesTrackBar.Marker.MarkerType.mtRight)
                        lastStartMarkerIndex = marker.Index;

                    if (marker.markerType == FramesTrackBar.Marker.MarkerType.mtLeft)
                        lastEndMarkerIndex = marker.Index;
                }

                range.start = lastStartMarkerIndex;
                range.end = lastEndMarkerIndex;
            }

        }

        private void processMedia(int procPluginIndex)
        {
            processingForm = new ProcessingForm(dataProcessor.cancelProcessing, 
                configManager.configuration.askBeforeCancellingProcessing);
            setGUIEnabled(false);

            processingForm.Show();
            processingForm.Focus();

            DataProcessor.FrameRange range = new DataProcessor.FrameRange(0, 0);

            computeFrameRange(range);

            if (dataProcessor.analyzeMedia(range, pluginManager, configManager.configuration,
                procPluginIndex, processingCallback, perFrameCallback, timeRemainingCallback) != 0)
                MessageBox.Show("Could not analyze media.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            else
            {
                refreshFrameInfo();
                firstFrame();
            }

            refreshDataRunsList();
            refreshRendering(true);
            if (dataRunsGrid.Rows.Count > 0)
                dataRunsGrid.Rows[dataRunsGrid.Rows.Count - 1].Selected = true;

            processingForm.Close();
            if (dataProcessor.currentRun != null)
                framesTrackBar.Value = dataProcessor.currentRun.firstFrame;
            setGUIEnabled(true);
            modify();
        }

        /**
         * \bug #0002 if package is null, something else has to be done
         */
        private void fillDataTree()
        {
            dataViewTree.Nodes.Clear();
            if (dataProcessor.currentRun != null && dataProcessor.currentRun.getPackageCount() > 0)
                for (int i = 0; i < dataProcessor.currentRun.getPackageCount(); i++)
                {
                    DataPackage package = dataProcessor.currentRun.getPackage(i);
                    if (package != null)
                        dataViewTree.Nodes.Add("Processing frames " + (package.frame1 + 1).ToString() + " and " +
                            (package.frame2 + 1).ToString() + " [double click]");
                    else
                        dataViewTree.Nodes.Add("Processing frames " + (i + 1).ToString() + " and " +
                            (i + 2).ToString() + " [double click]");
                }
        }

        private void fillDataTreeSub(TreeNode parent)
        {
            parent.Nodes.Clear();

            if (dataProcessor.currentRun == null)
                return;

            parent.Text = parent.Text.Substring(0, parent.Text.Length - " [double click]".Length);

            DataPackage package = dataProcessor.currentRun.getPackage(parent.Index);
            if (package != null && package.vectors.Length > 0)
                for (int j = 0; j < package.vectors.Length; j++)
                {
                    Vector v = package.vectors[j];
                    if (v.valid)
                        parent.Nodes.Add("(" + v.start.X.ToString() + "; " + v.start.Y.ToString() + ") -> (" +
                            v.target.X.ToString() + "; " + v.target.Y.ToString() + ")");
                }
        }

        private void newProject()
        {
            setGUIEnabled(false);
            if (closeProject(false))
            {
                projectManager.create("untitled", projectCreatedCallback);
                refreshProjectLayout();
            }
            setGUIEnabled(true);
        }

        private bool closeProject(bool force)
        {
            pausePlayingFrames();

            if (!force && projectManager.projectLoaded && projectManager.modified)
            {

                DialogResult result = DialogResult.No;

                if (configManager.configuration.askBeforeClosingModified)
                    result = MessageBox.Show("Current project was modified. Save first?",
                        "Modified", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question);

                if (result == DialogResult.Cancel)
                    return false;

                if (result == DialogResult.Yes)
                    saveProject(true);
            }

            if (pluginManager.activeImportPlugin != null)
                pluginManager.activeImportPlugin.clearFiles();

            projectManager.close();

            dataRunsGrid.Rows.Clear();
            dataProcessor.clearData();
            clearViews();
            renderer.reset();
            framesTrackBar.Frames = 0;

            refreshFrameInfo();
            refreshProjectLayout();
            return true;
        }

        private void loadProject(bool force)
        {
            setGUIEnabled(false);
            if (openProjectDialog.ShowDialog() == DialogResult.OK)
            {
                closeProject(force);
                if (projectManager.load(openProjectDialog.FileName, projectLoadedCallback, progressCallback) != 0)
                    progressCallback(0, "Failed to open project file");
                refreshProjectLayout();
            }
            setGUIEnabled(true);
        }

        private void saveProject(bool saveImmediately)
        {
            if (projectManager.projectLoaded)
            {
                setGUIEnabled(false);
                if (saveImmediately && projectManager.location != "")
                {
                    if (projectManager.save(projectManager.location, projectSavingCallback,
                        progressCallback) != 0)
                        progressCallback(0, "Failed to save project file");
                }
                else
                {
                    if (projectManager.location == "")
                    {
                        saveProjectDialog.FileName = projectManager.project.name +
                            ProjectManager.PROJ_ARCHIVE_EXT;
                        saveProjectDialog.FilterIndex = 0;
                    }
                    else
                        saveProjectDialog.FileName = projectManager.location;

                    if (saveProjectDialog.ShowDialog() == DialogResult.OK)
                        if (projectManager.save(saveProjectDialog.FileName, projectSavingCallback,
                            progressCallback) != 0)
                            progressCallback(0, "Failed to save project file");
                }
                refreshProjectLayout();
                setGUIEnabled(true);
            }
        }

        private void showAboutDialog()
        {
            AboutForm aboutForm = new AboutForm();
            aboutForm.setVersionLabel(VERSION);
            aboutForm.ShowDialog();
        }

        private void deleteRun(int index, bool all)
        {
            if (all && dataProcessor.getRunsCount() > 0)
            {
                if (MessageBox.Show("Delete all existing data?", "Warning", MessageBoxButtons.YesNo,
                    MessageBoxIcon.Warning) == DialogResult.No)
                    return;

                dataProcessor.clearData();
                modify();
                refreshProjectLayout();
                return;
            }

            if (dataProcessor.getRunsCount() > index)
            {
                dataProcessor.deleteRun(index);
                modify();

                if (dataRunsGrid.Rows.Count > 0)
                    dataRunsGrid.Rows[0].Selected = true;
                refreshProjectLayout();
            }
        }

        private void loadConfiguration(string filename)
        {
            if (configManager.load(filename, configLoadedCallback, progressCallback) != ErrorCodes.ERR_NOERR)
                progressCallback(0, "Failed to load configuration");
        }

        private void saveConfiguration(string filename)
        {
            if (configManager.save(filename, configSavingCallback, progressCallback) != ErrorCodes.ERR_NOERR)
                progressCallback(0, "Failed to save configuration");
        }

        private void importFiles(object sender, EventArgs e)
        {
            if (pluginManager.activeImportPlugin != null &&
                pluginManager.activeImportPlugin.importedFiles.Length > 0)
            {
                if (MessageBox.Show("Importing new files will drop your existing frames and data.\n" +
                        "Continue anyway?",
                        "Warning", MessageBoxButtons.YesNo, MessageBoxIcon.Warning) == DialogResult.No)
                    return;
            }

            // user clicked a dynamically created menuitem, its tag being the requested import plugin index
            if ((sender as ToolStripMenuItem).Tag != null)
                pluginManager.importPluginIndex = (int)(sender as ToolStripMenuItem).Tag;

            // we came here any other way than clicking such a menuitem
            if (pluginManager.activeImportPlugin != null)
            {
                setGUIEnabled(false);
                openMediaDialog.Multiselect = pluginManager.activeImportPlugin.supportsMultipleFiles;
                if (openMediaDialog.ShowDialog() == DialogResult.OK)
                {
                    pluginManager.activeImportPlugin.clearFiles();
                    dataProcessor.clearData();
                    framesTrackBar.Frames = 0;
                    modify();

                    string[] files = openMediaDialog.FileNames;
                    if (pluginManager.activeImportPlugin.loadFiles(files, progressCallback) != ErrorCodes.ERR_NOERR)
                        MessageBox.Show("Failed to load all files.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);

                    if (pluginManager.activeImportPlugin.frameCount > 0)
                        framesTrackBar.Frames = pluginManager.activeImportPlugin.frameCount;
                    else
                        framesTrackBar.Frames = 0;


                    framesTrackBar.Value = 0;
                    renderer.reset();
                    if (!fitWindowToolStripMenuItem.Checked)
                        fitWindowToolStripMenuItem_Click(this, null);
                    refreshFrameInfo();
                    refreshDataRunsList();
                    refreshDataViewTree();
                    refreshRendering(true);
                }
                setGUIEnabled(true);
            }
            else
                MessageBox.Show("No import plugin available. Load plugins first.", "Error", MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
        }

        private void exportResults(object sender, EventArgs e)
        {
            // user clicked a dynamically created menuitem, its tag being the requested import plugin index
            if ((sender as ToolStripMenuItem).Tag != null)
                pluginManager.exportPluginIndex = (int)(sender as ToolStripMenuItem).Tag;

            // we came here any other way than clicking such a menuitem
            if (pluginManager.activeExportPlugin != null)
            {
                if (pluginManager.activeExportPlugin.requiresConfig)
                    pluginManager.activeExportPlugin.configure();

                pluginManager.activeExportPlugin.exportData(dataProcessor.dataRuns.ToArray(),
                    pluginManager.activeImportPlugin.frameCount, getFrameCallback, renderViewCallback);
            }
        }

        private void viewBox_OnPaint(object sender, PaintEventArgs e)
        {
            PictureBox viewBox = (sender as PictureBox);
            TabPage thisPage = viewBox.Parent as TabPage;

            int viewID = contentTabControl.TabPages.IndexOf(thisPage) - 1;

            if (views.Count > viewID)
                e.Graphics.DrawImage(views[viewID] as Bitmap, new Point(0, 0));

            // MessageBox.Show("viewID = " + viewID.ToString());
        }

        private void removeView(int viewID)
        {
            if (views.Count > viewID)
                try
                {
                    views.RemoveAt(viewID);
                }
                catch (ArgumentOutOfRangeException)
                {
                    ErrorCodes.error(ErrorCodes.ERR_INDEXBOUNDS);
                }

            if (contentTabControl.TabCount - 1 > viewID)
                contentTabControl.TabPages.RemoveAt(viewID + 1);
        }

        private void clearViews()
        {
            if (contentTabControl.TabCount > 1)
                for (int i = contentTabControl.TabCount - 1; i > 0; i--)
                    removeView(i - 1);
        }

        private void closeViewButton_OnClick(object sender, EventArgs e)
        {
            Button closeViewButton = (sender as Button);
            TabPage thisPage = closeViewButton.Parent.Parent as TabPage;
            int viewID = contentTabControl.TabPages.IndexOf(thisPage) - 1;

            removeView(viewID);
        }

        private void addContentTabPage(string label)
        {
            TabPage newPage = new TabPage(label);

            /*Panel viewBoxPanel = new Panel();
            viewBoxPanel.Height = 30;
            viewBoxPanel.Dock = DockStyle.Bottom;

            viewBoxPanel.Controls.Add(closeViewButton);
            newPage.Controls.Add(viewBoxPanel);*/

            PictureBox viewBox = new PictureBox();
            viewBox.Dock = DockStyle.Fill;
            viewBox.Image = views[contentTabControl.TabPages.Count - 1];
            viewBox.SizeMode = PictureBoxSizeMode.Zoom;
            //viewBox.Paint += viewBox_OnPaint;

            Button closeViewButton = new Button();
            closeViewButton.Text = "";
            closeViewButton.Image = multidac.Properties.Resources.process_stop;
            closeViewButton.Size = new Size(26, 26);
            closeViewButton.Location = new Point(5, 5);
            closeViewButton.FlatStyle = FlatStyle.Flat;
            closeViewButton.FlatAppearance.BorderSize = 0;
            closeViewButton.Click += closeViewButton_OnClick;

            viewBox.Controls.Add(closeViewButton);

            newPage.Controls.Add(viewBox);

            try
            {
                contentTabControl.TabPages.Add(newPage);
            }
            catch (ArgumentNullException)
            {
                ErrorCodes.error(ErrorCodes.ERR_NULL);
            }

            try
            {
                contentTabControl.SelectedIndex = contentTabControl.TabPages.Count - 1;
            }
            catch (ArgumentOutOfRangeException)
            {
                ErrorCodes.error(ErrorCodes.ERR_INDEXBOUNDS);
            }
        }

        private int renderViewCallback(Bitmap view)
        {
            if (view == null)
                return ErrorCodes.ERR_NULL;

            views.Add(view);
            addContentTabPage("View " + views.Count.ToString());
            return ErrorCodes.ERR_NOERR;
        }

        private Bitmap getFrameCallback(int frame)
        {
            if (pluginManager.activeImportPlugin != null && pluginManager.activeImportPlugin.frameCount > frame)
            {
                Size frameDimension = new Size(0, 0);
                if (pluginManager.activeImportPlugin.getDimension(frame, ref frameDimension) != ErrorCodes.ERR_NOERR)
                    return null;

                Bitmap buffer = new Bitmap(frameDimension.Width, frameDimension.Height);
                Graphics bufferGraphics = Graphics.FromImage(buffer);

                if (pluginManager.activeImportPlugin.renderFrame(frame, bufferGraphics) != ErrorCodes.ERR_NOERR)
                    return null;

                return buffer;
            }
            else
                return null;
        }

        private void initProgram()
        {
            renderer = new Renderer();
            pluginManager = new PluginManager();
            dataProcessor = new DataProcessor();
            projectManager = new ProjectManager(dataProcessor, pluginManager);
            views = new List<Bitmap>();

            configManager = new ConfigManager();
            loadConfiguration(CONFIG_FILENAME);

            /*if (!Directory.Exists(pluginManager.pathToPlugins))
            {
                MessageBox.Show("The plugin directory could not be determined. Please select it" +
                    " manually. (e.g. C:/Program Files/multidac/plugins).", "Plugin directory",
                    MessageBoxButtons.OK, MessageBoxIcon.Information);
                pluginManager.configure(reloadPlugins, progressCallback);
            }
            else*/

            if (reloadPlugins(true) != PluginReloadAction.praLoaded)
                statusBarLabel.Text = "Could not load any plugins.";

            refreshProjectLayout();
        }

        private void configureProgram(int startTab)
        {
            if (configManager.showSettingsForm(startTab, configLoadedCallback) == DialogResult.OK)
                refreshProjectLayout();
        }

        private void framesTrackBar_OnValueChangedHandler(Object sender, FramesTrackBar.ValueChangedArgs e)
        {
            currentFrame = framesTrackBar.Value;

            refreshFrameInfo();
            refreshRendering(true);
        }

        private void initFileDialogs()
        {
            string projectDialogsFilter = "multiDAC Project Archive (*" + ProjectManager.PROJ_ARCHIVE_EXT +
                ")|*" + ProjectManager.PROJ_ARCHIVE_EXT + "|multiDAC Project File (*" +
                ProjectManager.PROJ_XML_EXT + ")|*" + ProjectManager.PROJ_XML_EXT;

            openProjectDialog.Filter = projectDialogsFilter;
            openProjectDialog.FilterIndex = 0;
            saveProjectDialog.Filter = projectDialogsFilter;
            saveProjectDialog.FilterIndex = 0;
        }

        private void openURL(string url)
        {
            try
            {
                System.Diagnostics.Process.Start(url);
            }
            catch (Exception)
            {
                ErrorCodes.error(ErrorCodes.ERR_UNKNOWN);
            }
        }

        private void openWebpage()
        {
            openURL("http://multidac.sourceforge.net");
        }

        private void reportBug()
        {
            openURL("http://sourceforge.net/apps/mantisbt/multidac/my_view_page.php");
        }

        private void zoomIn()
        {
            if (renderer != null)
            {
                renderer.Zoom = Convert.ToInt32(renderer.Zoom * 1.5f);
                refreshRendering(true);
                refreshFrameInfo();
            }
        }

        private void zoomOut()
        {
            if (renderer != null)
            {
                renderer.Zoom = Convert.ToInt32(renderer.Zoom * 0.5f);
                refreshRendering(true);
                refreshFrameInfo();
            }
        }

        private void zoomFitWindow()
        {
            if (renderer != null)
            {
                renderer.zoomFitWindow();
                refreshRendering(true);
                refreshFrameInfo();
            }
        }

        private void adjustZoom(int zoomLevel)
        {
            if (renderer != null)
            {
                renderer.Zoom = zoomLevel;
                refreshRendering(true);
                refreshFrameInfo();
            }
        }

        private void setFrameInfoLabel(Point mousePos)
        {
            if (!IsPlaying)
            {
                contentMousePos = mousePos;
                if (mousePos.X == -1 || mousePos.Y == -1)
                    mousePos = new Point(0, 0);

                paintBoxInfoLabel.Text = renderer.dimension.Width.ToString() + "x" +
                    renderer.dimension.Height.ToString() + " with " + renderer.Zoom.ToString() + "%" +
                    " ( " + mousePos.X.ToString() + " ; " + mousePos.Y.ToString() + " )";

                statusBar.Update();

                if (splitscreenToolStripMenuItem.Checked)
                    refreshRendering(false);
            }
        }

        /** constructor **/
        public MainForm()
        {
            InitializeComponent();

            framesTrackBar = new FramesTrackBar(1, frameBarContainer.ContentPanel.Size);
            framesTrackBar.Dock = DockStyle.Fill;
            framesTrackBar.OnValueChanged += framesTrackBar_OnValueChangedHandler;
            frameBarContainer.ContentPanel.Controls.Add(framesTrackBar);

            initProgram();
            initFileDialogs();
        }

        #region Eventhandlers

        #region MainForm handlers

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            // check if project should be saved before closing
            if (!closeProject(false))
                e.Cancel = true;
            else
                saveConfiguration(CONFIG_FILENAME);
        }

        private void MainForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            // force running processing operations to stop
            if (dataProcessor != null)
                dataProcessor.cancelProcessing(true);
        }

        #endregion

        #region renderingBox handlers

        private void renderingBox_Paint(object sender, PaintEventArgs e)
        {
            updatingGraphics = true;

            e.Graphics.FillRectangle(Brushes.DarkSlateGray, e.ClipRectangle);
            if (renderer != null)
                renderer.render(e.Graphics, e.ClipRectangle, pluginManager.activeImportPlugin,
                    currentFrame, dataProcessor.getPackage(dataProcessor.currentRunIndex,
                    currentFrame), contentMousePos, showVectorsToolStripMenuItem.Checked,
                    movingVectorsToolStripMenuItem.Checked, splitscreenToolStripMenuItem.Checked);

            updatingGraphics = false;
        }

        private void renderingBox_MouseDown(object sender, MouseEventArgs e)
        {
            if (renderer != null && renderer.draggable)
            {
                draggingFrame = true;
                draggingStart = e.Location;
            }
        }

        private void renderingBox_MouseUp(object sender, MouseEventArgs e)
        {
            if (renderer != null && draggingFrame)
            {
                draggingFrame = false;
                draggingStart = new Point(0, 0);
            }
        }

        private void renderingBox_MouseMove(object sender, MouseEventArgs e)
        {
            if (renderer != null && renderer.enabled && renderer.scaling.Width != 0.0f &&
                renderer.scaling.Height != 0.0f)
            {
                if (renderer.draggable)
                    renderingBox.Cursor = Cursors.Hand;
                else
                    renderingBox.Cursor = Cursors.Default;

                if (draggingFrame)
                {
                    int offsetX = Convert.ToInt32((e.X - draggingStart.X) * 1.2f);
                    int offsetY = Convert.ToInt32((e.Y - draggingStart.Y) * 1.2f);
                    renderer.DraggedOffset = new Size(renderer.DraggedOffset.Width + offsetX,
                        renderer.DraggedOffset.Height + offsetY);
                    draggingStart = e.Location;
                    refreshRendering(false);
                    return;
                }

                try
                {
                    if ((!splitscreenToolStripMenuItem.Checked && renderer.destination.Contains(e.Location)) ||
                        (splitscreenToolStripMenuItem.Checked && (e.Location.X < renderingBox.Bounds.Width / 2 &&
                        renderer.destination.Contains(e.Location))))
                    {
                        setFrameInfoLabel(new Point(
                            Convert.ToInt32((e.Location.X - renderer.destination.X) / renderer.scaling.Width),
                            Convert.ToInt32((e.Location.Y - renderer.destination.Y) / renderer.scaling.Height)));
                    }
                    else
                        setFrameInfoLabel(new Point(-1, -1));

                    return;
                }
                catch (OverflowException)
                {
                    ErrorCodes.error(ErrorCodes.ERR_OVERFLOW);
                }
            }

            setFrameInfoLabel(new Point(-1, -1));
        }

        private void renderingBox_MouseLeave(object sender, EventArgs e)
        {
            setFrameInfoLabel(new Point(-1, -1));
            draggingFrame = false;
        }

        private void renderingBox_Resize(object sender, EventArgs e)
        {
            refreshFrameInfo();
            refreshRendering(true);
        }

        #endregion

        #region renderingBox context menu handlers

        private void showVectorsToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            showVectorsToolStripMenuItem.Checked = showVectorsToolStripMenuItem1.Checked;
            refreshRendering(true);
        }

        private void movingVectorsToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            movingVectorsToolStripMenuItem.Checked = movingVectorsToolStripMenuItem1.Checked;
            refreshRendering(true);
        }

        private void splitscreenToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            splitscreenToolStripMenuItem.Checked = splitscreenToolStripMenuItem1.Checked;
            refreshRendering(true);
        }

        #endregion

        #region mainMenu handlers

        private void aktualisierenToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (reloadPlugins(false) == PluginReloadAction.praFailed)
                statusBarLabel.Text = "Could not load any plugins.";
        }

        private void processDataToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (pluginManager.activeImportPlugin != null && pluginManager.activeImportPlugin.frameCount > 1)
            {
                int procPluginIndex = pluginManager.processingPluginIndex;
                if ((sender as ToolStripMenuItem).Tag != null)
                    procPluginIndex = (int)(sender as ToolStripMenuItem).Tag;

                processMedia(procPluginIndex);
                fillDataTree();
            }
            else
                MessageBox.Show("No frames to analyze. Import media first.",
                    "No frames", MessageBoxButtons.OK, MessageBoxIcon.Information);
        }

        private void programSettingsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            configureProgram(0);
        }

        private void showVectorsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            showVectorsToolStripMenuItem1.Checked = showVectorsToolStripMenuItem.Checked;
            refreshRendering(true);
        }

        private void movingVectorsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            movingVectorsToolStripMenuItem1.Checked = movingVectorsToolStripMenuItem.Checked;
            refreshRendering(true);
        }

        private void splitscreenToolStripMenuItem_Click(object sender, EventArgs e)
        {
            splitscreenToolStripMenuItem1.Checked = splitscreenToolStripMenuItem.Checked;
            refreshRendering(true);
        }

        private void pluginsettingsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            pluginManager.configure(reloadPlugins, progressCallback);
            refreshRendering(true);
        }

        private void newToolStripMenuItem_Click(object sender, EventArgs e)
        {
            newProject();
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (projectManager.modified)
            {
                DialogResult result = DialogResult.No;

                if (configManager.configuration.askBeforeClosingModified)
                    result = MessageBox.Show("Current project was modified. Save first?",
                        "Modified", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question);

                if (result == DialogResult.Cancel)
                    return;

                if (result == DialogResult.Yes)
                    saveProject(true);
            }
            loadProject(true);
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            saveProject(true);
        }

        private void saveAsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            saveProject(false);
        }

        private void closeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            closeProject(false);
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (closeProject(false))
                Close();
        }

        private void propertiesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (projectManager.showForm() == DialogResult.OK)
            {
                refreshProjectLayout();
                modify();
            }
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            showAboutDialog();
        }

        private void onlineHelpToolStripMenuItem_Click(object sender, EventArgs e)
        {
            openWebpage();
        }

        private void reportBugToolStripMenuItem_Click(object sender, EventArgs e)
        {
            reportBug();
        }

        #endregion

        #region framesTrackBar handlers

        private void prevFrameButton_Click(object sender, EventArgs e)
        {
            prevFrame();
        }

        private void nextFrameButton_Click(object sender, EventArgs e)
        {
            nextFrame();
        }

        private void processingComboBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            setPlugins();
        }

        private void firstFrameButton_Click(object sender, EventArgs e)
        {
            firstFrame();
        }

        private void lastFrameButton_Click(object sender, EventArgs e)
        {
            lastFrame();
        }

        private void removeMarkerButton_Click(object sender, EventArgs e)
        {
            if (framesTrackBar != null)
            {
                framesTrackBar.removeMarker(framesTrackBar.Value);
                if ((dataProcessor != null) && (dataProcessor.currentRun != null))
                    for (int i = 0; i < dataProcessor.currentRun.notes.Count; i++)
                        if ((dataProcessor.currentRun.notes[i] as IndexedNote).index == framesTrackBar.Value)
                        {
                            dataProcessor.currentRun.notes.RemoveAt(i);
                            break;
                        }
                modify();
            }
            refreshFrameInfo();
        }



        private void zoomInToolStripMenuItem_Click(object sender, EventArgs e)
        {
            zoomIn();
        }

        private void zoomOutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            zoomOut();
        }

        private void fitWindowToolStripMenuItem_Click(object sender, EventArgs e)
        {
            zoomFitWindow();
        }

        private void zoom50ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            adjustZoom(50);
        }

        private void zoom100ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            adjustZoom(100);
        }

        private void zoom200ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            adjustZoom(200);
        }

        private void playFramesButton_Click(object sender, EventArgs e)
        {
            playFrames();
        }

        private void pauseFramesButton_Click(object sender, EventArgs e)
        {
            pausePlayingFrames();
        }

        private void frameBarContainer_ContentPanel_Resize(object sender, EventArgs e)
        {
            if (framesTrackBar != null)
                framesTrackBar.Invalidate();
        }

        private void addStartMarkerButton_Click(object sender, EventArgs e)
        {
            if (framesTrackBar != null)
                framesTrackBar.addMarker(framesTrackBar.Value, "Processing start", FramesTrackBar.Marker.MarkerType.mtRight);
            refreshFrameInfo();
        }

        private void addEndMarkerButton_Click(object sender, EventArgs e)
        {
            if (framesTrackBar != null)
                framesTrackBar.addMarker(framesTrackBar.Value, "Processing end", FramesTrackBar.Marker.MarkerType.mtLeft);
            refreshFrameInfo();
        }

        private void addNoteMarkerButton_Click(object sender, EventArgs e)
        {
            if (framesTrackBar != null)
            {
                framesTrackBar.addMarker(framesTrackBar.Value, "Enter note...",
                    FramesTrackBar.Marker.MarkerType.mtNote);
                if (!framesTrackBar.editMarker(framesTrackBar.Value))
                    framesTrackBar.removeMarker(framesTrackBar.Value);
                else
                {
                    dataProcessor.currentRun.addNote(framesTrackBar.Value,
                        framesTrackBar.getMarkerFromIndex(framesTrackBar.Value).Note);
                    modify();
                }
            }
            refreshFrameInfo();
        }

        private void editNoteMarkerButton_Click(object sender, EventArgs e)
        {
            if (framesTrackBar != null)
            {
                if (framesTrackBar.editMarker(framesTrackBar.Value))
                {
                    foreach (IndexedNote note in dataProcessor.currentRun.notes)
                        if (note.index == framesTrackBar.Value)
                        {
                            note.note = framesTrackBar.getMarkerFromIndex(framesTrackBar.Value).Note;
                            break;
                        }
                    modify();
                }
            }
            refreshFrameInfo();
        }

        private void jumpStartMarkerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            jumpStartMarker();
        }

        private void jumpEndMarkerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            jumpEndMarker();
        }

        private void runsContextMenu_Opening(object sender, CancelEventArgs e)
        {
            deleteRunToolStripMenuItem.Enabled = dataRunsGrid.SelectedRows.Count > 0;
            deleteAllToolStripMenuItem.Enabled = dataRunsGrid.Rows.Count > 0;
        }

        private void deleteRunToolStripMenuItem_Click(object sender, EventArgs e)
        {
            deleteRun(dataRunsGrid.SelectedRows[0].Index, false);
        }

        private void deleteAllToolStripMenuItem_Click(object sender, EventArgs e)
        {
            deleteRun(0, true);
        }

        #endregion 

        #region data view handlers

        private void dataViewTree_DoubleClick(object sender, EventArgs e)
        {
            if (dataViewTree.SelectedNode != null && dataViewTree.SelectedNode.Level == 0 &&
                dataViewTree.SelectedNode.Nodes.Count == 0)
                fillDataTreeSub(dataViewTree.SelectedNode);
        }

        private void dataRunsGrid_SelectionChanged(object sender, EventArgs e)
        {
            if (dataRunsGrid.SelectedRows.Count > 0 && dataRunsGrid.SelectedRows[0].Index < dataProcessor.dataRuns.Count)
            {
                dataProcessor.setRun(dataRunsGrid.SelectedRows[0].Index);
                deleteSelectedToolStripMenuItem.Enabled = true;

                if (framesTrackBar != null)
                {
                    framesTrackBar.removeAllMarkers();
                    framesTrackBar.addMarker(dataProcessor.currentRun.firstFrame, "Processing start",
                        FramesTrackBar.Marker.MarkerType.mtRight);
                    framesTrackBar.addMarker(dataProcessor.currentRun.lastFrame, "Processing end",
                        FramesTrackBar.Marker.MarkerType.mtLeft);
                    foreach (IndexedNote note in dataProcessor.currentRun.notes)
                        framesTrackBar.addMarker(note.index, note.note, FramesTrackBar.Marker.MarkerType.mtNote);
                }

                refreshFrameInfo();
                refreshDataViewTree();
                refreshRendering(true);
            }
            else
                deleteSelectedToolStripMenuItem.Enabled = false;
        }

        #endregion

        #region main toolbar handlers

        private void processDataButton_Click(object sender, EventArgs e)
        {
            processDataToolStripMenuItem_Click(processDataToolStripMenuItem, e);
        }

        #endregion

        #region sidebar handlers

        private void vectorCountTrackBar_ValueChanged(object sender, EventArgs e)
        {
            dataProcessor.vectorCount = vectorCountTrackBar.Value;
            vectorCountLabel.Text = "Vectors: " + (dataProcessor.vectorCount * dataProcessor.vectorCount).ToString();
            modify();
        }

        private void configProcessingButton_Click(object sender, EventArgs e)
        {
            configureProgram(1);
        }

        #endregion

        #endregion
    }
 
}
